# -*- coding: utf-8 -*-
# pylint: disable=bare-except,chained-comparison,invalid-name,len-as-condition,line-too-long,protected-access,redefined-outer-name,too-many-arguments,too-many-boolean-expressions,too-many-branches,too-many-lines,too-many-locals,too-many-statements,unneeded-not,unused-import
u"""
Interfaces for the NAG Mark 28.5 `rand` Chapter.

``rand`` - Random Number Generators

Base interfaces.

We do not intend for these lower-level base interfaces to be called
directly from user code. Please refer to the ``naginterfaces.library``
subpackage instead.

This module is concerned with the generation of sequences of independent
pseudorandom and quasi-random numbers from various distributions, and
models.

For full information please refer to the NAG Library document

https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05intro.html
"""

# NAG Copyright 2017-2022.

from __future__ import absolute_import

import collections as _collections
import logging as _logging
import sys as _sys

from math import ceil
from math import sqrt

from ..base import _nag_enum_consts
from ..base import _nag_p_consts
from ..base import utils

def init_repeat(genid, subid, seed, lseed):
    u"""
    Initializes a pseudorandom number generator to give a repeatable
    sequence.

    ``init_repeat`` initializes the selected base generator, as used by
    the group of pseudorandom number functions (see
    ``init_leapfrog``--``init_skipahead``, ``permute``, ``sample``,
    ``times_garch_asym1``--``times_mv_varma``,
    ``matrix_orthog``--``matrix_2waytable``, ``copula_students_t``,
    ``copula_normal``, ``multivar_students_t``, ``multivar_normal`` and
    ``dist_uniform01``--``int_uniform``) and the quasi-random scrambled
    sequence initialization function, ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05kf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05kff.html

    Parameters
    ----------
    genid : int
        Must contain the type of base generator to use.

        `genid` = 1

            NAG basic generator.

        `genid` = 2

            Wichmann Hill I generator.

        `genid` = 3

            Mersenne Twister.

        `genid` = 4

            Wichmann Hill II generator.

        `genid` = 5

            ACORN generator.

        `genid` = 6

            L'Ecuyer MRG32k3a generator.

        See [the G05 Introduction] for details of each of the base
        generators.

    subid : int
        If `genid` = 2, `subid` indicates which of the 273
        sub-generators to use. In this case, the ((|`subid`|+272) mod
        273)+1 sub-generator is used.

        If `genid` = 5, `subid` indicates the values of k and p to use,
        where k is the order of the generator, and p controls the size
        of the modulus, M, with M = 2^(p*30).

        If `subid` < 1, the default values of k = 10 and p = 2 are used,
        otherwise values for k and p are calculated from the formula,
        `subid` = k+1000*(p-1).

        If `genid` = 6 and mod(`subid`, 2) = 0 the range of the
        generator is set to (0,1], otherwise the range is set to (0,1);
        in this case the sequence is identical to the implementation of
        MRG32k3a in TestU01 (see L'Ecuyer and Simard (2002)) for
        identical seeds.

        For all other values of `genid`, `subid` is not referenced.

    seed : int, shape (`lseed`)
        The initial (seed) values for the selected base generator. The
        number of initial values required varies with each of the base
        generators.

        If `genid` = 1, one seed is required.

        If `genid` = 2, four seeds are required.

        If `genid` = 3, 624 seeds are required.

        If `genid` = 4, four seeds are required.

        If `genid` = 5, (k+1)p seeds are required, where k and p are
        defined by `subid`.

        For the ACORN generator it is recommended that an odd value is
        used for `seed`[0].

        If `genid` = 6, six seeds are required.

        If insufficient seeds are provided then the first `lseed`-1
        values supplied in `seed` are used and the remaining values are
        randomly generated using the NAG basic generator.

        In such cases the NAG basic generator is initialized using the
        value supplied in `seed`[`lseed`-1].

    lseed : int
        The size of the `seed` array.

    Returns
    -------
    statecomm : dict, RNG communication object
        RNG communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3, 4, 5 or 6.

        (`errno` 3)
            On entry, invalid `seed`.

        (`errno` 4)
            On entry, `lseed` = *<value>*.

            Constraint: `lseed` >= 1.

    Notes
    -----
    ``init_repeat`` selects a base generator through the input value of
    the arguments `genid` and `subid`, and then initializes it based on
    the values given in the array `seed`.

    A given base generator will yield different sequences of random
    numbers if initialized with different values of `seed`.
    Alternatively, the same sequence of random numbers will be generated
    if the same value of `seed` is used.
    It should be noted that there is no guarantee of statistical
    properties between sequences, only within sequences.

    A definition of some of the terms used in this description, along
    with details of the various base generators can be found in [the G05
    Introduction].

    References
    ----------
    L'Ecuyer, P and Simard, R, 2002, `TestU01: a software library in
    ANSI C for empirical testing of random number generators`,
    Departement d'Informatique et de Recherche Operationnelle,
    Universite de Montreal, https://www.iro.umontreal.ca/~lecuyer

    Maclaren, N M, 1989, `The generation of multiple independent
    sequences of pseudorandom numbers`, Appl. Statist. (38), 351--359

    Matsumoto, M and Nishimura, T, 1998, `Mersenne twister: a
    623-dimensionally equidistributed uniform pseudorandom number
    generator`, ACM Transactions on Modelling and Computer Simulations

    Wichmann, B A and Hill, I D, 2006, `Generating good pseudo-random
    numbers`, Computational Statistics and Data Analysis (51),
    1614--1622

    Wikramaratna, R S, 1989, `ACORN - a new method for generating
    sequences of uniformly distributed pseudo-random numbers`, Journal
    of Computational Physics (83), 16--31

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    :meth:`naginterfaces.library.examples.rand.copula_students_t_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.init_repeat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05kfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, genid.
        utils._BYREF_INT, # Input, subid.
        utils._EngineIntArrayType.c_type, # Input, seed[lseed].
        utils._BYREF_INT, # Input, lseed.
        utils._EngineIntArrayType.c_type, # Output, state[lstate].
        utils._BYREF_INT, # Input/Output, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _genid = utils._EngineIntScalarType(
        genid, {'fun_name': fun_name, 'entity': 'genid'},
    ).to_ctypes()
    _subid = utils._EngineIntScalarType(
        subid, {'fun_name': fun_name, 'entity': 'subid'},
    ).to_ctypes()
    _lseed = utils._EngineIntScalarType(
        lseed, {'fun_name': fun_name, 'entity': 'lseed'},
    ).to_ctypes()
    _lseed_py = utils._EngineIntScalarType.to_py(_lseed)
    _seed = utils._EngineIntArrayType(
        seed, {'fun_name': fun_name, 'entity': 'seed'},
        exp_shape=(_lseed_py,),
    ).to_ctypes()
    _lstate = utils._EngineIntScalarType.c_type(-1)
    _lstate_py = utils._EngineIntScalarType.to_py(_lstate)
    _state = utils._EngineIntArrayType.empty_ctypes(_lstate_py)
    _statecomm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _genid,
        _subid,
        _seed,
        _lseed,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == -1:
        _lstate_py = utils._EngineIntScalarType.to_py(_lstate)
        _state = utils._EngineIntArrayType.empty_ctypes(_lstate_py)
        _statecomm_py['state'] = _state
        engine_ctypes(
            _en_data,
            _genid,
            _subid,
            _seed,
            _lseed,
            _state,
            _lstate,
            _errbuf,
            _ierr,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, genid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: genid = 1, 2, 3, 4, 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid seed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lseed = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lseed >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _statecomm_py
    _err_class.manage()
    return return_data

def init_nonrepeat(genid, subid):
    u"""
    Initializes a pseudorandom number generator to give a non-repeatable
    sequence.

    ``init_nonrepeat`` initializes the selected base generator to
    generate a non-repeatable sequence of variates.
    The base generator can then be used by the group of pseudorandom
    number functions (see ``init_leapfrog``--``init_skipahead``,
    ``permute``, ``sample``, ``times_garch_asym1``--``times_mv_varma``,
    ``matrix_orthog``--``matrix_2waytable``, ``copula_students_t``,
    ``copula_normal``, ``multivar_students_t``, ``multivar_normal`` and
    ``dist_uniform01``--``int_uniform``) and the quasi-random scrambled
    sequence initialization function, ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05kg

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05kgf.html

    Parameters
    ----------
    genid : int
        Must contain the type of base generator to use.

        `genid` = 1

            NAG basic generator.

        `genid` = 2

            Wichmann Hill I generator.

        `genid` = 3

            Mersenne Twister.

        `genid` = 4

            Wichmann Hill II generator.

        `genid` = 5

            ACORN generator.

        `genid` = 6

            L'Ecuyer MRG32k3a generator.

        See [the G05 Introduction] for details of each of the base
        generators.

    subid : int
        If `genid` = 2, `subid` indicates which of the 273
        sub-generators to use. In this case, the ((|`subid`|+272) mod
        273)+1 sub-generator is used.

        If `genid` = 5, `subid` indicates the values of k and p to use,
        where k is the order of the generator, and p controls the size
        of the modulus, M, with M = 2^(p*30).

        If `subid` < 1, the default values of k = 10 and p = 2 are used,
        otherwise values for k and p are calculated from the formula,
        `subid` = k+1000*(p-1).

        If `genid` = 6 and mod(`subid`, 2) = 0 the range of the
        generator is set to (0,1], otherwise the range is set to (0,1);
        in this case the sequence is identical to the implementation of
        MRG32k3a in TestU01 (see L'Ecuyer and Simard (2002)) for
        identical seeds.

        For all other values of `genid`, `subid` is not referenced.

    Returns
    -------
    statecomm : dict, RNG communication object
        RNG communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3, 4, 5 or 6.

    Notes
    -----
    ``init_nonrepeat`` selects a base generator through the input value
    of the arguments `genid` and `subid`, and then initializes it based
    on the values taken from the real-time clock, resulting in the same
    base generator yielding different sequences of random numbers each
    time the calling program is run.
    It should be noted that there is no guarantee of statistical
    properties between sequences, only within sequences.

    A definition of some of the terms used in this description, along
    with details of the various base generators can be found in [the G05
    Introduction].

    References
    ----------
    L'Ecuyer, P and Simard, R, 2002, `TestU01: a software library in
    ANSI C for empirical testing of random number generators`,
    Departement d'Informatique et de Recherche Operationnelle,
    Universite de Montreal, https://www.iro.umontreal.ca/~lecuyer

    Maclaren, N M, 1989, `The generation of multiple independent
    sequences of pseudorandom numbers`, Appl. Statist. (38), 351--359

    Matsumoto, M and Nishimura, T, 1998, `Mersenne twister: a
    623-dimensionally equidistributed uniform pseudorandom number
    generator`, ACM Transactions on Modelling and Computer Simulations

    Wichmann, B A and Hill, I D, 2006, `Generating good pseudo-random
    numbers`, Computational Statistics and Data Analysis (51),
    1614--1622

    Wikramaratna, R S, 1989, `ACORN - a new method for generating
    sequences of uniformly distributed pseudo-random numbers`, Journal
    of Computational Physics (83), 16--31
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.init_nonrepeat'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05kgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, genid.
        utils._BYREF_INT, # Input, subid.
        utils._EngineIntArrayType.c_type, # Output, state[lstate].
        utils._BYREF_INT, # Input/Output, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _genid = utils._EngineIntScalarType(
        genid, {'fun_name': fun_name, 'entity': 'genid'},
    ).to_ctypes()
    _subid = utils._EngineIntScalarType(
        subid, {'fun_name': fun_name, 'entity': 'subid'},
    ).to_ctypes()
    _lstate = utils._EngineIntScalarType.c_type(-1)
    _lstate_py = utils._EngineIntScalarType.to_py(_lstate)
    _state = utils._EngineIntArrayType.empty_ctypes(_lstate_py)
    _statecomm_py = {}
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _genid,
        _subid,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == -1:
        _lstate_py = utils._EngineIntScalarType.to_py(_lstate)
        _state = utils._EngineIntArrayType.empty_ctypes(_lstate_py)
        _statecomm_py['state'] = _state
        engine_ctypes(
            _en_data,
            _genid,
            _subid,
            _state,
            _lstate,
            _errbuf,
            _ierr,
            _errbuf._length_,
        )
        utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == -1:
            _err_class = utils._EngineNoErrorExit()
        elif _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, genid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: genid = 1, 2, 3, 4, 5 or 6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _statecomm_py
    _err_class.manage()
    return return_data

def init_leapfrog(n, k, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using leap-frog.

    ``init_leapfrog`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the leap-frog
    method.

    For full information please refer to the NAG Library document for
    g05kh

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05khf.html

    Parameters
    ----------
    n : int
        n, the total number of sequences required.

    k : int
        k, the number of the current sequence.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: 0 < `k` <= `n`.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 4)
            On entry, cannot use leap-frog with the base generator
            defined by `statecomm`['state'].

    Notes
    -----
    ``init_leapfrog`` adjusts a base generator to allow multiple,
    independent, sequences of pseudorandom numbers to be generated via
    the leap-frog method (see [the G05 Introduction] for details).

    If, prior to calling ``init_leapfrog`` the base generator defined by
    `statecomm`['state'] would produce random numbers x_1,x_2,x_3,...,
    then after calling ``init_leapfrog`` the generator will produce
    random numbers x_k,x_{k+n},x_{k+2n},x_{k+3n},....

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``init_leapfrog``.

    The leap-frog algorithm can be used in conjunction with the NAG
    basic generator, both the Wichmann--Hill I and Wichmann--Hill II
    generators, the Mersenne Twister and L'Ecuyer.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.init_leapfrog'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05khft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _k,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 < k <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cannot use leap-frog with the base generator defined by statecomm[\'state\'].',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def init_skipahead(n, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using skip-ahead.

    ``init_skipahead`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the skip-ahead
    method.

    The base pseudorandom number sequence defined by
    `statecomm`['state'] is advanced n places.

    For full information please refer to the NAG Library document for
    g05kj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05kjf.html

    Parameters
    ----------
    n : int
        n, the number of places to skip ahead.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 3)
            On entry, cannot use skip-ahead with the base generator
            defined by `statecomm`['state'].

        (`errno` 4)
            On entry, the base generator is Mersenne Twister, but the
            `statecomm`['state'] vector defined on initialization is not
            large enough to perform a skip ahead. See the initialization
            function ``init_repeat`` or ``init_nonrepeat``.

    Notes
    -----
    ``init_skipahead`` adjusts a base generator to allow multiple,
    independent, sequences of pseudorandom numbers to be generated via
    the skip-ahead method (see [the G05 Introduction] for details).

    If, prior to calling ``init_skipahead`` the base generator defined
    by `statecomm`['state'] would produce random numbers
    x_1,x_2,x_3,..., then after calling ``init_skipahead`` the generator
    will produce random numbers x_{n+1},x_{n+2},x_{n+3},....

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``init_skipahead``.

    The skip-ahead algorithm can be used in conjunction with any of the
    six base generators discussed in submodule ``rand``.

    References
    ----------
    Haramoto, H, Matsumoto, M, Nishimura, T, Panneton, F and L'Ecuyer,
    P, 2008, `Efficient jump ahead for F2-linear random number
    generators`, INFORMS J. on Computing (20(3)), 385--390

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.init_skipahead'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05kjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cannot use skip-ahead with the base generator defined by statecomm[\'state\'].',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the base generator is Mersenne Twister, but',
                    'the statecomm[\'state\'] vector defined on initialization is not large enough',
                    'to perform a skip ahead. See the initialization function init_repeat or init_nonrepeat.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def init_skipahead_power2(n, statecomm):
    u"""
    Primes a pseudorandom number generator for generating multiple
    streams using skip-ahead, skipping ahead a power of 2.

    ``init_skipahead_power2`` allows for the generation of multiple,
    independent, sequences of pseudorandom numbers using the skip-ahead
    method.
    The base pseudorandom number sequence defined by
    `statecomm`['state'] is advanced 2^n places.

    For full information please refer to the NAG Library document for
    g05kk

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05kkf.html

    Parameters
    ----------
    n : int
        n, where the number of places to skip-ahead is defined as 2^n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 3)
            On entry, cannot use skip-ahead with the base generator
            defined by `statecomm`['state'].

        (`errno` 4)
            On entry, the `statecomm`['state'] vector defined on
            initialization is not large enough to perform a skip-ahead
            (applies to Mersenne Twister base generator). See the
            initialization function ``init_repeat`` or
            ``init_nonrepeat``.

    Notes
    -----
    ``init_skipahead_power2`` adjusts a base generator to allow
    multiple, independent, sequences of pseudorandom numbers to be
    generated via the skip-ahead method (see [the G05 Introduction] for
    details).

    If, prior to calling ``init_skipahead_power2`` the base generator
    defined by `statecomm`['state'] would produce random numbers
    x_1,x_2,x_3,..., then after calling ``init_skipahead_power2`` the
    generator will produce random numbers
    x_{2^n+1},x_{2^n+2},x_{2^n+3},....

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``init_skipahead_power2``.

    The skip-ahead algorithm can be used in conjunction with any of the
    six base generators discussed in [the G05 Introduction].

    References
    ----------
    Haramoto, H, Matsumoto, M, Nishimura, T, Panneton, F and L'Ecuyer,
    P, 2008, `Efficient jump ahead for F2-linear random number
    generators`, INFORMS J. on Computing (20(3)), 385--390

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.init_skipahead_power2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05kkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, cannot use skip-ahead with the base generator defined by statecomm[\'state\'].',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the statecomm[\'state\'] vector defined on initialization is not large',
                    'enough to perform a skip-ahead (applies to Mersenne Twister base generator).',
                    'See the initialization function init_repeat or init_nonrepeat.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = None
    _err_class.manage()
    return return_data

def permute(indx, n, statecomm):
    u"""
    Pseudorandom permutation of an integer vector.

    ``permute`` performs a pseudorandom permutation of a vector of
    integers.

    For full information please refer to the NAG Library document for
    g05nc

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ncf.html

    Parameters
    ----------
    indx : int, shape (`n`), modified in place
        `On entry`: the n integer values to be permuted.

        `On exit`: the n permuted integer values.

    n : int
        The number of values to be permuted.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``permute`` permutes the elements of an integer array without
    inspecting their values.
    Each of the n! possible permutations of the n values may be regarded
    as being equally probable.

    Even for modest values of n it is theoretically impossible that all
    n! permutations may occur, as n! is likely to exceed the cycle
    length of any of the base generators.
    For practical purposes this is irrelevant, as the time necessary to
    generate all possible permutations is many millenia.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``permute``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.permute'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ncft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input/Output, indx[n].
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _indx_cls = utils._EngineIntArrayType(
        indx, {'fun_name': fun_name, 'entity': 'indx'},
        exp_shape=(_n_py,),
    )
    _indx = _indx_cls.to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _indx,
        _n,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _indx_cls.refresh_py(_indx)
        return_data = None
    _err_class.manage()
    return return_data

def sample(ipop, n, isampl, m, statecomm, lipop):
    u"""
    Pseudorandom sample from an integer vector.

    ``sample`` selects a pseudorandom sample without replacement from an
    integer vector.

    For full information please refer to the NAG Library document for
    g05nd

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ndf.html

    Parameters
    ----------
    ipop : None or int, shape (`lipop`)
        The population to be sampled.

        If `ipop` is None, then the population is assumed to be the set
        of values (1,2,...,`n`) and the array `ipop` is not referenced.

    n : None or int
        Note: if this argument is None then a default value will be
        used, determined as follows: `ipop`.shape[0].

        The number of elements in the population to be sampled.

    isampl : int, shape (`m`), modified in place
        `On exit`: the selected sample.

    m : int
        The sample size.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    lipop : int
        The dimension of the array `ipop`.

    Raises
    ------
    NagValueError
        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 4)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `m` <= `n`.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 101)
            On entry, `lipop` = *<value>* and `n` = *<value>*.

            Constraint: `lipop` = 0 or `lipop` = `n`.

    Notes
    -----
    ``sample`` selects m elements from a population vector `ipop` of
    length n and places them in a sample vector `isampl`.
    Their order in `ipop` will be preserved in `isampl`.
    Each of the (n;m) possible combinations of elements of `isampl` may
    be regarded as being equally probable.

    For moderate or large values of n it is theoretically impossible
    that all combinations of size m may occur, unless m is near 1 or
    near n.
    This is because (n;m) exceeds the cycle length of any of the base
    generators.
    For practical purposes this is irrelevant, as the time taken to
    generate all possible combinations is many millenia.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``sample``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.sample'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ndft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, ipop[lipop].
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Output, isampl[m].
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_INT, # Input, lipop.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _lipop = utils._EngineIntScalarType(
        lipop, {'fun_name': fun_name, 'entity': 'lipop'},
    ).to_ctypes()
    if n is None:
        if ipop is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'ipop'},
                'must not be None when n is None',
            )
    if ipop is None:
        if n is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'n'},
                'must not be None when ipop is None',
            )
    if n is None:
        _n = utils._EngineIntScalarType(
            utils._EngineIntArrayType.get_shape(ipop, {'fun_name': fun_name, 'entity': 'ipop'}, exp_rank=1, allow_none=True)[0], {'fun_name': fun_name, 'entity': 'n'},
            allow_none=True,
        ).to_ctypes()
    else:
        _n = utils._EngineIntScalarType(
            n, {'fun_name': fun_name, 'entity': 'n'},
            allow_none=True,
        ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ipop = utils._EngineIntArrayType(
        ipop, {'fun_name': fun_name, 'entity': 'ipop'},
        exp_shape=(_n_py,),
        check_contents=n is None,
        allow_none=True,
    ).to_ctypes(can_reference_contents=n is None)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _isampl_cls = utils._EngineIntArrayType(
        isampl, {'fun_name': fun_name, 'entity': 'isampl'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _isampl = _isampl_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ipop,
        _n,
        _isampl,
        _m,
        _state,
        _lstate,
        _lipop,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= m <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 101 and _ifmt == 101:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lipop = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: lipop = 0 or lipop = n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _isampl_cls.refresh_py(_isampl)
        return_data = None
    _err_class.manage()
    return return_data

def sample_wgt(order, wt, ipop, n, isampl, m, statecomm):
    u"""
    Pseudorandom sample, without replacement, unequal weights.

    ``sample_wgt`` selects a pseudorandom sample, without replacement
    and allowing for unequal probabilities.

    For full information please refer to the NAG Library document for
    g05ne

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05nef.html

    Parameters
    ----------
    order : str, length 1
        A flag indicating the sorted status of the `wt` vector.

        `order` = 'A'

            `wt` is sorted in ascending order,

        `order` = 'D'

            `wt` is sorted in descending order,

        `order` = 'U'

            `wt` is unsorted and ``sample_wgt`` will sort the weights
            prior to using them.

        Irrespective of the value of `order`, no checks are made on the
        sorted status of `wt`, e.g., it is possible to supply `order` =
        'A', even when `wt` is not sorted.

        In such cases the `wt` array will not be sorted internally, but
        ``sample_wgt`` will still work correctly except, possibly, in
        cases of extreme weight values.

        It is usually more efficient to specify a value of `order` that
        is consistent with the status of `wt`.

    wt : float, shape (`n`)
        w_i, the relative probability weights. These weights need not
        sum to 1.0.

    ipop : None or int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ipop` is not None: `n`; otherwise: 0.

        The population to be sampled. If `ipop` is None then the
        population is assumed to be the set of values (1,2,...,`n`) and
        the array `ipop` is not referenced. Elements of `ipop` with the
        same value are not combined, therefore, if `wt`[i-1] !=
        0,`wt`[j-1] != 0 and i != j then there is a nonzero probability
        that the sample will contain both `ipop`[i-1] and `ipop`[j-1].
        If `ipop`[i-1] = `ipop`[j-1] then that value can appear in
        `isampl` more than once.

    n : int
        n, the size of the population.

    isampl : int, shape (`m`), modified in place
        `On exit`: the selected sample.

    m : int
        m, the size of the sample required.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `order` had an illegal value.

        (`errno` 2)
            On entry, at least one weight was less than zero.

        (`errno` 5)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 7)
            On entry, `m` = *<value>* and `n` = *<value>*.

            Constraint: 0 <= `m` <= `n`.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 21)
            On entry, `m` = *<value>*, number of nonzero weights =
            *<value>*.

            Constraint: must be at least `m` nonzero weights.

    Notes
    -----
    ``sample_wgt`` selects m elements from either the set of values
    (1,2,...,n) or a supplied population vector of length n.
    The probability of selecting the ith element is proportional to a
    user-supplied weight, w_i.
    Each element will appear at most once in the sample, i.e., the
    sampling is done without replacement.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``sample_wgt``.
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.sample_wgt'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05neft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, iorder.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineIntArrayType.c_type, # Input, ipop[lipop].
        utils._BYREF_INT, # Input, lipop.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Output, isampl[m].
        utils._BYREF_INT, # Input, m.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _order = utils._EngineCharScalarType(
        order, {'fun_name': fun_name, 'entity': 'order'}, exp_chrlen=1,
    ).to_ctypes()
    _order_py = utils._EngineCharScalarType.to_py(_order)
    if _order_py.upper() == 'A':
        _iorder = utils._EngineIntScalarType.c_type(1)
    elif _order_py.upper() == 'D':
        _iorder = utils._EngineIntScalarType.c_type(2)
    elif _order_py.upper() == 'U':
        _iorder = utils._EngineIntScalarType.c_type(0)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, order had an illegal value.',
            ],
            errcodes=(1,),
        )
        _err_class.manage()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _lipop_py = utils._EngineIntArrayType.get_shape(ipop, {'fun_name': fun_name, 'entity': 'ipop'}, exp_rank=1, allow_none=True)[0]
    _lipop = utils._EngineIntScalarType(
        _lipop_py, {'fun_name': fun_name, 'entity': '_lipop'},
    ).to_ctypes()
    if ipop is not None:
        _ipop_dim1 = _n_py
    else:
        _ipop_dim1 = 0
    exp_shape_ipop = (_ipop_dim1,)
    _ipop = utils._EngineIntArrayType(
        ipop, {'fun_name': fun_name, 'entity': 'ipop'},
        exp_shape=exp_shape_ipop,
        allow_none=True,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _isampl_cls = utils._EngineIntArrayType(
        isampl, {'fun_name': fun_name, 'entity': 'isampl'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _isampl = _isampl_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _iorder,
        _wt,
        _ipop,
        _lipop,
        _n,
        _isampl,
        _m,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one weight was less than zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= m <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'number of nonzero weights = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: must be at least m nonzero weights.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _isampl_cls.refresh_py(_isampl)
        return_data = None
    _err_class.manage()
    return return_data

def resample(rtype, n, wt, ipop, m, nrs, otype, isampl, statecomm):
    u"""
    Pseudorandom resampling, unequal weights.

    ``resample`` selects a pseudorandom sample (or resamples) from a
    vector of integers based on the supplied weights.

    For full information please refer to the NAG Library document for
    g05nf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05nff.html

    Parameters
    ----------
    rtype : int
        A flag indicating the resampling method to use.

        `rtype` = 1

            Multinomial resampling will be used.

        `rtype` = 2

            Systematic resampling will be used.

        `rtype` = 3

            Residual resampling will be used.

    n : int
        n, the size of the population.

    wt : float, shape (`n`)
        w_i, the weights. These weights need not sum to 1.0.

    ipop : None or int, shape (:)
        Note: the required length for this argument is determined as
        follows: if `ipop` is not None: `n`; otherwise: 0.

        V, the vector to be sampled from. If `ipop` is None then the V
        is assumed to be the set of values (1,2,...,`n`).

        Elements of `ipop` with the same value are not combined,
        therefore, if `wt`[i] != 0,`wt`[j] != 0 and i != j then there is
        a nonzero probability that the sample will contain both
        `ipop`[i] and `ipop`[j], irrespective of their values.

        If the values to be returned in `isampl` are counts, i.e.,
        `otype` = 2, then `ipop` is not referenced.

    m : int
        m, the size of the sample required.

    nrs : int
        The number of times to resample.

    otype : int
        A flag indicating what is returned in `isampl`.

        `otype` = 1

            The values returned in `isampl` are taken from the
            population.

        `otype` = 2

            The values returned in `isampl` are counts.

        See `isampl` for more information.

    isampl : int, shape (:, :), modified in place
        `On exit`: the selected samples.

        If `otype` = 1 then each column of `isampl` contains the m
        values from V that make up the sample.

        If `otype` = 2 then `isampl`[j-1,k-1] contains the number of
        times that V_j appears in the kth sample.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `rtype` = *<value>*.

            Constraint: `rtype` = 1, 2 or 3.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 31)
            On entry, i = *<value>* and `wt`[i-1] = *<value>*.

            Constraint: `wt`[i-1] >= 0.0.

        (`errno` 32)
            On entry, all the weights are zero.

            Constraint: at least one weight must be nonzero.

        (`errno` 61)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 71)
            On entry, `nrs` = *<value>*.

            Constraint: `nrs` >= 0.

        (`errno` 81)
            On entry, `otype` = *<value>*.

            Constraint: `otype` = 1 or 2.

        (`errno` 121)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 91)
            There was no random component to the sample. Check the
            sample size and weights are as expected.

    Notes
    -----
    Given a vector V, of n integer values, ``resample`` selects m
    elements with the probability of selecting V_j proportional to a
    user-supplied weight, w_j.
    The sampling is done with replacement, so that each value, V_j, may
    appear more than once in the sample.

    The most common usage case for ``resample`` is where V was obtained
    using some other sampling method, for example, importance sampling.
    In such a case, this function is used to perform resampling.

    Several methods of calculating N_j, the number of times V_j appears
    in the sample, are available:

    -   Multinomial Resampling: The vector of counts; {N_j:j =
        1,2,...,n} is drawn from a multinomial distribution with
        probabilities given by the normalised weights, w~.

    -   Systematic Resampling:

            N_j = |{u_i:i = 1,2,...,m and sum_k = 1^j-1w~_k < u_i <=
            sum_k = 1^jw~_k}|

        where w~ are the normalised weights, u_1~U(0,1/m), u_i =
        u_1+(i-1)/m, for i = 2,3,...,m and sum_k = 1^0w~_k is defined to
        be zero. In other words, N_j is the number of shifted and scaled
        uniform variates contained in bins defined by the partial sums
        of normalised weights.

    -   Residual Resampling: N_j = N_j^S+N_j^R, where N_j^S =
        floor(m*w~_j) (i.e. the integer part of m*w~_j), and the vector
        of residual counts, {N_j^R:j = 1,2,...,n} is drawn from a
        multinomial distribution with probabilities given by
        {w~_j-(N_j^S)/m:j = 1,2,...,n}.

    See ``int_multinomial`` for more information on the multinomial
    distribution and Douc `et al.` (2005) for more details on the
    resampling methods.

    If multiple samples are requested (`nrs` > 1) then the chosen
    resampling method is performed independently for each sample.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``resample``.

    References
    ----------
    Douc, R, Cappe, O, and Moulines, E, 2005, `Comparison of resampling
    schemes for particle filtering`, Proceedings of the 4th
    International Symposium on Image and Signal Processing and Analysis,
    64--69, https://dx.doi.org/10.1109/ISPA.2005.195385

    Li, T, Bolic, M, and Djuric, P M, 2015, `Resampling Methods for
    Particle Filtering: Classification, implementation, and strategies`,
    IEEE Signal Processing Magazine (vol. 32, no. 3), 70--86,
    https://dx.doi.org/10.1109/MSP.2014.2330626
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.resample'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05nfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, rtype.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, wt[n].
        utils._EngineIntArrayType.c_type, # Input, ipop[lipop].
        utils._BYREF_INT, # Input, lipop.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, nrs.
        utils._BYREF_INT, # Input, otype.
        utils._EngineIntArrayType.c_type, # Output, isampl[ldisampl, sdisampl].
        utils._BYREF_INT, # Input, ldisampl.
        utils._BYREF_INT, # Input, sdisampl.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((isampl, 'isampl'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _rtype = utils._EngineIntScalarType(
        rtype, {'fun_name': fun_name, 'entity': 'rtype'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _wt = utils._EngineFloat64ArrayType(
        wt, {'fun_name': fun_name, 'entity': 'wt'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _lipop_py = utils._EngineIntArrayType.get_shape(ipop, {'fun_name': fun_name, 'entity': 'ipop'}, exp_rank=1, allow_none=True)[0]
    _lipop = utils._EngineIntScalarType(
        _lipop_py, {'fun_name': fun_name, 'entity': '_lipop'},
    ).to_ctypes()
    if ipop is not None:
        _ipop_dim1 = _n_py
    else:
        _ipop_dim1 = 0
    exp_shape_ipop = (_ipop_dim1,)
    _ipop = utils._EngineIntArrayType(
        ipop, {'fun_name': fun_name, 'entity': 'ipop'},
        exp_shape=exp_shape_ipop,
        allow_none=True,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _nrs = utils._EngineIntScalarType(
        nrs, {'fun_name': fun_name, 'entity': 'nrs'},
    ).to_ctypes()
    _otype = utils._EngineIntScalarType(
        otype, {'fun_name': fun_name, 'entity': 'otype'},
    ).to_ctypes()
    _ldisampl_py = utils._EngineIntArrayType.get_nd(isampl, {'fun_name': fun_name, 'entity': 'isampl'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldisampl = utils._EngineIntScalarType(
        _ldisampl_py, {'fun_name': fun_name, 'entity': '_ldisampl'},
    ).to_ctypes()
    _sdisampl_py = utils._EngineIntArrayType.get_nd(isampl, {'fun_name': fun_name, 'entity': 'isampl'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdisampl = utils._EngineIntScalarType(
        _sdisampl_py, {'fun_name': fun_name, 'entity': '_sdisampl'},
    ).to_ctypes()
    _isampl_cls = utils._EngineIntArrayType(
        isampl, {'fun_name': fun_name, 'entity': 'isampl'},
        exp_shape=(None, None),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _isampl = _isampl_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _rtype,
        _n,
        _wt,
        _ipop,
        _lipop,
        _m,
        _nrs,
        _otype,
        _isampl,
        _ldisampl,
        _sdisampl,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rtype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rtype = 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and wt[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 32 and _ifmt == 32:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, all the weights are zero.',
                    'Constraint: at least one weight must be nonzero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrs = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrs >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, otype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: otype = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'There was no random component to the sample.',
                    'Check the sample size and weights are as expected.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 122 and _ifmt == 122:
            _ierr_fl = 121
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _isampl_cls.refresh_py(_isampl)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def times_garch_asym1(dist, num, ip, iq, theta, gamma, df, ht, et, fcall, comm, statecomm):
    u"""
    Generates a realization of a time series from a GARCH process with
    asymmetry of the form (epsilon_{t-1}+gamma)^2.

    ``times_garch_asym1`` generates a given number of terms of a type I
    AGARCH(p,q) process (see Engle and Ng (1993)).

    For full information please refer to the NAG Library document for
    g05pd

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pdf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    ht : float, shape (`num`), modified in place
        `On exit`: the conditional variances h_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    et : float, shape (`num`), modified in place
        `On exit`: the observations epsilon_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `comm`['r'].

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `fcall` = False, this argument must have been initialized by
        a prior call to ``times_garch_asym1``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i-1] >= 0.0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, sum of `theta`[i-1], for i = 2,3,...,`ip`+`iq`+1
            is >= 1.0: sum = *<value>*.

    Notes
    -----
    A type I AGARCH(p,q) process can be represented by:

        h_t = alpha_0+sum_i = 1^q alpha_i(epsilon_{t-i}+gamma)^2+sum_i =
        1^p beta_i h_{t-i}, t = 1,2,...,T;

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of observations in the
    sequence, epsilon_t is the observed value of the GARCH(p,q) process
    at time t, h_t is the conditional variance at time t, and psi_t the
    set of all information up to time t.
    Symmetric GARCH sequences are generated when gamma is zero,
    otherwise asymmetric GARCH sequences are generated with gamma
    specifying the amount by which positive (or negative) shocks are to
    be enhanced.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_garch_asym1``.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_garch_asym1'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._BYREF_INT, # Input, df.
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._BYREF_LOGICAL, # Input, fcall.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _fcall = utils._EngineLogicalScalarType(
        fcall, {'fun_name': fun_name, 'entity': 'fcall'},
    ).to_ctypes()
    _fcall_py = utils._EngineLogicalScalarType.to_py(_fcall)
    if not _fcall_py:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(2*(_ip_py+_iq_py+2))
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dist,
        _num,
        _ip,
        _iq,
        _theta,
        _gamma,
        _df,
        _ht,
        _et,
        _fcall,
        _r,
        _lr,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist is not valid: dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, theta[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: theta[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip or iq is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of ip = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum of theta[i-1], for i = 2,3,...,ip+iq+1 is >= 1.0:',
                    'sum = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ht_cls.refresh_py(_ht)
        _et_cls.refresh_py(_et)
        return_data = None
    _err_class.manage()
    return return_data

def times_garch_asym2(dist, num, ip, iq, theta, gamma, df, ht, et, fcall, comm, statecomm):
    u"""
    Generates a realization of a time series from a GARCH process with
    asymmetry of the form (|epsilon_{t-1}|+gamma*epsilon_{t-1})^2.

    ``times_garch_asym2`` generates a given number of terms of a type II
    AGARCH(p,q) process (see Engle and Ng (1993)).

    For full information please refer to the NAG Library document for
    g05pe

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pef.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    ht : float, shape (`num`), modified in place
        `On exit`: the conditional variances h_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    et : float, shape (`num`), modified in place
        `On exit`: the observations epsilon_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `comm`['r'].

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `fcall` = False, this argument must have been initialized by
        a prior call to ``times_garch_asym2``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i-1] >= 0.0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, sum of `theta`[i-1], for i = 2,3,...,`ip`+`iq`+1
            is >= 1.0: sum = *<value>*.

    Notes
    -----
    A type II AGARCH(p,q) process can be represented by:

        h_t = alpha_0 +
        sum(alpha_i(|epsilon_{t-i}|+gammaepsilon_{t-i})^2)+ sum(beta_i
        h_{t-i}), t = 1,2,...,T;

    where epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of observations in the
    sequence, epsilon_t is the observed value of the GARCH(p,q) process
    at time t, h_t is the conditional variance at time t, and psi_t the
    set of all information up to time t.
    Symmetric GARCH sequences are generated when gamma is zero,
    otherwise asymmetric GARCH sequences are generated with gamma
    specifying the amount by which positive (or negative) shocks are to
    be enhanced.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_garch_asym2``.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_garch_asym2'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05peft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._BYREF_INT, # Input, df.
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._BYREF_LOGICAL, # Input, fcall.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _fcall = utils._EngineLogicalScalarType(
        fcall, {'fun_name': fun_name, 'entity': 'fcall'},
    ).to_ctypes()
    _fcall_py = utils._EngineLogicalScalarType.to_py(_fcall)
    if not _fcall_py:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(2*(_ip_py+_iq_py+2))
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dist,
        _num,
        _ip,
        _iq,
        _theta,
        _gamma,
        _df,
        _ht,
        _et,
        _fcall,
        _r,
        _lr,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist is not valid: dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, theta[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: theta[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip or iq is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of ip = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum of theta[i-1], for i = 2,3,...,ip+iq+1 is >= 1.0:',
                    'sum = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ht_cls.refresh_py(_ht)
        _et_cls.refresh_py(_et)
        return_data = None
    _err_class.manage()
    return return_data

def times_garch_gjr(dist, num, ip, iq, theta, gamma, df, ht, et, fcall, comm, statecomm):
    u"""
    Generates a realization of a time series from an asymmetric Glosten,
    Jagannathan and Runkle (GJR) GARCH process.

    ``times_garch_gjr`` generates a given number of terms of a GJR
    GARCH(p,q) process (see Glosten `et al.` (1993)).

    For full information please refer to the NAG Library document for
    g05pf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pff.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (`iq`+`ip`+1)
        The first element must contain the coefficient alpha_o, the next
        `iq` elements must contain the coefficients alpha_i, for i =
        1,2,...,q. The remaining `ip` elements must contain the
        coefficients beta_j, for j = 1,2,...,p.

    gamma : float
        The asymmetry parameter gamma for the GARCH(p,q) sequence.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    ht : float, shape (`num`), modified in place
        `On exit`: the conditional variances h_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    et : float, shape (`num`), modified in place
        `On exit`: the observations epsilon_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `comm`['r'].

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `fcall` = False, this argument must have been initialized by
        a prior call to ``times_garch_gjr``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 5)
            On entry, `theta`[*<value>*] = *<value>* and gamma =
            *<value>*.

            Constraint: alpha_i+gamma >= 0.

        (`errno` 7)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 11)
            `ip` or `iq` is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 51)
            On entry, `theta`[*<value>*] = *<value>*.

            Constraint: `theta`[i] >= 0.0.

        (`errno` 52)
            On entry, sum of `theta`[i] = *<value>*.

            Constraint: sum of `theta`[i], for i = 1,2,...,`ip`+`iq` is
            < 1.0.

    Notes
    -----
    A GJR GARCH(p,q) process is represented by:

        h_t = alpha_0+sum_i = 1^q(alpha_i+gamma
        I_{t-i})epsilon_t-i^2+sum_i = 1^p beta_i h_{t-i}, t = 1,2,...,T;

    where I_t = 1 if epsilon_t < 0, I_t = 0 if epsilon_t >= 0, and
    epsilon_t | psi_{t-1} = N(0,h_t) or epsilon_t | psi_{t-1} =
    S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of observations in the
    sequence, epsilon_t is the observed value of the GARCH(p,q) process
    at time t, h_t is the conditional variance at time t, and psi_t the
    set of all information up to time t.
    Symmetric GARCH sequences are generated when gamma is zero,
    otherwise asymmetric GARCH sequences are generated with gamma
    specifying the amount by which negative shocks are to be enhanced.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_garch_gjr``.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_garch_gjr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq+ip+1].
        utils._BYREF_FLOAT64, # Input, gamma.
        utils._BYREF_INT, # Input, df.
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._BYREF_LOGICAL, # Input, fcall.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py+_ip_py+1,),
    ).to_ctypes()
    _gamma = utils._EngineFloat64ScalarType(
        gamma, {'fun_name': fun_name, 'entity': 'gamma'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _fcall = utils._EngineLogicalScalarType(
        fcall, {'fun_name': fun_name, 'entity': 'fcall'},
    ).to_ctypes()
    _fcall_py = utils._EngineLogicalScalarType.to_py(_fcall)
    if not _fcall_py:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(2*(_ip_py+_iq_py+2))
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dist,
        _num,
        _ip,
        _iq,
        _theta,
        _gamma,
        _df,
        _ht,
        _et,
        _fcall,
        _r,
        _lr,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist is not valid: dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, theta[{:s}] = {:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'and gamma = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: alpha_i+gamma >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip or iq is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of ip = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, theta[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: theta[i] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum of theta[i] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sum of theta[i], for i = 1,2,...,ip+iq is < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ht_cls.refresh_py(_ht)
        _et_cls.refresh_py(_et)
        return_data = None
    _err_class.manage()
    return return_data

def times_garch_exp(dist, num, ip, iq, theta, df, ht, et, fcall, comm, statecomm):
    u"""
    Generates a realization of a time series from an exponential GARCH
    (EGARCH) process.

    ``times_garch_exp`` generates a given number of terms of an
    exponential GARCH(p,q) process (see Engle and Ng (1993)).

    For full information please refer to the NAG Library document for
    g05pg

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pgf.html

    Parameters
    ----------
    dist : str, length 1
        The type of distribution to use for epsilon_t.

        `dist` = 'N'

            A Normal distribution is used.

        `dist` = 'T'

            A Student's t-distribution is used.

    num : int
        T, the number of terms in the sequence.

    ip : int
        The number of coefficients, beta_i, for i = 1,2,...,p.

    iq : int
        The number of coefficients, alpha_i, for i = 1,2,...,q.

    theta : float, shape (2*`iq`+`ip`+1)
        The initial parameter estimates for the vector theta. The first
        element must contain the coefficient alpha_o and the next `iq`
        elements must contain the autoregressive coefficients alpha_i,
        for i = 1,2,...,q. The next `iq` elements must contain the
        coefficients phi_i, for i = 1,2,...,q. The next `ip` elements
        must contain the moving average coefficients beta_j, for j =
        1,2,...,p.

    df : int
        The number of degrees of freedom for the Student's
        t-distribution.

        If `dist` = 'N', `df` is not referenced.

    ht : float, shape (`num`), modified in place
        `On exit`: the conditional variances h_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    et : float, shape (`num`), modified in place
        `On exit`: the observations epsilon_t, for t = 1,2,...,T, for
        the GARCH(p,q) sequence.

    fcall : bool
        If `fcall` = True, a new sequence is to be generated, otherwise
        a given sequence is to be continued using the information in
        `comm`['r'].

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `fcall` = False, this argument must have been initialized by
        a prior call to ``times_garch_exp``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `dist` is not valid: `dist` = *<value>*.

        (`errno` 2)
            On entry, `num` = *<value>*.

            Constraint: `num` >= 0.

        (`errno` 3)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 4)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 1.

        (`errno` 6)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 10)
            `ip` or `iq` is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 12)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 20)
            Invalid sequence generated, use different parameters.

    Notes
    -----
    An exponential GARCH(p,q) process is represented by:

        ln(h_t) = alpha_0+sum_i = 1^q alpha_i z_{t-i}+sum_i = 1^q
        phi_i(|z_{t-i}|-E[|z_{t-i}|])+sum_j = 1^p beta_j ln(h_{t-j}), t
        = 1,2,...,T;

    where z_t = (epsilon_t)/sqrt(h_t), E[|z_{t-i}|] denotes the expected
    value of |z_{t-i}|, and epsilon_t | psi_{t-1} = N(0,h_t) or
    epsilon_t | psi_{t-1} = S_t(df,h_t).
    Here S_t is a standardized Student's t-distribution with df degrees
    of freedom and variance h_t, T is the number of observations in the
    sequence, epsilon_t is the observed value of the GARCH(p,q) process
    at time t, h_t is the conditional variance at time t, and psi_t the
    set of all information up to time t.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_garch_exp``.

    References
    ----------
    Bollerslev, T, 1986, `Generalised autoregressive conditional
    heteroskedasticity`, Journal of Econometrics (31), 307--327

    Engle, R, 1982, `Autoregressive conditional heteroskedasticity with
    estimates of the variance of United Kingdom inflation`, Econometrica
    (50), 987--1008

    Engle, R and Ng, V, 1993, `Measuring and testing the impact of news
    on volatility`, Journal of Finance (48), 1749--1777

    Glosten, L, Jagannathan, R and Runkle, D, 1993, `Relationship
    between the expected value and the volatility of nominal excess
    return on stocks`, Journal of Finance (48), 1779--1801

    Hamilton, J, 1994, `Time Series Analysis`, Princeton University
    Press
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_garch_exp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, dist(len=1).
        utils._BYREF_INT, # Input, num.
        utils._BYREF_INT, # Input, ip.
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[2*iq+ip+1].
        utils._BYREF_INT, # Input, df.
        utils._EngineFloat64ArrayType.c_type, # Output, ht[num].
        utils._EngineFloat64ArrayType.c_type, # Output, et[num].
        utils._BYREF_LOGICAL, # Input, fcall.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of dist.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _dist = utils._EngineCharScalarType(
        dist, {'fun_name': fun_name, 'entity': 'dist'}, exp_chrlen=1,
    ).to_ctypes()
    _num = utils._EngineIntScalarType(
        num, {'fun_name': fun_name, 'entity': 'num'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(2*_iq_py+_ip_py+1,),
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _num_py = utils._EngineIntScalarType.to_py(_num)
    _ht_cls = utils._EngineFloat64ArrayType(
        ht, {'fun_name': fun_name, 'entity': 'ht'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _ht = _ht_cls.to_ctypes(can_reference_contents=False)
    _et_cls = utils._EngineFloat64ArrayType(
        et, {'fun_name': fun_name, 'entity': 'et'},
        exp_shape=(_num_py,),
        check_contents=False,
    )
    _et = _et_cls.to_ctypes(can_reference_contents=False)
    _fcall = utils._EngineLogicalScalarType(
        fcall, {'fun_name': fun_name, 'entity': 'fcall'},
    ).to_ctypes()
    _fcall_py = utils._EngineLogicalScalarType.to_py(_fcall)
    if not _fcall_py:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(2*(_ip_py+2*_iq_py+2))
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _dist,
        _num,
        _ip,
        _iq,
        _theta,
        _df,
        _ht,
        _et,
        _fcall,
        _r,
        _lr,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _dist._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, dist is not valid: dist = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, num = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: num >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip or iq is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of ip = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 12
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 20 and _ifmt == 20:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Invalid sequence generated, use different parameters.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _ht_cls.refresh_py(_ht)
        _et_cls.refresh_py(_et)
        return_data = None
    _err_class.manage()
    return return_data

def times_arma(mode, n, xmean, ip, phi, iq, theta, avar, comm, statecomm, x):
    u"""
    Generates a realization of a time series from an ARMA model.

    ``times_arma`` generates a realization of a univariate time series
    from an autoregressive moving average (ARMA) model.
    The realization may be continued or a new realization generated at
    subsequent calls to ``times_arma``.

    For full information please refer to the NAG Library document for
    g05ph

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05phf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate terms in the time series using reference vector set
            up in a prior call to ``times_arma``.

        `mode` = 2

            Set up reference vector and generate terms in the time
            series.

    n : int
        n, the number of observations to be generated.

    xmean : float
        The mean of the time series.

    ip : int
        p, the number of autoregressive coefficients supplied.

    phi : float, shape (`ip`)
        The autoregressive coefficients of the model,
        phi_1,phi_2,...,phi_p.

    iq : int
        q, the number of moving average coefficients supplied.

    theta : float, shape (`iq`)
        The moving average coefficients of the model,
        theta_1,theta_2,...,theta_q.

    avar : float
        sigma^2, the variance of the Normal perturbations.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``times_arma``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or float, shape (`n`), modified in place
        `On exit`, if not None on entry: contains the next n
        observations from the time series.

    Returns
    -------
    var : float
        The proportion of the variance of a term in the series that is
        due to the moving-average (error) terms in the model. The
        smaller this is, the nearer is the model to non-stationarity.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 5)
            On entry, the AR parameters are outside the stationarity
            region.

        (`errno` 6)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 8)
            On entry, `avar` = *<value>*.

            Constraint: `avar` >= 0.0.

        (`errno` 9)
            Reference vector `comm`['r'] has been corrupted or not
            initialized correctly.

        (`errno` 9)
            `ip` or `iq` is not the same as when `comm`['r'] was set up
            in a previous call.

            Previous value of `ip` = *<value>* and `ip` = *<value>*.

            Previous value of `iq` = *<value>* and `iq` = *<value>*.

        (`errno` 11)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    Let the vector x_t, denote a time series which is assumed to follow
    an autoregressive moving average (ARMA) model of the form:

        [table omitted]

    where epsilon_t, is a residual series of independent random
    perturbations assumed to be Normally distributed with zero mean and
    variance sigma^2.
    The parameters {phi_i}, for i = 1,2,...,p, are called the
    autoregressive (AR) parameters, and {theta_j}, for j = 1,2,...,q,
    the moving average (MA) parameters.
    The parameters in the model are thus the p phi values, the q theta
    values, the mean mu and the residual variance sigma^2.

    ``times_arma`` sets up a reference vector containing initial values
    corresponding to a stationary position using the method described in
    Tunnicliffe--Wilson (1979).
    The function can then return a realization of x_1,x_2,...,x_n.
    On a successful exit, the recent history is updated and saved in the
    reference vector `comm`['r'] so that ``times_arma`` may be called
    again to generate a realization of x_{n+1},x_{n+2},..., etc.
    See the description of the argument `mode` in [Parameters] for
    details.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_arma``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Tunnicliffe--Wilson, G, 1979, `Some efficient computational
    procedures for high order ARMA models`, J. Statist. Comput.
    Simulation (8), 301--309
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_arma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05phft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmean.
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, phi[ip].
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[iq].
        utils._BYREF_FLOAT64, # Input, avar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_FLOAT64, # Output, var.
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmean = utils._EngineFloat64ScalarType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _phi = utils._EngineFloat64ArrayType(
        phi, {'fun_name': fun_name, 'entity': 'phi'},
        exp_shape=(_ip_py,),
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_iq_py,),
    ).to_ctypes()
    _avar = utils._EngineFloat64ScalarType(
        avar, {'fun_name': fun_name, 'entity': 'avar'},
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_ip_py+_iq_py+6+max(_ip_py, _iq_py+1))
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _var = utils._EngineFloat64ScalarType.empty_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _xmean,
        _ip,
        _phi,
        _iq,
        _theta,
        _avar,
        _r,
        _lr,
        _state,
        _var,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the AR parameters are outside the stationarity region.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, avar = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: avar >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Reference vector comm[\'r\'] has been corrupted or not initialized correctly.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 91 and _ifmt == 91:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'ip or iq is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of ip = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of iq = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 11
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        _var_py = utils._EngineFloat64ScalarType.to_py(_var)
        return_data = _var_py
    _err_class.manage()
    return return_data

def times_mv_varma(mode, n, k, xmean, ip, phi, iq, theta, var, comm, statecomm, x):
    u"""
    Generates a realization of a multivariate time series from a VARMA
    model.

    ``times_mv_varma`` generates a realization of a multivariate time
    series from a vector autoregressive moving average (VARMA) model.
    The realization may be continued or a new realization generated at
    subsequent calls to ``times_mv_varma``.

    For full information please refer to the NAG Library document for
    g05pj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pjf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector and compute a realization of the
            recent history.

        `mode` = 1

            Generate terms in the time series using reference vector set
            up in a prior call to ``times_mv_varma``.

        `mode` = 2

            Combine the operations of `mode` = 0 or 1.

        `mode` = 3

            A new realization of the recent history is computed using
            information stored in the reference vector, and the
            following sequence of time series values are generated.

        If `mode` = 1 or 3, you must ensure that the reference vector
        `comm`['r'] and the values of `k`, `ip`, `iq`, `xmean`, `phi`,
        `theta`, `var` and ldvar have not been changed between calls to
        ``times_mv_varma``.

    n : int
        n, the number of observations to be generated.

    k : int
        k, the dimension of the multivariate time series.

    xmean : float, shape (`k`)
        mu, the vector of means of the multivariate time series.

    ip : int
        p, the number of autoregressive parameter matrices.

    phi : float, shape (`k`*`k`*`ip`)
        Must contain the elements of the `ip`*`k`*`k` autoregressive
        parameter matrices of the model, phi_1,phi_2,...,phi_p. If `phi`
        is considered as a three-dimensional array, dimensioned as
        `phi`[`k`-1,`k`-1,`ip`-1], the (i,j)th element of phi_l would be
        stored in `phi`[i-1,j-1,l-1]; that is,
        `phi`[(l-1)*k*k+(j-1)*k+i-1] must be set equal to the (i,j)th
        element of phi_l, for l = 1,2,...,p, i = 1,2,...,k and j =
        1,2,...,k.

    iq : int
        q, the number of moving average parameter matrices.

    theta : float, shape (`k`*`k`*`iq`)
        Must contain the elements of the `iq`*`k`*`k` moving average
        parameter matrices of the model, theta_1,theta_2,...,theta_q. If
        `theta` is considered as a three-dimensional array, dimensioned
        as `theta`(`k`,`k`,`iq`), the (i,j)th element of theta_l would
        be stored in `theta`[i-1,j-1,l-1]; that is,
        `theta`[(l-1)*k*k+(j-1)*k+i-1] must be set equal to the (i,j)th
        element of theta_l, for j = 1,2,...,k, for i = 1,2,...,k, for l
        = 1,2,...,q.

    var : float, shape (`k`, `k`)
        `var`[i-1,j-1] must contain the (i,j)th element of Sigma, for j
        = 1,2,...,`k`, for i = 1,2,...,`k`. Only the lower triangle is
        required.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1 or 3, this argument must have been initialized by
        a prior call to ``times_mv_varma``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`k`, `n`), modified in place
        `On exit`: `x`[i-1,t-1] will contain a realization of the ith
        component of X_t, for t = 1,2,...,n, for i = 1,2,...,k.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 1.

        (`errno` 5)
            On entry, `ip` = *<value>*.

            Constraint: `ip` >= 0.

        (`errno` 6)
            On entry, the AR parameters are outside the stationarity
            region.

        (`errno` 7)
            On entry, `iq` = *<value>*.

            Constraint: `iq` >= 0.

        (`errno` 8)
            On entry, the moving average parameter matrices are such
            that the model is non-invertible.

        (`errno` 9)
            On entry, the covariance matrix `var` is not positive
            semidefinite to machine precision.

        (`errno` 11)
            `k` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `k` = *<value>* and `k` = *<value>*.

        (`errno` 13)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 20)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues of the matrices
            used to test for stationarity or invertibility.

        (`errno` 21)
            The reference vector cannot be computed because the AR
            parameters are too close to the boundary of the stationarity
            region.

        (`errno` 22)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues of the covariance
            matrix.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 23)
            An excessive number of iterations were required by the NAG
            function used to evaluate the eigenvalues stored in the
            reference vector.

    Notes
    -----
    Let the vector X_t = (x_{1t},x_{2t},...,x_{kt})^T, denote a
    k-dimensional time series which is assumed to follow a vector
    autoregressive moving average (VARMA) model of the form:

        [table omitted]

    where epsilon_t = (epsilon_{1t},epsilon_{2t},...,epsilon_{kt})^T, is
    a vector of k residual series assumed to be Normally distributed
    with zero mean and covariance matrix Sigma.
    The components of epsilon_t are assumed to be uncorrelated at
    non-simultaneous lags.
    The phi_i's and theta_j's are k*k matrices of parameters. {phi_i},
    for i = 1,2,...,p, are called the autoregressive (AR) parameter
    matrices, and {theta_j}, for j = 1,2,...,q, the moving average (MA)
    parameter matrices.
    The parameters in the model are thus the p k*k phi-matrices, the q
    k*k theta-matrices, the mean vector mu and the residual error
    covariance matrix Sigma.
    Let

        A(phi) = [[table omitted]]_{pk*pk} and B(theta) = [[table
        omitted]]_{qk*qk}

    where I denotes the k*k identity matrix.

    The model (1) must be both stationary and invertible.
    The model is said to be stationary if the eigenvalues of A(phi) lie
    inside the unit circle and invertible if the eigenvalues of B(theta)
    lie inside the unit circle.

    For k >= 6 the VARMA model (1) is recast into state space form and a
    realization of the state vector at time zero computed.
    For all other cases the function computes a realization of the
    pre-observed vectors X_0,X_-1,...,X_{1-p},
    epsilon_0,epsilon_-1,...,epsilon_{1-q}, from (1), see Shea (1988).
    This realization is then used to generate a sequence of successive
    time series observations.
    Note that special action is taken for pure MA models, that is for p
    = 0.

    At your request a new realization of the time series may be
    generated more efficiently using the information in a reference
    vector created during a previous call to ``times_mv_varma``.
    See the description of the argument `mode` in [Parameters] for
    details.

    The function returns a realization of X_1,X_2,...,X_n.
    On a successful exit, the recent history is updated and saved in the
    array `comm`['r'] so that ``times_mv_varma`` may be called again to
    generate a realization of X_{n+1},X_{n+2},..., etc.
    See the description of the argument `mode` in [Parameters] for
    details.

    Further computational details are given in Shea (1988).
    Note, however, that ``times_mv_varma`` uses a spectral decomposition
    rather than a Cholesky factorization to generate the multivariate
    Normals.
    Although this method involves more multiplications than the Cholesky
    factorization method and is thus slightly slower it is more stable
    when faced with ill-conditioned covariance matrices. A method of
    assigning the AR and MA coefficient matrices so that the
    stationarity and invertibility conditions are satisfied is described
    in Barone (1987).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``times_mv_varma``.

    References
    ----------
    Barone, P, 1987, `A method for generating independent realisations
    of a multivariate normal stationary and invertible ARMA(p,q)
    process`, J. Time Ser. Anal. (8), 125--130

    Shea, B L, 1988, `A note on the generation of independent
    realisations of a vector autoregressive moving average process`, J.
    Time Ser. Anal. (9), 403--410
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_mv_varma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, xmean[k].
        utils._BYREF_INT, # Input, ip.
        utils._EngineFloat64ArrayType.c_type, # Input, phi[k*k*ip].
        utils._BYREF_INT, # Input, iq.
        utils._EngineFloat64ArrayType.c_type, # Input, theta[k*k*iq].
        utils._EngineFloat64ArrayType.c_type, # Input, var[ldvar, k].
        utils._BYREF_INT, # Input, ldvar.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, n].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((var, 'var'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _xmean = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_k_py,),
    ).to_ctypes()
    _ip = utils._EngineIntScalarType(
        ip, {'fun_name': fun_name, 'entity': 'ip'},
    ).to_ctypes()
    _ip_py = utils._EngineIntScalarType.to_py(_ip)
    _phi = utils._EngineFloat64ArrayType(
        phi, {'fun_name': fun_name, 'entity': 'phi'},
        exp_shape=(_k_py*_k_py*_ip_py,),
    ).to_ctypes()
    _iq = utils._EngineIntScalarType(
        iq, {'fun_name': fun_name, 'entity': 'iq'},
    ).to_ctypes()
    _iq_py = utils._EngineIntScalarType.to_py(_iq)
    _theta = utils._EngineFloat64ArrayType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
        exp_shape=(_k_py*_k_py*_iq_py,),
    ).to_ctypes()
    _ldvar_py = utils._EngineFloat64ArrayType.get_nd(var, {'fun_name': fun_name, 'entity': 'var'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldvar = utils._EngineIntScalarType(
        _ldvar_py, {'fun_name': fun_name, 'entity': '_ldvar'},
    ).to_ctypes()
    _var = utils._EngineFloat64ArrayType(
        var, {'fun_name': fun_name, 'entity': 'var'},
        exp_shape=(_k_py, _k_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py in (1, 3):
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _ip_py == 0:
        _vl_l_py = _k_py*(_k_py+1)//2
    elif _ip_py >= 1:
        _vl_l_py = _k_py*(_k_py+1)//2+(_ip_py-1)*_k_py**2
    else:
        _vl_l_py = 0
    if _comm_needs_init:
        if _mode_py in (0, 2):
            if _k_py >= 6:
                _r_dim1 = (5*max(_ip_py, _iq_py)**2+1)*_k_py**2+(4*max(_ip_py, _iq_py)+3)*_k_py+4
            else:
                _r_dim1 = ((_ip_py+_iq_py)**2+1)*_k_py**2+(4*(_ip_py+_iq_py)+3)*_k_py+max(_k_py*max(_ip_py, _iq_py)*(_k_py*max(_ip_py, _iq_py)+2), _k_py**2*(_ip_py+_iq_py)**2+_vl_l_py*(_vl_l_py+3)+_k_py**2*(_iq_py+1))+4
        elif _mode_py in (1, 3):
            _r_dim1 = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
        else:
            _r_dim1 = 0
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_k_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _k,
        _xmean,
        _ip,
        _phi,
        _iq,
        _theta,
        _var,
        _ldvar,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ip >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the AR parameters are outside the stationarity region.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iq = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: iq >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the moving average parameter matrices are such that the model is',
                    'non-invertible.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix var is not positive semidefinite',
                    'to machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'k is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of k = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 16 and _ifmt == 16:
            _ierr_fl = 13
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 20
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An excessive number of iterations were required by the NAG function used',
                    'to evaluate the eigenvalues of the matrices used to test for',
                    'stationarity or invertibility.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 36 and _ifmt == 36:
            _ierr_fl = 21
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The reference vector cannot be computed because the AR parameters are too',
                    'close to the boundary of the stationarity region.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 29 and _ifmt == 29:
            _ierr_fl = 22
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An excessive number of iterations were required by the NAG function used',
                    'to evaluate the eigenvalues of the covariance matrix.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 211 and _ifmt == 211:
            _ierr_fl = 23
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'An excessive number of iterations were required by the NAG function used',
                    'to evaluate the eigenvalues stored in the reference vector.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        return_data = None
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = return_data
    _err_class.manage()
    return return_data

def times_smooth_exp(mode, n, itype, p, param, init, var, comm, statecomm, e, en, x):
    u"""
    Generates a realization of a time series from an exponential
    smoothing model.

    ``times_smooth_exp`` simulates from an exponential smoothing model,
    where the model uses either single exponential, double exponential
    or a Holt--Winters method.

    For full information please refer to the NAG Library document for
    g05pm

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pmf.html

    Parameters
    ----------
    mode : int
        Indicates if ``times_smooth_exp`` is continuing from a previous
        call or, if not, how the initial values are computed.

        `mode` = 0

            Values for m_0, r_0 and s_{-j}, for j = 0,1,...,p-1, are
            supplied in `init`.

        `mode` = 1

            ``times_smooth_exp`` continues from a previous call using
            values that are supplied in `comm`['r']. `comm`['r'] is not
            updated.

        `mode` = 2

            ``times_smooth_exp`` continues from a previous call using
            values that are supplied in `comm`['r']. `comm`['r'] is
            updated.

    n : int
        The number of terms of the time series being generated.

    itype : int
        The smoothing function.

        `itype` = 1

            Single exponential.

        `itype` = 2

            Brown's double exponential.

        `itype` = 3

            Linear Holt.

        `itype` = 4

            Additive Holt--Winters.

        `itype` = 5

            Multiplicative Holt--Winters.

    p : int
        If `itype` = 4 or 5, the seasonal order, p, otherwise `p` is not
        referenced.

    param : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype` in (1, 2): 1; if `itype`=3: 3; if `itype` in
        (4, 5): 4; otherwise: 0.

        The smoothing parameters.

        If `itype` = 1 or 2, `param`[0] = alpha and any remaining
        elements of `param` are not referenced.

        If `itype` = 3, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = phi and any remaining elements of `param` are not
        referenced.

        If `itype` = 4 or 5, `param`[0] = alpha, `param`[1] = gamma,
        `param`[2] = beta and `param`[3] = phi and any remaining
        elements of `param` are not referenced.

    init : float, shape (:)
        Note: the required length for this argument is determined as
        follows: if `itype`=1: 1; if `itype` in (2, 3): 2; if `itype` in
        (4, 5): 2+`p`; otherwise: 0.

        If `mode` = 0, the initial values for m_0, r_0 and s_{-j}, for j
        = 0,1,...,p-1, used to initialize the smoothing.

        If `itype` = 1, `init`[0] = m_0 and any remaining elements of
        `init` are not referenced.

        If `itype` = 2 or 3, `init`[0] = m_0 and `init`[1] = r_0 and any
        remaining elements of `init` are not referenced.

        If `itype` = 4 or 5, `init`[0] = m_0, `init`[1] = r_0 and
        `init`[2] to `init`[2+p-1] hold the values for s_{-j}, for j =
        0,1,...,p-1.

        Any remaining elements of `init` are not referenced.

    var : float
        The variance, sigma^2 of the Normal distribution used to
        generate the errors epsilon_i. If `var` <= 0.0 then Normally
        distributed errors are not used.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1 or 2, this argument must have been initialized by
        a prior call to ``times_smooth_exp``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    e : float, shape (`en`)
        If `en` > 0 and `var` <= 0.0, a vector from which the errors,
        epsilon_t are randomly drawn, with replacement.

        If `en` <= 0, `e` is not referenced.

    en : int
        If `en` > 0, the length of the vector `e`.

        If both `var` <= 0.0 and `en` <= 0 then epsilon_t = 0.0, for t =
        1,2,...,n.

    x : float, shape (`n`), modified in place
        `On exit`: the generated time series, x_t, for t = 1,2,...,n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1, 2, 3, 4 or 5.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: if `itype` = 4 or 5, `p` >= 2.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: 0 <= `param`[i] <= 1.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: if `itype` = 2, 0 < `param`[i] <= 1.

        (`errno` 5)
            On entry, `param`[*<value>*] = *<value>*.

            Constraint: `param`[i] >= 0.

        (`errno` 8)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 12)
            Model unsuitable for multiplicative Holt--Winter, try a
            different set of parameters.

    Notes
    -----
    ``times_smooth_exp`` returns {x_t:t = 1,2,...,n}, a realization of a
    time series from an exponential smoothing model defined by one of
    five smoothing functions:

        Single Exponential Smoothing

                [table omitted]

        Brown Double Exponential Smoothing

                [table omitted]

        Linear Holt Exponential Smoothing

                [table omitted]

        Additive Holt--Winters Smoothing

                [table omitted]

        Multiplicative Holt--Winters Smoothing

                [table omitted]

    where m_t is the mean, r_t is the trend and s_t is the seasonal
    component at time t with p being the seasonal order.
    The errors, epsilon_t are either drawn from a normal distribution
    with mean zero and variance sigma^2 or randomly sampled, with
    replacement, from a user-supplied vector.

    References
    ----------
    Chatfield, C, 1980, `The Analysis of Time Series`, Chapman and Hall
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.times_smooth_exp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, itype.
        utils._BYREF_INT, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input, param[lparam].
        utils._EngineFloat64ArrayType.c_type, # Input, init[linit].
        utils._BYREF_FLOAT64, # Input, var.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Input, e[en].
        utils._BYREF_INT, # Input, en.
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lparam.
        utils._BYREF_INT, # Input, linit.
        utils._BYREF_INT, # Input, lr.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _p = utils._EngineIntScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _lparam_py = utils._EngineFloat64ArrayType.get_shape(param, {'fun_name': fun_name, 'entity': 'param'}, exp_rank=1)[0]
    _lparam = utils._EngineIntScalarType(
        _lparam_py, {'fun_name': fun_name, 'entity': '_lparam'},
    ).to_ctypes()
    _itype_py = utils._EngineIntScalarType.to_py(_itype)
    if _itype_py in (1, 2):
        _param_dim1 = 1
    elif _itype_py == 3:
        _param_dim1 = 3
    elif _itype_py in (4, 5):
        _param_dim1 = 4
    else:
        _param_dim1 = 0
    exp_shape_param = (_param_dim1,)
    _param = utils._EngineFloat64ArrayType(
        param, {'fun_name': fun_name, 'entity': 'param'},
        exp_shape=exp_shape_param,
    ).to_ctypes()
    _linit_py = utils._EngineFloat64ArrayType.get_shape(init, {'fun_name': fun_name, 'entity': 'init'}, exp_rank=1)[0]
    _linit = utils._EngineIntScalarType(
        _linit_py, {'fun_name': fun_name, 'entity': '_linit'},
    ).to_ctypes()
    _p_py = utils._EngineIntScalarType.to_py(_p)
    if _itype_py == 1:
        _init_dim1 = 1
    elif _itype_py in (2, 3):
        _init_dim1 = 2
    elif _itype_py in (4, 5):
        _init_dim1 = 2+_p_py
    else:
        _init_dim1 = 0
    exp_shape_init = (_init_dim1,)
    _init = utils._EngineFloat64ArrayType(
        init, {'fun_name': fun_name, 'entity': 'init'},
        exp_shape=exp_shape_init,
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py in (1, 2):
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        if _itype_py in (1, 2, 3):
            _r_dim1 = 13
        elif _itype_py in (4, 5):
            _r_dim1 = 13+_p_py
        else:
            _r_dim1 = 0
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _en = utils._EngineIntScalarType(
        en, {'fun_name': fun_name, 'entity': 'en'},
    ).to_ctypes()
    _en_py = utils._EngineIntScalarType.to_py(_en)
    _e = utils._EngineFloat64ArrayType(
        e, {'fun_name': fun_name, 'entity': 'e'},
        exp_shape=(_en_py,),
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _itype,
        _p,
        _param,
        _init,
        _var,
        _r,
        _state,
        _e,
        _en,
        _x,
        _lparam,
        _linit,
        _lr,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = 1, 2, 3, 4 or 5.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if itype = 4 or 5, p >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= param[i] <= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 52 and _ifmt == 52:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if itype = 2, 0 < param[i] <= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 53 and _ifmt == 53:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, param[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: param[i] >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'Model unsuitable for multiplicative Holt--Winter, try a different set of',
                    'parameters.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def kfold_xyw(k, fold, n, m, sordx, x, y, w, statecomm):
    u"""
    Permutes a matrix, vector, vector triplet into a form suitable for
    K-fold cross validation.

    ``kfold_xyw`` generates training and validation datasets suitable
    for use in cross-validation or jack-knifing.

    For full information please refer to the NAG Library document for
    g05pv

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pvf.html

    Parameters
    ----------
    k : int
        K, the number of folds.

    fold : int
        The number of the fold to return as the validation dataset.

        On the first call to ``kfold_xyw`` `fold` should be set to 1 and
        then incremented by one at each subsequent call until all K sets
        of training and validation datasets have been produced.

        See [Further Comments] for more details on how a different
        calling sequence can be used.

    n : int
        n, the number of observations.

    m : int
        m, the number of variables.

    sordx : int
        Determines how variables are stored in `x`.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sordx`=2: `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sordx`=1: `m`; if `sordx`=2: `n`;
        otherwise: 0.

        The way the data is stored in `x` is defined by `sordx`.

        If `sordx` = 1, `x`[i-1,j-1] contains the ith observation for
        the jth variable, for i = 1,2,...,`n` and j = 1,2,...,`m`.

        If `sordx` = 2, `x`[j-1,i-1] contains the ith observation for
        the jth variable, for i = 1,2,...,`n` and j = 1,2,...,`m`.

        `On entry`: if `fold` = 1, `x` must hold X_o, the values of X
        for the original dataset, otherwise, `x` must not be changed
        since the last call to ``kfold_xyw``.

        `On exit`: values of X for the training and validation datasets,
        with X_t held in observations 1 to `nt` and X_v in observations
        `nt`+1 to `n`.

    y : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `y` is not None: `n`; otherwise: 0.

        If the original dataset does not include y_o then `y` must be
        set to None.

        `Optionally, on entry`: y_o, the values of y for the original
        dataset. If `fold` != 1, `y` must hold the vector returned in sy
        by the last call to ``kfold_xyw``.

        `On exit`, if not None on entry: values of y for the training
        and validation datasets, with y_t held in elements 1 to `nt` and
        y_v in elements `nt`+1 to `n`.

    w : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `w` is not None: `n`; otherwise: 0.

        `Optionally, on entry`: if `fold` != 1, `w` must hold the vector
        returned in sw by the last call to ``kfold_xyw``.

        `On exit`, if not None on entry: values of w for the training
        and validation datasets, with w_t held in elements 1 to `nt` and
        w_v in elements `nt`+1 to `n`.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    nt : int
        n_t, the number of observations in the training dataset.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `k` = *<value>* and `n` = *<value>*.

            Constraint: 2 <= `k` <= `n`.

        (`errno` 21)
            On entry, `fold` = *<value>* and `k` = *<value>*.

            Constraint: 1 <= `fold` <= `k`.

        (`errno` 31)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 41)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 51)
            On entry, `sordx` = *<value>*.

            Constraint: `sordx` = 1 or 2.

        (`errno` 71)
            On entry, ldx = *<value>* and `n` = *<value>*.

            Constraint: if `sordx` = 1, ldx >= `n`.

        (`errno` 72)
            On entry, ldx = *<value>* and `m` = *<value>*.

            Constraint: if `sordx` = 2, ldx >= `m`.

        (`errno` 131)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Warns
    -----
    NagAlgorithmicWarning
        (`errno` 61)
            More than 50% of the data did not move when the data was
            shuffled. *<value>* of the *<value>* observations stayed
            put.

    Notes
    -----
    Let X_o denote a matrix of n observations on m variables and y_o and
    w_o each denote a vector of length n.
    For example, X_o might represent a matrix of independent variables,
    y_o the dependent variable and w_o the associated weights in a
    weighted regression.

    ``kfold_xyw`` generates a series of training datasets, denoted by
    the matrix, vector, vector triplet (X_t,y_t,w_t) of n_t
    observations, and validation datasets, denoted (X_v,y_v,w_v) with
    n_v observations.
    These training and validation datasets are generated as follows.

    Each of the original n observations is randomly assigned to one of K
    equally sized groups or folds.
    For the kth sample the validation dataset consists of those
    observations in group k and the training dataset consists of all
    those observations not in group k.
    Therefore, at most K samples can be generated.

    If n is not divisible by K then the observations are assigned to
    groups as evenly as possible, therefore, any group will be at most
    one observation larger or smaller than any other group.

    When using K = n the resulting datasets are suitable for
    leave-one-out cross-validation, or the training dataset on its own
    for jack-knifing.
    When using K != n the resulting datasets are suitable for K-fold
    cross-validation.
    Datasets suitable for reversed cross-validation can be obtained by
    switching the training and validation datasets, i.e., use the kth
    group as the training dataset and the rest of the data as the
    validation dataset.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``kfold_xyw``.

    See Also
    --------
    :meth:`naginterfaces.library.examples.correg.glm_binomial_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.kfold_xyw'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pvft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, k.
        utils._BYREF_INT, # Input, fold.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, sordx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, ly.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[ly].
        utils._BYREF_INT, # Input, lw.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, w[lw].
        utils._BYREF_INT, # Output, nt.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_INT, # Input, matlab.
        utils._BYREF_INT, # Input, sordsx.
        utils._EngineFloat64ArrayType.c_type, # Output, sx[ldsx, 0].
        utils._BYREF_INT, # Input, ldsx.
        utils._BYREF_INT, # Input, lsy.
        utils._EngineFloat64ArrayType.c_type, # Output, sy[lsy].
        utils._BYREF_INT, # Input, lsw.
        utils._EngineFloat64ArrayType.c_type, # Output, sw[lsw].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _fold = utils._EngineIntScalarType(
        fold, {'fun_name': fun_name, 'entity': 'fold'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _sordx = utils._EngineIntScalarType(
        sordx, {'fun_name': fun_name, 'entity': 'sordx'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sordx_py = utils._EngineIntScalarType.to_py(_sordx)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _sordx_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = _n_py
    if _sordx_py == 1:
        _x_dim2 = _m_py
    elif _sordx_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ly_py = utils._EngineFloat64ArrayType.get_shape(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1, allow_none=True)[0]
    _ly = utils._EngineIntScalarType(
        _ly_py, {'fun_name': fun_name, 'entity': '_ly'},
    ).to_ctypes()
    if y is not None:
        _y_dim1 = _n_py
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        allow_none=True,
    )
    _y = _y_cls.to_ctypes()
    _lw_py = utils._EngineFloat64ArrayType.get_shape(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=1, allow_none=True)[0]
    _lw = utils._EngineIntScalarType(
        _lw_py, {'fun_name': fun_name, 'entity': '_lw'},
    ).to_ctypes()
    if w is not None:
        _w_dim1 = _n_py
    else:
        _w_dim1 = 0
    exp_shape_w = (_w_dim1,)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=exp_shape_w,
        allow_none=True,
    )
    _w = _w_cls.to_ctypes()
    _nt = utils._EngineIntScalarType.empty_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _matlab = utils._EngineIntScalarType.c_type(0)
    _sordsx = utils._EngineIntScalarType.c_type(1)
    _ldsx = utils._EngineIntScalarType.c_type(0)
    _sx = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lsy = utils._EngineIntScalarType.c_type(0)
    _sy = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lsw = utils._EngineIntScalarType.c_type(0)
    _sw = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _k,
        _fold,
        _n,
        _m,
        _sordx,
        _x,
        _ldx,
        _ly,
        _y,
        _lw,
        _w,
        _nt,
        _state,
        _lstate,
        _matlab,
        _sordsx,
        _sx,
        _ldsx,
        _lsy,
        _sy,
        _lsw,
        _sw,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 2 <= k <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fold = {:s} and k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= fold <= k.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 51 and _ifmt == 51:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sordx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sordx = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineWarningExit(
                fun_name=fun_name,
                msg_lines=[
                    'More than 50% of the data did not move when the data was shuffled.',
                    '{:s} of the {:s} observations stayed put.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldx = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if sordx = 1, ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 72 and _ifmt == 72:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldx = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if sordx = 2, ldx >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 131 and _ifmt == 131:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 141 and _ifmt == 141:
            _ierr_fl = 131
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, (utils._EngineNoErrorExit, utils._EngineWarningExit)):
        _x_cls.refresh_py(_x)
        if y is not None:
            _y_cls.refresh_py(_y)
        if w is not None:
            _w_cls.refresh_py(_w)
        _nt_py = utils._EngineIntScalarType.to_py(_nt)
        return_data = _nt_py
        if isinstance(_err_class, utils._EngineWarningExit):
            _err_class.return_data = _collections.namedtuple(
                'nag_return_data',
                ['nt'],
            )(
                _nt_py,
            )
    _err_class.manage()
    return return_data

def subsamp_xyw(nt, n, m, sordx, x, y, w, statecomm):
    u"""
    Permutes a matrix, vector, vector triplet into a form suitable for
    random sub-sampling validation.

    ``subsamp_xyw`` generates a dataset suitable for use with repeated
    random sub-sampling validation.

    For full information please refer to the NAG Library document for
    g05pw

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pwf.html

    Parameters
    ----------
    nt : int
        n_t, the number of observations in the training dataset.

    n : int
        n, the number of observations.

    m : int
        m, the number of variables.

    sordx : int
        Determines how variables are stored in `x`.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sordx`=2: `m`; otherwise: `n`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sordx`=1: `m`; if `sordx`=2: `n`;
        otherwise: 0.

        The way the data is stored in `x` is defined by `sordx`.

        If `sordx` = 1, `x`[i-1,j-1] contains the ith observation for
        the jth variable, for i = 1,2,...,`n` and j = 1,2,...,`m`.

        If `sordx` = 2, `x`[j-1,i-1] contains the ith observation for
        the jth variable, for i = 1,2,...,`n` and j = 1,2,...,`m`.

        `On entry`: `x` must hold X_o, the values of X for the original
        dataset. This may be the same `x` as updated by a previous call
        to ``subsamp_xyw``.

        `On exit`: values of X for the training and validation datasets,
        with X_t held in observations 1 to `nt` and X_v in observations
        `nt`+1 to `n`.

    y : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `y` is not None: `n`; otherwise: 0.

        `Optionally, on entry`: `y` must hold y_o, the values of y for
        the original dataset. This may be the same `y` as updated by a
        previous call to ``subsamp_xyw``.

        `On exit`, if not None on entry: values of y for the training
        and validation datasets, with y_t held in elements 1 to `nt` and
        y_v in elements `nt`+1 to `n`.

    w : None or float, shape (:), modified in place
        Note: the required length for this argument is determined as
        follows: if `w` is not None: `n`; otherwise: 0.

        `Optionally, on entry`: `w` must hold w_o, the values of w for
        the original dataset. This may be the same `w` as updated by a
        previous call to ``subsamp_xyw``.

        `On exit`, if not None on entry: values of w for the training
        and validation datasets, with w_t held in elements 1 to `nt` and
        w_v in elements `nt`+1 to `n`.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `nt` = *<value>* and `n` = *<value>*.

            Constraint: 1 <= `nt` <= `n`.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 31)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 41)
            On entry, `sordx` = *<value>*.

            Constraint: `sordx` = 1 or 2.

        (`errno` 61)
            On entry, ldx = *<value>* and `n` = *<value>*.

            Constraint: if `sordx` = 1, ldx >= `n`.

        (`errno` 62)
            On entry, ldx = *<value>* and `m` = *<value>*.

            Constraint: if `sordx` = 2, ldx >= `m`.

        (`errno` 111)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    Let X_o denote a matrix of n observations on m variables and y_o and
    w_o each denote a vector of length n.
    For example, X_o might represent a matrix of independent variables,
    y_o the dependent variable and w_o the associated weights in a
    weighted regression.

    ``subsamp_xyw`` generates a series of training datasets, denoted by
    the matrix, vector, vector triplet (X_t,y_t,w_t) of n_t
    observations, and validation datasets, denoted (X_v,y_v,w_v) with
    n_v observations.
    These training and validation datasets are generated by randomly
    assigning each observation to either the training dataset or the
    validation dataset.

    The resulting datasets are suitable for use with repeated random
    sub-sampling validation.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``subsamp_xyw``.
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.subsamp_xyw'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pwft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, nt.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, sordx.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, x[ldx, *].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, ly.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, y[ly].
        utils._BYREF_INT, # Input, lw.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, w[lw].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_INT, # Input, matlab.
        utils._BYREF_INT, # Input, sordsx.
        utils._EngineFloat64ArrayType.c_type, # Output, sx[ldsx, 0].
        utils._BYREF_INT, # Input, ldsx.
        utils._BYREF_INT, # Input, lsy.
        utils._EngineFloat64ArrayType.c_type, # Output, sy[lsy].
        utils._BYREF_INT, # Input, lsw.
        utils._EngineFloat64ArrayType.c_type, # Output, sw[lsw].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _nt = utils._EngineIntScalarType(
        nt, {'fun_name': fun_name, 'entity': 'nt'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _sordx = utils._EngineIntScalarType(
        sordx, {'fun_name': fun_name, 'entity': 'sordx'},
    ).to_ctypes()
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sordx_py = utils._EngineIntScalarType.to_py(_sordx)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _sordx_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = _n_py
    if _sordx_py == 1:
        _x_dim2 = _m_py
    elif _sordx_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
    )
    _x = _x_cls.to_ctypes()
    _ly_py = utils._EngineFloat64ArrayType.get_shape(y, {'fun_name': fun_name, 'entity': 'y'}, exp_rank=1, allow_none=True)[0]
    _ly = utils._EngineIntScalarType(
        _ly_py, {'fun_name': fun_name, 'entity': '_ly'},
    ).to_ctypes()
    if y is not None:
        _y_dim1 = _n_py
    else:
        _y_dim1 = 0
    exp_shape_y = (_y_dim1,)
    _y_cls = utils._EngineFloat64ArrayType(
        y, {'fun_name': fun_name, 'entity': 'y'},
        exp_shape=exp_shape_y,
        allow_none=True,
    )
    _y = _y_cls.to_ctypes()
    _lw_py = utils._EngineFloat64ArrayType.get_shape(w, {'fun_name': fun_name, 'entity': 'w'}, exp_rank=1, allow_none=True)[0]
    _lw = utils._EngineIntScalarType(
        _lw_py, {'fun_name': fun_name, 'entity': '_lw'},
    ).to_ctypes()
    if w is not None:
        _w_dim1 = _n_py
    else:
        _w_dim1 = 0
    exp_shape_w = (_w_dim1,)
    _w_cls = utils._EngineFloat64ArrayType(
        w, {'fun_name': fun_name, 'entity': 'w'},
        exp_shape=exp_shape_w,
        allow_none=True,
    )
    _w = _w_cls.to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _matlab = utils._EngineIntScalarType.c_type(0)
    _sordsx = utils._EngineIntScalarType.c_type(1)
    _ldsx = utils._EngineIntScalarType.c_type(0)
    _sx = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lsy = utils._EngineIntScalarType.c_type(0)
    _sy = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _lsw = utils._EngineIntScalarType.c_type(0)
    _sw = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _nt,
        _n,
        _m,
        _sordx,
        _x,
        _ldx,
        _ly,
        _y,
        _lw,
        _w,
        _state,
        _lstate,
        _matlab,
        _sordsx,
        _sx,
        _ldsx,
        _lsy,
        _sy,
        _lsw,
        _sw,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nt = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= nt <= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sordx = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sordx = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldx = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if sordx = 1, ldx >= n.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldx = {:s} and m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if sordx = 2, ldx >= m.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 121 and _ifmt == 121:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        if y is not None:
            _y_cls.refresh_py(_y)
        if w is not None:
            _w_cls.refresh_py(_w)
        return_data = None
    _err_class.manage()
    return return_data

def matrix_orthog(side, init, m, n, statecomm, a):
    u"""
    Generates a random orthogonal matrix.

    ``matrix_orthog`` generates a random orthogonal matrix.

    For full information please refer to the NAG Library document for
    g05px

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pxf.html

    Parameters
    ----------
    side : str, length 1
        Indicates whether the matrix A is multiplied on the left or
        right by the random orthogonal matrix U.

        `side` = 'L'

            The matrix A is multiplied on the left, i.e., premultiplied.

        `side` = 'R'

            The matrix A is multiplied on the right, i.e.,
            post-multiplied.

    init : str, length 1
        Indicates whether or not `a` should be initialized to the
        identity matrix.

        `init` = 'I'

            `a` is initialized to the identity matrix.

        `init` = 'N'

            `a` is not initialized and the matrix A must be supplied in
            `a`.

    m : int
        m, the number of rows of the matrix A.

    n : int
        n, the number of columns of the matrix A.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    a : float, shape (`m`, `n`), modified in place
        `On entry`: if `init` = 'N', `a` must contain the matrix A.

        `On exit`: the matrix UA when `side` = 'L' or the matrix AU when
        `side` = 'R'.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `side` is not valid: `side` = *<value>*.

        (`errno` 2)
            On entry, `init` is not valid: `init` = *<value>*.

        (`errno` 3)
            On entry, `side` = *<value>*, `m` = *<value>*.

            Constraint: if `side` = 'L', `m` > 1; otherwise `m` >= 1.

        (`errno` 4)
            On entry, `side` = *<value>*, `n` = *<value>*.

            Constraint: if `side` = 'R', `n` > 1; otherwise `n` >= 1.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 8)
            On entry, `side` = *<value>*, `m` = *<value>*.

            Constraint: if `side` = 'L', `m` > 1; otherwise `m` >= 1.

        (`errno` 8)
            On entry, `side` = *<value>*, `n` = *<value>*.

            Constraint: if `side` = 'R', `n` > 1; otherwise `n` >= 1.

    Notes
    -----
    ``matrix_orthog`` pre - or post-multiplies an m*n matrix A by a
    random orthogonal matrix U, overwriting A.
    The matrix A may optionally be initialized to the identity matrix
    before multiplying by U, hence U is returned. U is generated using
    the method of Stewart (1980).
    The algorithm can be summarised as follows.

    Let x_1,x_2,...,x_{n-1} follow independent multinormal distributions
    with zero mean and variance I sigma^2 and dimensions n,n-1,...,2;
    let H_j = diag(I_{j-1},H_j^*), where I_{j-1} is the identity matrix
    and H_j^* is the Householder transformation that reduces x_j to
    r_{jj}e_1, e_1 being the vector with first element one and the
    remaining elements zero and r_{jj} being a scalar, and let D =
    diag(sign(r_11),sign(r_22),...,sign(r_{nn})).
    Then the product U = DH_1 H_2...H_{n-1} is a random orthogonal
    matrix distributed according to the Haar measure over the set of
    orthogonal matrices of n.
    See Theorem 3.3 in Stewart (1980).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``matrix_orthog``.

    References
    ----------
    Stewart, G W, 1980, `The efficient generation of random orthogonal
    matrices with an application to condition estimates`, SIAM J. Numer.
    Anal. (17), 403--409
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.matrix_orthog'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pxft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_STRING, # Input, side(len=1).
        utils._BYREF_STRING, # Input, init(len=1).
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, a[lda, n].
        utils._BYREF_INT, # Input, lda.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of side.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of init.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((a, 'a'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _side = utils._EngineCharScalarType(
        side, {'fun_name': fun_name, 'entity': 'side'}, exp_chrlen=1,
    ).to_ctypes()
    _init = utils._EngineCharScalarType(
        init, {'fun_name': fun_name, 'entity': 'init'}, exp_chrlen=1,
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _lda_py = utils._EngineFloat64ArrayType.get_nd(a, {'fun_name': fun_name, 'entity': 'a'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _lda = utils._EngineIntScalarType(
        _lda_py, {'fun_name': fun_name, 'entity': '_lda'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _a_cls = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py, _n_py),
        mutual_sorder=_m_sorder,
    )
    _a = _a_cls.to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _side,
        _init,
        _m,
        _n,
        _state,
        _a,
        _lda,
        _lstate,
        _errbuf,
        _ierr,
        _side._length_,
        _init._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, side is not valid: side = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, init is not valid: init = \'{:s}\'.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, side = \'{:s}\', m = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if side = \'L\', m > 1;',
                    'otherwise m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, side = \'{:s}\', n = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if side = \'R\', n > 1;',
                    'otherwise n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, side = \'{:s}\', m = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if side = \'L\', m > 1;',
                    'otherwise m >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, side = \'{:s}\', n = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if side = \'R\', n > 1;',
                    'otherwise n >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _a_cls.refresh_py(_a)
        return_data = None
    _err_class.manage()
    return return_data

def matrix_corr(n, d, eps, statecomm, c):
    u"""
    Generates a random correlation matrix.

    ``matrix_corr`` generates a random correlation matrix with given
    eigenvalues.

    For full information please refer to the NAG Library document for
    g05py

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pyf.html

    Parameters
    ----------
    n : int
        n, the dimension of the correlation matrix to be generated.

    d : float, shape (`n`)
        The n eigenvalues, lambda_i, for i = 1,2,...,n.

    eps : float
        The maximum acceptable error in the diagonal elements.

        Suggested value: `eps` = 0.00001.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    c : float, shape (`n`, `n`), modified in place
        `On exit`: a random correlation matrix, C, of dimension n.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 1.

        (`errno` 2)
            On entry, an eigenvalue is negative.

        (`errno` 2)
            On entry, the eigenvalues do not sum to `n`.

        (`errno` 3)
            On entry, `eps` = *<value>*.

            Constraint: `eps` >= `n`*machine precision.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 5)
            The diagonals of the returned matrix are not unity, try
            increasing the value of `eps`, or rerun the code using a
            different seed.

    Notes
    -----
    Given n eigenvalues, lambda_1,lambda_2,...,lambda_n, such that

        sum_i = 1^n lambda_i = n

    and

        lambda_i >= 0, i = 1,2,...,n,

    ``matrix_corr`` will generate a random correlation matrix, C, of
    dimension n, with eigenvalues lambda_1,lambda_2,...,lambda_n.

    The method used is based on that described by Lin and Bendel (1985).
    Let D be the diagonal matrix with values
    lambda_1,lambda_2,...,lambda_n and let A be a random orthogonal
    matrix generated by ``matrix_orthog`` then the matrix C_0 = ADA^T is
    a random covariance matrix with eigenvalues
    lambda_1,lambda_2,...,lambda_n.
    The matrix C_0 is transformed into a correlation matrix by means of
    n-1 elementary rotation matrices P_i such that C =
    P_{n-1}P_{n-2}...P_1 C_0 P_1^T... P_{n-2}^T P_{n-1}^T.
    The restriction on the sum of eigenvalues implies that for any
    diagonal element of C_0 > 1, there is another diagonal element < 1.
    The P_i are constructed from such pairs, chosen at random, to
    produce a unit diagonal element corresponding to the first element.
    This is repeated until all diagonal elements are 1 to within a given
    tolerance epsilon.

    The randomness of C should be interpreted only to the extent that A
    is a random orthogonal matrix and C is computed from A using the P_i
    which are chosen as arbitrarily as possible.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``matrix_corr``.

    References
    ----------
    Lin, S P and Bendel, R B, 1985, `Algorithm AS 213: Generation of
    population correlation on matrices with specified eigenvalues`,
    Appl. Statist. (34), 193--198
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.matrix_corr'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pyft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, d[n].
        utils._BYREF_FLOAT64, # Input, eps.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, c[ldc, n].
        utils._BYREF_INT, # Input, ldc.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _d = utils._EngineFloat64ArrayType(
        d, {'fun_name': fun_name, 'entity': 'd'},
        exp_shape=(_n_py,),
    ).to_ctypes()
    _eps = utils._EngineFloat64ScalarType(
        eps, {'fun_name': fun_name, 'entity': 'eps'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c_cls = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_n_py, _n_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _c = _c_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _d,
        _eps,
        _state,
        _c,
        _ldc,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, an eigenvalue is negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the eigenvalues do not sum to n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, eps = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: eps >= n*machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The diagonals of the returned matrix are not unity, try increasing',
                    'the value of eps, or rerun the code using a different seed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _c_cls.refresh_py(_c)
        return_data = None
    _err_class.manage()
    return return_data

def matrix_2waytable(mode, nrow, ncol, totr, totc, comm, statecomm, x):
    u"""
    Generates a random two-way table.

    ``matrix_2waytable`` generates a random two-way table.

    For full information please refer to the NAG Library document for
    g05pz

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05pzf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate two-way table using reference vector set up in a
            prior call to ``matrix_2waytable``.

        `mode` = 2

            Set up reference vector and generate two-way table.

    nrow : int
        m, the number of rows in the table.

    ncol : int
        n, the number of columns in the table.

    totr : int, shape (`nrow`)
        The m row totals, R_i, for i = 1,2,...,m.

    totc : int, shape (`ncol`)
        The n column totals, C_j, for j = 1,2,...,n.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``matrix_2waytable``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`nrow`, `ncol`), modified in place
        `On exit`, if not None on entry: if `mode` = 1 or 2, a
        pseudorandom two-way m*n table, X, with element `x`[i-1,j-1]
        containing the (i,j)th entry in the table such that
        sum(`x`[i-1,j-1]) = `totc`[j-1] and sum(`x`[i-1,j-1]) =
        `totr`[i-1]

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `nrow` = *<value>*.

            Constraint: `nrow` >= 2.

        (`errno` 3)
            On entry, `ncol` = *<value>*.

            Constraint: `ncol` >= 2.

        (`errno` 4)
            On entry, at least one element of `totr` is negative or
            `totr` sums to zero.

        (`errno` 5)
            On entry, `totc` has at least one negative element.

        (`errno` 6)
            `nrow` or `ncol` is not the same as when `comm`['r'] was set
            up in a previous call.

            Previous value of `nrow` = *<value>* and `nrow` = *<value>*.

            Previous value of `ncol` = *<value>* and `ncol` = *<value>*.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

        (`errno` 15)
            On entry, the arrays `totr` and `totc` do not sum to the
            same total: `totr` array total is *<value>*, `totc` array
            total is *<value>*.

    Notes
    -----
    Given m row totals R_i and n column totals C_j (with sum_i = 1^mR_i
    = sum_j = 1^nC_j = T, say), ``matrix_2waytable`` will generate a
    pseudorandom two-way table of integers such that the row and column
    totals are satisfied.

    The method used is based on that described by Patefield (1981) which
    is most efficient when T is large relative to the number of table
    entries m*n (i.e., T > 2mn).
    Entries are generated one row at a time and one entry at a time
    within a row.
    Each entry is generated using the conditional probability
    distribution for that entry given the entries in the previous rows
    and the previous entries in the same row.

    A reference vector is used to store computed values that can be
    reused in the generation of new tables with the same row and column
    totals. ``matrix_2waytable`` can be called to simply set up the
    reference vector, or to generate a two-way table using a reference
    vector set up in a previous call, or it can combine both functions
    in a single call.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``matrix_2waytable``.

    References
    ----------
    Patefield, W M, 1981, `An efficient method of generating R*C tables
    with given row and column totals`, Appl. Stats. (30), 91--97
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.matrix_2waytable'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05pzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, nrow.
        utils._BYREF_INT, # Input, ncol.
        utils._EngineIntArrayType.c_type, # Input, totr[nrow].
        utils._EngineIntArrayType.c_type, # Input, totc[ncol].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[ldx, ncol].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _nrow = utils._EngineIntScalarType(
        nrow, {'fun_name': fun_name, 'entity': 'nrow'},
    ).to_ctypes()
    _ncol = utils._EngineIntScalarType(
        ncol, {'fun_name': fun_name, 'entity': 'ncol'},
    ).to_ctypes()
    _nrow_py = utils._EngineIntScalarType.to_py(_nrow)
    _totr_cls = utils._EngineIntArrayType(
        totr, {'fun_name': fun_name, 'entity': 'totr'},
        exp_shape=(_nrow_py,),
    )
    _totr = _totr_cls.to_ctypes()
    _ncol_py = utils._EngineIntScalarType.to_py(_ncol)
    _totc = utils._EngineIntArrayType(
        totc, {'fun_name': fun_name, 'entity': 'totc'},
        exp_shape=(_ncol_py,),
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        if _mode_py in (0, 2):
            _r_dim1 = _totr_cls.get_sum(0, _nrow_py-1)+5
        elif _mode_py == 1:
            _r_dim1 = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
        else:
            _r_dim1 = 0
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineIntArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_nrow_py, _ncol_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _nrow,
        _ncol,
        _totr,
        _totc,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nrow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nrow >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ncol >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of totr is negative or totr sums to zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of totr is negative or totr sums to zero.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, totc has at least one negative element.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'nrow or ncol is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of nrow = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'nrow = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of ncol = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'ncol = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _ierr_fl = 15
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the arrays totr and totc do not sum to the same total:',
                    'totr array total is {:s}, totc array total is'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_students_t(mode, n, df, m, c, comm, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a Student's
    t-copula.

    ``copula_students_t`` sets up a reference vector and generates an
    array of pseudorandom numbers from a Student's t copula with nu
    degrees of freedom and covariance matrix nu/(nu-2)C.

    For full information please refer to the NAG Library document for
    g05rc

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rcf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``copula_students_t``.

        `mode` = 2

            Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    df : int
        nu, the number of degrees of freedom of the distribution.

    m : int
        m, the number of dimensions of the distribution.

    c : float, shape (`m`, `m`)
        Matrix which, along with `df`, defines the covariance of the
        distribution. Only the upper triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``copula_students_t``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `mode`=0: 0; otherwise: `m`.

        `On exit`, if not None on entry: the array of values from a
        multivariate Student's t copula, with `x`[i-1,j-1] holding the
        jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 5)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 7)
            `m` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The Student's t copula, G, is defined by

        G(u_1,u_2,...,u_m;C) =
        T_nu,C^m(t_nu,C_11^-1(u_1),t_nu,C_22^-1(u_2),...,t_nu,C_mm^-1(u_m))

    where m is the number of dimensions, T_nu,C^m is the multivariate
    Student's t density function with nu degrees of freedom, mean zero
    and covariance matrix nu/(nu-2)C and t_nu,C_ii^-1 is the inverse of
    the univariate Student's t density function with nu degrees of
    freedom, zero mean and variance nu/(nu-2)C_ii.

    ``multivar_students_t`` is used to generate a vector from a
    multivariate Student's t distribution and ``stat.prob_students_t``
    is used to convert each element of that vector into a uniformly
    distributed value between zero and one.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_students_t``.

    References
    ----------
    Nelsen, R B, 1998, `An Introduction to Copulas. Lecture Notes in
    Statistics 139`, Springer

    Sklar, A, 1973, `Random variables: joint distribution functions and
    copulas`, Kybernetika (9), 499--460

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.copula_students_t_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_students_t'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, df.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_m_py*(_m_py+1)+2)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _m_py
    exp_shape_x = (_n_py, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _uplo = utils._EngineCharScalarType(
        'U', {'fun_name': fun_name, 'entity': '_uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mode,
        _n,
        _df,
        _m,
        _c,
        _ldc,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _uplo,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix C is not positive semidefinite',
                    'to machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'm is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_normal(mode, n, m, c, comm, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a Gaussian copula.

    ``copula_normal`` sets up a reference vector and generates an array
    of pseudorandom numbers from a Normal (Gaussian) copula with
    covariance matrix C.

    For full information please refer to the NAG Library document for
    g05rd

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rdf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``copula_normal``.

        `mode` = 2

            Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    m : int
        m, the number of dimensions of the distribution.

    c : float, shape (`m`, `m`)
        The covariance matrix of the distribution. Only the upper
        triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``copula_normal``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `mode`=0: 0; otherwise: `m`.

        `On exit`, if not None on entry: the array of values from a
        multivariate Gaussian copula, with `x`[i-1,j-1] holding the jth
        dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 4)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 6)
            `m` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The Gaussian copula, G, is defined by

        G(u_1,u_2,...,u_m;C) =
        Phi_C(phi_C_11^-1(u_1),phi_C_22^-1(u_2),...,phi_C_mm^-1(u_m))

    where m is the number of dimensions, Phi_C is the multivariate
    Normal density function with mean zero and covariance matrix C and
    phi_C_ii^-1 is the inverse of the univariate Normal density function
    with mean zero and variance C_ii.

    ``multivar_normal`` is used to generate a vector from a multivariate
    Normal distribution and ``stat.prob_normal`` is used to convert each
    element of that vector into a uniformly distributed value between
    zero and one.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_normal``.

    References
    ----------
    Nelsen, R B, 1998, `An Introduction to Copulas. Lecture Notes in
    Statistics 139`, Springer

    Sklar, A, 1973, `Random variables: joint distribution functions and
    copulas`, Kybernetika (9), 499--460
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_m_py*(_m_py+1)+1)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _m_py
    exp_shape_x = (_n_py, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _uplo = utils._EngineCharScalarType(
        'U', {'fun_name': fun_name, 'entity': '_uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mode,
        _n,
        _m,
        _c,
        _ldc,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _uplo,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix C is not positive semidefinite',
                    'to machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'm is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_clayton_bivar(n, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate
    Clayton/Cook--Johnson copula.

    ``copula_clayton_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Clayton/Cook--Johnson Archimedean
    copula.

    For full information please refer to the NAG Library document for
    g05re

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ref.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: 2;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: 2; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the n bivariate uniforms with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= -1.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates pseudorandom uniform bivariates {u_1,u_2} in (0,1]^2 whose
    joint distribution is the Clayton/Cook--Johnson Archimedean copula
    C_theta with parameter theta, given by

        C_theta = [max(u_1^(-theta)+u_2^(-theta)-1,0)]^(-1/theta), theta
        in (-1, infinity)∖{0}

    with the special cases:

        C_-1 = max(u_1+u_2-1,0), the Fréchet--Hoeffding lower bound;

        C_0 = u_1 u_2, the product copula;

        C_infinity = min(u_1,u_2), the Fréchet--Hoeffding upper bound.

    The generation method uses conditional sampling.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_clayton_bivar``.

    References
    ----------
    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_clayton_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05reft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid theta: theta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta >= -1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_frank_bivar(n, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate Frank
    copula.

    ``copula_frank_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Frank Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rff.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: 2;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: 2; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the n bivariate uniforms with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates pseudorandom uniform bivariates {u_1,u_2} in [0,1]^2 whose
    joint distribution is the Frank Archimedean copula C_theta with
    parameter theta, given by

        C_theta = -1/thetaln[1+(e^(-theta u_1)-1)(e^(-theta
        u_2)-1)/(e^(-theta)-1)], theta in (- infinity, infinity)∖{0}

    with the special cases:

        C_{ - infinity } = max(u_1+u_2-1,0), the Fréchet--Hoeffding
        lower bound;

        C_0 = u_1 u_2, the product copula;

        C_infinity = min(u_1,u_2), the Fréchet--Hoeffding upper bound.

    The generation method uses conditional sampling.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_frank_bivar``.

    References
    ----------
    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_frank_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_plackett_bivar(n, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a bivariate Plackett
    copula.

    ``copula_plackett_bivar`` generates pseudorandom uniform bivariates
    with joint distribution of a Plackett copula.

    For full information please refer to the NAG Library document for
    g05rg

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rgf.html

    Parameters
    ----------
    n : int
        n, the number of bivariates to generate.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,2, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: 2;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: 2; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the n bivariate uniforms with joint distribution
        described by C_theta, with `x`[i-1,j-1] holding the ith value
        for the jth dimension if `sorder` = 1 and the jth value for the
        ith dimension if `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 0.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates pseudorandom uniform bivariates {u_1,u_2} in [0,1]^2 whose
    joint distribution is the Plackett copula C_theta with parameter
    theta, given by

        C_theta =
        ([1+(theta-1)(u_1+u_2)]-sqrt([1+(theta-1)(u_1+u_2)]^2-4u_1
        u_2theta(theta-1)))/(2*(theta-1)), theta in (0, infinity)∖{1}

    with the special cases:

        C_0 = max(u_1+u_2-1,0), the Fréchet--Hoeffding lower bound;

        C_1 = u_1 u_2, the product copula;

        C_infinity = min(u_1,u_2), the Fréchet--Hoeffding upper bound.

    The generation method uses conditional sampling.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_plackett_bivar``.

    References
    ----------
    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_plackett_bivar'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = 2
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = 2
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid theta: theta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_clayton(n, m, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Clayton/Cook--Johnson copula.

    ``copula_clayton`` generates pseudorandom uniform variates with
    joint distribution of a Clayton/Cook--Johnson Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rh

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rhf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: `m`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `m`; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the pseudorandom uniform variates with joint
        distribution described by C_theta, with `x`[i-1,j-1] holding the
        ith value for the jth dimension if `sorder` = 1 and the jth
        value for the ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0*10^-6.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates n pseudorandom uniform m-variates whose joint distribution
    is the Clayton/Cook--Johnson Archimedean copula C_theta, given by

        C_theta = (u_1^-theta+u_2^-theta+...+u_m^-theta-m+1)^(-1/theta),
        {[table omitted]

    with the special case:

        C_infinity = min(u_1,u_2,...,u_m), the Fréchet--Hoeffding upper
        bound.

    The generation method uses mixture of powers.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_clayton``.

    References
    ----------
    Marshall, A W and Olkin, I, 1988, `Families of multivariate
    distributions`, Journal of the American Statistical Association
    (83), 403

    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_clayton'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rhft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid theta: theta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta >= 1.0*10^-6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_frank(n, m, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate Frank
    copula.

    ``copula_frank`` generates pseudorandom uniform variates with joint
    distribution of a Frank Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rjf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: `m`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `m`; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the pseudorandom uniform variates with joint
        distribution described by C_theta, with `x`[i-1,j-1] holding the
        ith value for the jth dimension if `sorder` = 1 and the jth
        value for the ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0*10^-6.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates n pseudorandom uniform m-variates whose joint distribution
    is the Frank Archimedean copula C_theta, given by

        C_theta = -1/thetaln[1+((e^(-theta u_1)-1)(e^(-theta
        u_2)-1)...(e^(-theta u_m)-1))/(e^(-theta)-1)^(m-1)], {[table
        omitted]

    with the special case:

        C_infinity = min(u_1,u_2,...,u_m), the Fréchet--Hoeffding upper
        bound.

    The generation method uses mixture of powers.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_frank``.

    References
    ----------
    Marshall, A W and Olkin, I, 1988, `Families of multivariate
    distributions`, Journal of the American Statistical Association
    (83), 403

    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_frank'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid theta: theta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta >= 1.0*10^-6.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def copula_gumbel(n, m, theta, sorder, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a Gumbel--Hougaard
    copula.

    ``copula_gumbel`` generates pseudorandom uniform variates with joint
    distribution of a Gumbel--Hougaard Archimedean copula.

    For full information please refer to the NAG Library document for
    g05rk

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rkf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom uniform variates to generate.

    m : int
        m, the number of dimensions.

    theta : float
        theta, the copula parameter.

    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2: `m`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `m`; if `sorder`=2: `n`;
        otherwise: 0.

        `On exit`: the pseudorandom uniform variates with joint
        distribution described by C_theta, with `x`[i-1,j-1] holding the
        ith value for the jth dimension if `sorder` = 1 and the jth
        value for the ith dimension of `sorder` = 2.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, corrupt `statecomm`['state'] argument.

        (`errno` 2)
            On entry, invalid `theta`: `theta` = *<value>*.

            Constraint: `theta` >= 1.0.

        (`errno` 3)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 2.

        (`errno` 5)
            On entry, invalid `sorder`.

            Constraint: `sorder` = 1 or 2.

    Notes
    -----
    Generates n pseudorandom uniform m-variates whose joint distribution
    is the Gumbel--Hougaard Archimedean copula C_theta, given by

        C_theta =
        exp{-[(-ln(u_1))^theta+(-ln(u_2))^theta+...+(-ln(u_m))^theta]},
        {[table omitted]

    with the special cases:

        C_1 = u_1 u_2...u_m, the product copula;

        C_infinity = min(u_1,u_2,...,u_m), the Fréchet--Hoeffding upper
        bound.

    The generation method uses mixture of powers.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``copula_gumbel``.

    References
    ----------
    Marshall, A W and Olkin, I, 1988, `Families of multivariate
    distributions`, Journal of the American Statistical Association
    (83), 403

    Nelsen, R B, 2006, `An Introduction to Copulas`, (2nd Edition),
    Springer Series in Statistics
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.copula_gumbel'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, theta.
        utils._BYREF_INT, # Input, sorder.
        utils._EngineIntArrayType.c_type, # Input/Output, state[*].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, sdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, sdx.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _theta = utils._EngineFloat64ScalarType(
        theta, {'fun_name': fun_name, 'entity': 'theta'},
    ).to_ctypes()
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _vl_lstate_py = max(1, utils._EngineIntArrayType.get_element(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, (0,), validate=False))
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _sdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _sdx = utils._EngineIntScalarType(
        _sdx_py, {'fun_name': fun_name, 'entity': '_sdx'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if _sorder_py == 1:
        _x_dim1 = _n_py
    elif _sorder_py == 2:
        _x_dim1 = _m_py
    else:
        _x_dim1 = 0
    if _sorder_py == 1:
        _x_dim2 = _m_py
    elif _sorder_py == 2:
        _x_dim2 = _n_py
    else:
        _x_dim2 = 0
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _m,
        _theta,
        _sorder,
        _state,
        _x,
        _ldx,
        _sdx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, corrupt statecomm[\'state\'] argument.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid theta: theta = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: theta >= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, invalid sorder.',
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def multivar_students_t(mode, n, df, m, xmu, c, comm, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Student's t-distribution.

    ``multivar_students_t`` sets up a reference vector and generates an
    array of pseudorandom numbers from a multivariate Student's t
    distribution with nu degrees of freedom, mean vector a and
    covariance matrix nu/(nu-2)C.

    For full information please refer to the NAG Library document for
    g05ry

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ryf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``multivar_students_t``.

        `mode` = 2

            Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    df : int
        nu, the number of degrees of freedom of the distribution.

    m : int
        m, the number of dimensions of the distribution.

    xmu : float, shape (`m`)
        a, the vector of means of the distribution.

    c : float, shape (`m`, `m`)
        Matrix which, along with `df`, defines the covariance of the
        distribution. Only the upper triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``multivar_students_t``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or float, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `mode`=0: 0; otherwise: `m`.

        `On exit`, if not None on entry: the array of pseudorandom
        multivariate Student's t vectors generated by the function, with
        `x`[i-1,j-1] holding the jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 3.

        (`errno` 4)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 6)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 8)
            `m` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 10)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    When the covariance matrix is nonsingular (i.e., strictly positive
    definite), the distribution has probability density function

        f(x) = Gamma(((nu+m))/2)/((pi
        v)^(m/2)Gamma(nu/2)|C|^(1/2))[1+(x-a)^TC^-1(x-a)/nu]^((-(nu+m))/2)

    where m is the number of dimensions, nu is the degrees of freedom, a
    is the vector of means, x is the vector of positions and nu/(nu-2)C
    is the covariance matrix.

    The function returns the value

        x = a+sqrt(nu/s)z

    where z is generated by ``dist_normal`` from a Normal distribution
    with mean zero and covariance matrix C and s is generated by
    ``dist_chisq`` from a chi^2-distribution with nu degrees of freedom.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``multivar_students_t``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.multivar_students_t'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ryft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, df.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, xmu[m].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _xmu = utils._EngineFloat64ArrayType(
        xmu, {'fun_name': fun_name, 'entity': 'xmu'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_m_py*(_m_py+1)+2)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _m_py
    exp_shape_x = (_n_py, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _uplo = utils._EngineCharScalarType(
        'U', {'fun_name': fun_name, 'entity': '_uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mode,
        _n,
        _df,
        _m,
        _xmu,
        _c,
        _ldc,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _uplo,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix C is not positive semidefinite',
                    'to machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'm is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 14 and _ifmt == 14:
            _ierr_fl = 10
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def multivar_normal(sorder, mode12, n, m, xmu, c, comm, statecomm, x):
    u"""
    Generates a matrix of pseudorandom numbers from a multivariate
    Normal distribution.

    ``multivar_normal`` sets up a reference vector and generates an
    array of pseudorandom numbers from a multivariate Normal
    distribution with mean vector a and covariance matrix C.

    For full information please refer to the NAG Library document for
    g05rz

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05rzf.html

    Parameters
    ----------
    sorder : int
        Determines the storage order of variates; the (i,j)th variate is
        stored in `x`[i-1,j-1] if `sorder` = 1, and `x`[j-1,i-1] if
        `sorder` = 2, for j = 1,2,...,m, for i = 1,2,...,n.

    mode12 : int
        A code for selecting the operation to be performed by the
        function.

        `mode12` = 0

            Set up reference vector only.

        `mode12` = 1

            Generate variates using reference vector set up in a prior
            call to ``multivar_normal``.

        `mode12` = 2

            Set up reference vector and generate variates.

    n : int
        n, the number of random variates required.

    m : int
        m, the number of dimensions of the distribution.

    xmu : float, shape (`m`)
        a, the vector of means of the distribution.

    c : float, shape (`m`, `m`)
        The covariance matrix of the distribution. Only the upper
        triangle need be set.

    comm : dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode12` = 1, this argument must have been initialized by a
        prior call to ``multivar_normal``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `mode12`=0: 0; if `sorder`=1: `n`;
        otherwise: `m`.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `mode12`=0: 0; if `sorder`=1: `m`;
        otherwise: `n`.

        `On exit`, if not None on entry: the array of pseudorandom
        multivariate Normal vectors generated by the function.

        Two possible storage orders are available.

        If `sorder` = 1 then `x`[i-1,j-1] holds the jth dimension for
        the ith variate.

        If `sorder` = 2 this ordering is reversed and `x`[j-1,i-1] holds
        the jth dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode12` = *<value>*.

            Constraint: `mode12` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 5)
            On entry, the covariance matrix C is not positive
            semidefinite to machine precision.

        (`errno` 7)
            `m` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    When the covariance matrix is nonsingular (i.e., strictly positive
    definite), the distribution has probability density function

        f(x) = sqrt(|C^-1|/(2 pi)^m)exp(-1/2(x-a)^TC^-1(x-a))

    where m is the number of dimensions, C is the covariance matrix, a
    is the vector of means and x is the vector of positions.

    Covariance matrices are symmetric and positive semidefinite.
    Given such a matrix C, there exists a lower triangular matrix L such
    that LL^T = C. L is not unique, if C is singular.

    ``multivar_normal`` decomposes C to find such an L.
    It then stores m, a and L in the reference vector r which is used to
    generate a vector x of independent standard Normal pseudorandom
    numbers.
    It then returns the vector a+Lx, which has the required multivariate
    Normal distribution.

    It should be noted that this function will work with a singular
    covariance matrix C, provided C is positive semidefinite, despite
    the fact that the above formula for the probability density function
    is not valid in that case. Wilkinson (1965) should be consulted if
    further information is required.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``multivar_normal``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Wilkinson, J H, 1965, `The Algebraic Eigenvalue Problem`, Oxford
    University Press, Oxford
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.multivar_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05rzft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mode12.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, xmu[m].
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, m].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._BYREF_STRING, # Input, uplo(len=1).
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of uplo.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((c, 'c'), (x, 'x')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _mode12 = utils._EngineIntScalarType(
        mode12, {'fun_name': fun_name, 'entity': 'mode12'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _xmu = utils._EngineFloat64ArrayType(
        xmu, {'fun_name': fun_name, 'entity': 'xmu'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_m_py, _m_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _mode12_py = utils._EngineIntScalarType.to_py(_mode12)
    if _mode12_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
        )
    if _comm_needs_init:
        _r = utils._EngineFloat64ArrayType.empty_ctypes(_m_py*(_m_py+1)+1)
        comm['r'] = _r
    else:
        _r = comm['r']
    _lr_py = utils._EngineFloat64ArrayType.get_shape(comm['r'], {'fun_name': fun_name, 'entity': 'comm[\'r\']'}, exp_rank=1, validate=False)[0]
    _lr = utils._EngineIntScalarType(
        _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    if _mode12_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode12 != 0',
            )
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode12_py == 0:
        _x_dim1 = 0
    elif _sorder_py == 1:
        _x_dim1 = _n_py
    else:
        _x_dim1 = _m_py
    if _mode12_py == 0:
        _x_dim2 = 0
    elif _sorder_py == 1:
        _x_dim2 = _m_py
    else:
        _x_dim2 = _n_py
    exp_shape_x = (_x_dim1, _x_dim2)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _uplo = utils._EngineCharScalarType(
        'U', {'fun_name': fun_name, 'entity': '_uplo'}, exp_chrlen=1,
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mode12,
        _n,
        _m,
        _xmu,
        _c,
        _ldc,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _uplo,
        _errbuf,
        _ierr,
        _uplo._length_,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode12 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode12 = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the covariance matrix C is not positive semidefinite',
                    'to machine precision.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'm is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_uniform01(n, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a uniform
    distribution over (0,1].

    ``dist_uniform01`` generates a vector of pseudorandom numbers taken
    from a uniform distribution between 0 and 1.

    For full information please refer to the NAG Library document for
    g05sa

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05saf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from a uniform
        distribution over the half closed interval (0,1].

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``dist_uniform01`` generates n values from a uniform distribution
    over the half closed interval (0,1].

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_uniform01``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    See Also
    --------
    :meth:`naginterfaces.library.examples.fit.dim2_spline_ts_sctr_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_uniform01'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05saft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_beta(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a beta distribution.

    ``dist_beta`` generates a vector of pseudorandom numbers taken from
    a beta distribution with parameters a and b.

    For full information please refer to the NAG Library document for
    g05sb

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sbf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the beta distribution.

    b : float
        b, the parameter of the beta distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified beta
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The beta distribution has PDF (probability density function)

        [table omitted]

    One of four algorithms is used to generate the variates depending on
    the values of a and b.
    Let alpha be the maximum and beta be the minimum of a and b.
    Then the algorithms are as follows:

    (i) if alpha < 0.5, Johnk's algorithm is used, see for example
        Dagpunar (1988). This generates the beta variate as
        u_1^1/a/(u_1^1/a+u_2^1/b), where u_1 and u_2 are uniformly
        distributed random variates;

    (#) if beta > 1, the algorithm BB given by Cheng (1978) is used.
        This involves the generation of an observation from a beta
        distribution of the second kind by the envelope rejection method
        using a log-logistic target distribution and then transforming
        it to a beta variate;

    (#) if alpha > 1 and beta < 1, the switching algorithm given by
        Atkinson (1979) is used. The two target distributions used are
        f_1(x) = beta x^beta and f_2(x) = alpha(1-x)^(beta-1), along
        with the approximation to the switching parameter of t =
        (1-beta)/(alpha+1-beta);

    (#) in all other cases, Cheng's BC algorithm (see Cheng (1978)) is
        used with modifications suggested by Dagpunar (1988). This
        algorithm is similar to BB, used when beta > 1, but is tuned for
        small values of a and b.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_beta``.

    References
    ----------
    Atkinson, A C, 1979, `A family of switching algorithms for the
    computer generation of beta random variates`, Biometrika (66),
    141--5

    Cheng, R C H, 1978, `Generating beta variates with nonintegral shape
    parameters`, Comm. ACM (21), 317--322

    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_beta'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_cauchy(n, xmed, semiqr, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a Cauchy
    distribution.

    ``dist_cauchy`` generates a vector of pseudorandom numbers from a
    Cauchy distribution with median a and semi-interquartile range b.

    For full information please refer to the NAG Library document for
    g05sc

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05scf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmed : float
        a, the median of the distribution.

    semiqr : float
        b, the semi-interquartile range of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified Cauchy
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `semiqr` = *<value>*.

            Constraint: `semiqr` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        f(x) = 1/(pi b(1+((x-a)/b)^2)).

    ``dist_cauchy`` returns the value

        a+b(2y_1-1)/(y_2),

    where y_1 and y_2 are a pair of consecutive pseudorandom numbers
    from a uniform distribution over (0,1), such that

        (2y_1-1)^2+y_2^2 <= 1.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_cauchy``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_cauchy'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05scft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmed.
        utils._BYREF_FLOAT64, # Input, semiqr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmed = utils._EngineFloat64ScalarType(
        xmed, {'fun_name': fun_name, 'entity': 'xmed'},
    ).to_ctypes()
    _semiqr = utils._EngineFloat64ScalarType(
        semiqr, {'fun_name': fun_name, 'entity': 'semiqr'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmed,
        _semiqr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, semiqr = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: semiqr >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_chisq(n, df, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a chi^2
    distribution.

    ``dist_chisq`` generates a vector of pseudorandom numbers taken from
    a chi^2-distribution with nu degrees of freedom.

    For full information please refer to the NAG Library document for
    g05sd

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sdf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        chi^2-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    This is the same as a gamma distribution with parameters nu/2 and 2.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_chisq``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_chisq'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, df.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _df,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_dirichlet(n, m, a, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a Dirichlet
    distribution.

    ``dist_dirichlet`` generates a vector of pseudorandom numbers taken
    from a Dirichlet distribution.

    For full information please refer to the NAG Library document for
    g05se

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sef.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of dimensions of the distribution.

    a : float, shape (`m`)
        The parameter vector for the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`, `m`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        Dirichlet distribution, with `x`[i-1,j-1] holding the jth
        dimension for the ith variate.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `m` = *<value>*.

            Constraint: `m` > 0.

        (`errno` 3)
            On entry, at least one `a`[i] <= 0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    where x = {x_1,x_2,...,x_m} is a vector of dimension m, such that
    x_i > 0 for all i and sum(x_i) = 1.

    ``dist_dirichlet`` generates a draw from a Dirichlet distribution by
    first drawing m independent samples, y_i~gamma(alpha_i,1), i.e.,
    independent draws from a gamma distribution with parameters alpha_i
    > 0 and one, and then setting x_i = y_i/ sum(y_j).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_dirichlet``.

    References
    ----------
    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_dirichlet'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05seft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, a[m].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineFloat64ArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py, _m_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _m,
        _a,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one a[i] <= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_exp(n, a, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from an exponential
    distribution.

    ``dist_exp`` generates a vector of pseudorandom numbers from a
    (negative) exponential distribution with mean a.

    For full information please refer to the NAG Library document for
    g05sf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sff.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the mean of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        exponential distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The exponential distribution has PDF (probability density function):

        [table omitted]

    ``dist_exp`` returns the values

        x_i = -aln(y_i)

    where y_i are the next n numbers generated by a uniform (0,1]
    generator.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_exp``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_exp'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_expmix(n, nmix, a, wgt, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from an exponential mix
    distribution.

    ``dist_expmix`` generates a vector of pseudorandom numbers from an
    exponential mix distribution composed of m exponential distributions
    each having a mean a_i and weight w_i.

    For full information please refer to the NAG Library document for
    g05sg

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sgf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    nmix : int
        m, the number of exponential distributions in the mix.

    a : float, shape (`nmix`)
        The m parameters a_i for the m exponential distributions in the
        mix.

    wgt : float, shape (`nmix`)
        The m weights w_i for the m exponential distributions in the
        mix.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        exponential mix distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `nmix` = *<value>*.

            Constraint: `nmix` >= 1.

        (`errno` 3)
            On entry, `a`[*<value>*] = *<value>*.

            Constraint: `a`[i-1] > 0.0.

        (`errno` 4)
            On entry, `wgt`[*<value>*] = *<value>*.

            Constraint: `wgt`[i-1] >= 0.0.

        (`errno` 4)
            On entry, sum of `wgt` = *<value>*.

            Constraint: sum of `wgt` = 1.0.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    where sum_i = 1^mw_i = 1 and a_i > 0, w_i >= 0.

    ``dist_expmix`` returns the values x_i by selecting, with
    probability w_j, random variates from an exponential distribution
    with parameter a_j.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_expmix``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_expmix'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, nmix.
        utils._EngineFloat64ArrayType.c_type, # Input, a[nmix].
        utils._EngineFloat64ArrayType.c_type, # Input, wgt[nmix].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _nmix = utils._EngineIntScalarType(
        nmix, {'fun_name': fun_name, 'entity': 'nmix'},
    ).to_ctypes()
    _nmix_py = utils._EngineIntScalarType.to_py(_nmix)
    _a = utils._EngineFloat64ArrayType(
        a, {'fun_name': fun_name, 'entity': 'a'},
        exp_shape=(_nmix_py,),
    ).to_ctypes()
    _wgt = utils._EngineFloat64ArrayType(
        wgt, {'fun_name': fun_name, 'entity': 'wgt'},
        exp_shape=(_nmix_py,),
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _nmix,
        _a,
        _wgt,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmix = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: nmix >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: a[i-1] > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, wgt[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: wgt[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sum of wgt = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sum of wgt = 1.0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_f(n, df1, df2, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from an F-distribution.

    ``dist_f`` generates a vector of pseudorandom numbers taken from an
    F (or Fisher's variance ratio) distribution with mu and nu degrees
    of freedom.

    For full information please refer to the NAG Library document for
    g05sh

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05shf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df1 : int
        mu, the number of degrees of freedom of the distribution.

    df2 : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        F-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df1` = *<value>*.

            Constraint: `df1` >= 1.

        (`errno` 3)
            On entry, `df2` = *<value>*.

            Constraint: `df2` >= 1.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    ``dist_f`` calculates the values

        nu y_i/mu z_i, i = 1,2,...,n,

    where y_i and z_i are generated by ``dist_gamma`` from gamma
    distributions with parameters (1/2 mu,2) and (1/2 nu,2) respectively
    (i.e., from chi^2-distributions with mu and nu degrees of freedom).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_f``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_f'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05shft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, df1.
        utils._BYREF_INT, # Input, df2.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _df1 = utils._EngineIntScalarType(
        df1, {'fun_name': fun_name, 'entity': 'df1'},
    ).to_ctypes()
    _df2 = utils._EngineIntScalarType(
        df2, {'fun_name': fun_name, 'entity': 'df2'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _df1,
        _df2,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df1 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df1 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df2 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df2 >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_gamma(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a gamma
    distribution.

    ``dist_gamma`` generates a vector of pseudorandom numbers taken from
    a gamma distribution with parameters a and b.

    For full information please refer to the NAG Library document for
    g05sj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sjf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the gamma distribution.

    b : float
        b, the parameter of the gamma distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified gamma
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The gamma distribution has PDF (probability density function)

        [table omitted]

    One of three algorithms is used to generate the variates depending
    upon the value of a:

    (i) if a < 1, a switching algorithm described by Dagpunar (1988)
        (called G6) is used. The target distributions are f_1(x) =
        cax^(a-1)/t^a and f_2(x) = (1-c)e^(-(x-t)), where c =
        t/(t+ae^(-t)), and the switching parameter, t, is taken as 1-a.
        This is similar to Ahrens and Dieter's GS algorithm (see Ahrens
        and Dieter (1974)) in which t = 1;

    (#) if a = 1, the gamma distribution reduces to the exponential
        distribution and the method based on the logarithmic
        transformation of a uniform random variate is used;

    (#) if a > 1, the algorithm given by Best (1978) is used. This is
        based on using a Student's t-distribution with two degrees of
        freedom as the target distribution in an envelope rejection
        method.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_gamma``.

    References
    ----------
    Ahrens, J H and Dieter, U, 1974, `Computer methods for sampling from
    gamma, beta, Poisson and binomial distributions`, Computing (12),
    223--46

    Best, D J, 1978, `Letter to the Editor`, Appl. Statist. (27), 181

    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Hastings, N A J and Peacock, J B, 1975, `Statistical Distributions`,
    Butterworth
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_gamma'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_normal(n, xmu, var, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a Normal
    distribution.

    ``dist_normal`` generates a vector of pseudorandom numbers taken
    from a Normal (Gaussian) distribution with mean mu and variance
    sigma^2.

    For full information please refer to the NAG Library document for
    g05sk

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05skf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmu : float
        mu, the mean of the distribution.

    var : float
        sigma^2, the variance of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified Normal
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability distribution function)

        f(x) = 1/(sigma*sqrt(2 pi))exp(-((x-mu)^2)/2 sigma^2).

    ``dist_normal`` uses the algorithm of Wichura (1988).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_normal``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley

    Wichura, 1988, `Algorithm AS 241: the percentage points of the
    Normal distribution`, Appl. Statist. (37), 477--484
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05skft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmu.
        utils._BYREF_FLOAT64, # Input, var.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmu = utils._EngineFloat64ScalarType(
        xmu, {'fun_name': fun_name, 'entity': 'xmu'},
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmu,
        _var,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_logistic(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a logistic
    distribution.

    ``dist_logistic`` generates a vector of pseudorandom numbers from a
    logistic distribution with mean a and spread b.

    For full information please refer to the NAG Library document for
    g05sl

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05slf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the mean of the distribution.

    b : float
        b, the spread of the distribution, where 'spread' is
        (sqrt(3))/pi*standard deviation.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        logistic distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        f(x) = (e^((x-a)/b))/b(1+e^((x-a)/b))^2.

    ``dist_logistic`` returns the value

        a+bln(y/(1-y)),

    where y is a pseudorandom number uniformly distributed over (0,1).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_logistic``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_logistic'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05slft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_lognormal(n, xmu, var, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a log-normal
    distribution.

    ``dist_lognormal`` generates a vector of pseudorandom numbers from a
    log-normal distribution with parameters mu and sigma^2.

    For full information please refer to the NAG Library document for
    g05sm

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05smf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmu : float
        mu, the mean of the distribution of ln(x).

    var : float
        sigma^2, the variance of the distribution of ln(x).

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        log-normal distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `xmu` is too large to take the exponential of
            `xmu` = *<value>*.

        (`errno` 3)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    i.e., ln(x) is normally distributed with mean mu and variance
    sigma^2. ``dist_lognormal`` evaluates exp(y_i), where the y_i are
    generated by ``dist_normal`` from a Normal distribution with mean mu
    and variance sigma^2, for i = 1,2,...,n.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_lognormal``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_lognormal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05smft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmu.
        utils._BYREF_FLOAT64, # Input, var.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmu = utils._EngineFloat64ScalarType(
        xmu, {'fun_name': fun_name, 'entity': 'xmu'},
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmu,
        _var,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmu is too large to take the exponential of',
                    'xmu = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_students_t(n, df, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a Student's
    t-distribution.

    ``dist_students_t`` generates a vector of pseudorandom numbers taken
    from a Student's t-distribution with nu degrees of freedom.

    For full information please refer to the NAG Library document for
    g05sn

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05snf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    df : int
        nu, the number of degrees of freedom of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        Student's t-distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `df` = *<value>*.

            Constraint: `df` >= 1.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        f(x) = (((nu-1)/2)!)/((1/2
        nu-1)!sqrt(pinu)(1+(x^2)/nu)^(1/2(nu+1))).

    ``dist_students_t`` calculates the values

        y_i sqrt(nu/(z_i)), i = 1,...,n

    where the y_i are generated by ``dist_normal`` from a Normal
    distribution with mean 0 and variance 1.0, and the z_i are generated
    by ``dist_gamma`` from a gamma distribution with parameters 1/2 nu
    and 2 (i.e., from a chi^2-distribution with nu degrees of freedom).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_students_t``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_students_t'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05snft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, df.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _df = utils._EngineIntScalarType(
        df, {'fun_name': fun_name, 'entity': 'df'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _df,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, df = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: df >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_triangular(n, xmin, xmed, xmax, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a triangular
    distribution.

    ``dist_triangular`` generates a vector of pseudorandom numbers from
    a triangular distribution with parameters x_min, x_med and x_max.

    For full information please refer to the NAG Library document for
    g05sp

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05spf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    xmin : float
        The end point x_min of the triangular distribution.

    xmed : float
        The median of the distribution x_med (also the location of the
        vertex of the triangular distribution at which the PDF reaches a
        maximum).

    xmax : float
        The end point x_max of the triangular distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified
        triangular distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `xmed` = *<value>* and `xmin` = *<value>*.

            Constraint: `xmed` >= `xmin`.

        (`errno` 4)
            On entry, `xmax` = *<value>* and `xmed` = *<value>*.

            Constraint: `xmax` >= `xmed`.

        (`errno` 5)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The triangular distribution has a PDF (probability density function)
    that is triangular in profile.
    The base of the triangle ranges from x = x_min to x = x_max and the
    PDF has a maximum value of 2/(x_max-x_min) at x = x_med.
    If x_min = x_med = x_max then x = x_med with probability 1;
    otherwise the triangular distribution has PDF:

        [table omitted]

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_triangular``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_triangular'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05spft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmed.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmed = utils._EngineFloat64ScalarType(
        xmed, {'fun_name': fun_name, 'entity': 'xmed'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _xmin,
        _xmed,
        _xmax,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmed = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmin = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmed >= xmin.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmed = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmax >= xmed.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_uniform(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a uniform
    distribution over [a,b].

    ``dist_uniform`` generates a vector of pseudorandom numbers
    uniformly distributed over the interval [a,b].

    For full information please refer to the NAG Library document for
    g05sq

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05sqf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        The end points a and b of the uniform distribution.

    b : float
        The end points a and b of the uniform distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified uniform
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` >= `a`.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    If a = 0 and b = 1, ``dist_uniform`` returns the next n values y_i
    from a uniform (0,1] generator (see ``dist_uniform01`` for details).

    For other values of a and b, ``dist_uniform`` applies the
    transformation

        x_i = a+(b-a)y_i.

    The function ensures that the values x_i lie in the closed interval
    [a,b].

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_uniform``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_uniform'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05sqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s} and b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: b >= a.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_vonmises(n, vk, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a von Mises
    distribution.

    ``dist_vonmises`` generates a vector of pseudorandom numbers from a
    von Mises distribution with concentration parameter kappa.

    For full information please refer to the NAG Library document for
    g05sr

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05srf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    vk : float
        kappa, the concentration parameter of the required von Mises
        distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified von
        Mises distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `vk` <= 0.0 or `vk` too large: `vk` = *<value>*.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The von Mises distribution is a symmetric distribution used in the
    analysis of circular data.
    The PDF (probability density function) of this distribution on the
    circle with mean direction mu_0 = 0 and concentration parameter
    kappa, can be written as:

        f(theta) = e^(kappa cos(theta))/(2 pi I_0(kappa)),

    where theta is reduced modulo 2 pi so that -pi <= theta < pi and
    kappa >= 0.
    For very small kappa the distribution is almost the uniform
    distribution, whereas for kappa→ infinity all the probability is
    concentrated at one point.

    The n variates, theta_1,theta_2,...,theta_n, are generated using an
    envelope rejection method with a wrapped Cauchy target distribution
    as proposed by Best and Fisher (1979) and described by Dagpunar
    (1988).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_vonmises``.

    References
    ----------
    Best, D J and Fisher, N I, 1979, `Efficient simulation of the von
    Mises distribution`, Appl. Statist. (28), 152--157

    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Mardia, K V, 1972, `Statistics of Directional Data`, Academic Press
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_vonmises'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05srft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, vk.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _vk = utils._EngineFloat64ScalarType(
        vk, {'fun_name': fun_name, 'entity': 'vk'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _vk,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, vk <= 0.0 or vk too large: vk = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def dist_weibull(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom numbers from a Weibull
    distribution.

    ``dist_weibull`` generates a vector of pseudorandom numbers from a
    two parameter Weibull distribution with shape parameter a and scale
    parameter b.

    For full information please refer to the NAG Library document for
    g05ss

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ssf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the shape parameter of the distribution.

    b : float
        b, the scale parameter of the distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : float, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified Weibull
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `a` = *<value>*.

            Constraint: `a` > 0.0.

        (`errno` 3)
            On entry, `b` = *<value>*.

            Constraint: `b` > 0.0.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The distribution has PDF (probability density function)

        [table omitted]

    ``dist_weibull`` returns the value (-bln(y))^(1/a), where y is a
    pseudorandom number from a uniform distribution over (0,1].

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``dist_weibull``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.dist_weibull'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ssft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._BYREF_FLOAT64, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineFloat64ArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineFloat64ScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineFloat64ArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, b = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: b > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_binomial(mode, n, m, p, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a binomial
    distribution.

    ``int_binomial`` generates a vector of pseudorandom integers from
    the discrete binomial distribution with parameters m and p.

    For full information please refer to the NAG Library document for
    g05ta

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05taf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_binomial``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of trials of the distribution.

    p : float
        p, the probability of success of the binomial distribution.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_binomial``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the n pseudorandom numbers from
        the specified binomial distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` <= 1.0.

        (`errno` 5)
            `p` or `m` is not the same as when `comm`['r'] was set up in
            a previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 5)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 7)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_binomial`` generates n integers x_i from a discrete binomial
    distribution, where the probability of x_i = I is

        P(x_i = I) = (m!)/(I!(m-I)!)*p^I*(1-p)^(m-I), I = 0,1,...,m,

    where m >= 0 and 0 <= p <= 1.
    This represents the probability of achieving I successes in m trials
    when the probability of success at a single trial is p.

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_binomial`` with the
    same parameter values can then use this reference vector to generate
    further variates.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_binomial``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_binomial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05taft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _p_py = utils._EngineFloat64ScalarType.to_py(_p)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 22+20*int(ceil(sqrt(abs(_m_py*_p_py*(1-_p_py)))))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _m,
        _p,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= p <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'p or m is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of p = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def logical(n, p, statecomm, x):
    u"""
    Generates a vector of pseudorandom logical values.

    ``logical`` generates a vector of pseudorandom logical values --
    True with probability p and False with probability (1-p).

    For full information please refer to the NAG Library document for
    g05tb

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tbf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom logical values to be generated.

    p : float
        Must contain the probability of ``logical`` returning True.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : bool, shape (`n`), modified in place
        `On exit`: the n logical values.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 2)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` <= 1.0.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``logical`` generates n logical values x_i from the relation

        y_i < p

    where y_i is a pseudorandom number from a uniform distribution over
    (0,1], generated by ``dist_uniform01`` using the values of
    `statecomm`['state'] as input to this function.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``logical``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.logical'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, p.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineLogicalArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineLogicalArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _p,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= p <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_geom(mode, n, p, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a geometric
    distribution.

    ``int_geom`` generates a vector of pseudorandom integers from the
    discrete geometric distribution with probability p of success at a
    trial.

    For full information please refer to the NAG Library document for
    g05tc

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tcf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_geom``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    p : float
        The parameter p of the geometric distribution representing the
        probability of success at a single trial.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_geom``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the n pseudorandom numbers from
        the specified geometric distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `p` = *<value>*.

            Constraint: machine precision <= `p` <= 1.0.

        (`errno` 3)
            `p` is so small that lr would have to be larger than the
            largest representable integer. Use `mode` = 3 instead. `p` =
            *<value>*

        (`errno` 4)
            `p` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_geom`` generates n integers x_i from a discrete geometric
    distribution, where the probability of x_i = I (a first success
    after I+1 trials) is

        P(x_i = I) = p*(1-p)^I, I = 0,1,....

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_geom`` with the same
    parameter value can then use this reference vector to generate
    further variates.
    If the search table is not used (as recommended for small values of
    p) then a direct transformation of uniform variates is used.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_geom``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_geom'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _p_py = utils._EngineFloat64ScalarType.to_py(_p)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 8+int(ceil(float(42)/float(_p_py)))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _p,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: machine precision <= p <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'p is so small that lr would have to be larger than the largest representable',
                    'integer. Use mode = 3 instead. p = {:s}'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'p is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of p = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_general(mode, n, p, np, ip1, itype, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a general discrete
    distribution.

    ``int_general`` generates a vector of pseudorandom integers from a
    discrete distribution with a given PDF (probability density
    function) or CDF (cumulative distribution function) p.

    For full information please refer to the NAG Library document for
    g05td

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tdf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_general``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    p : float, shape (`np`)
        The PDF or CDF of the distribution.

    np : int
        The number of values supplied in `p` defining the PDF or CDF of
        the discrete distribution.

    ip1 : int
        The value of the variate, a whole number, to which the
        probability in `p`[0] corresponds.

    itype : int
        Indicates the type of information contained in `p`.

        `itype` = 1

            `p` contains a probability distribution function (PDF).

        `itype` = 2

            `p` contains a cumulative distribution function (CDF).

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_general``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: contains n pseudorandom numbers
        from the specified discrete distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1 or 2.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, at least one element of the vector `p` is less
            than 0.0 or greater than 1.0.

        (`errno` 3)
            On entry, `itype` = 1 and the sum of the elements of `p` do
            not equal one.

        (`errno` 3)
            On entry, `itype` = 2 and the values of `p` are not all in
            stricly ascending order.

        (`errno` 3)
            On entry, `p`[`np`-1] = *<value>*.

            Constraint: if `itype` = 2, `p`[`np`-1] = 1.0.

        (`errno` 4)
            On entry, `np` = *<value>*.

            Constraint: `np` > 0.

        (`errno` 6)
            On entry, `itype` = *<value>*.

            Constraint: `itype` = 1 or 2.

        (`errno` 7)
            The value of `np` or `ip1` is not the same as when
            `comm`['r'] was set up in a previous call.

            Previous value of `np` = *<value>* and `np` = *<value>*.

            Previous value of `ip1` = *<value>* and `ip1` = *<value>*.

        (`errno` 7)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 9)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_general`` generates a sequence of n integers x_i, from a
    discrete distribution defined by information supplied in `p`.
    This may either be the PDF or CDF of the distribution.
    A reference vector is first set up to contain the CDF of the
    distribution in its higher elements, followed by an index.

    Setting up the reference vector and subsequent generation of
    variates can each be performed by separate calls to ``int_general``
    or may be combined in a single call.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_general``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_general'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, p[np].
        utils._BYREF_INT, # Input, np.
        utils._BYREF_INT, # Input, ip1.
        utils._BYREF_INT, # Input, itype.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _np = utils._EngineIntScalarType(
        np, {'fun_name': fun_name, 'entity': 'np'},
    ).to_ctypes()
    _np_py = utils._EngineIntScalarType.to_py(_np)
    _p = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_np_py,),
    ).to_ctypes()
    _ip1 = utils._EngineIntScalarType(
        ip1, {'fun_name': fun_name, 'entity': 'ip1'},
    ).to_ctypes()
    _itype = utils._EngineIntScalarType(
        itype, {'fun_name': fun_name, 'entity': 'itype'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 10+int(ceil(1.4*_np_py))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _p,
        _np,
        _ip1,
        _itype,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of the vector p is less than 0.0 or',
                    'greater than 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = 1 and the sum of the elements of p do not equal one.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 33 and _ifmt == 33:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = 2 and the values of p are not all in',
                    'stricly ascending order.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 43 and _ifmt == 43:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p[np-1] = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: if itype = 2, p[np-1] = 1.0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: np > 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, itype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: itype = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of np or ip1 is not the same as when comm[\'r\'] was set up',
                    'in a previous call.',
                    'Previous value of np = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of ip1 = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'ip1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 27 and _ifmt == 27:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _ierr_fl = 9
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_hypergeom(mode, n, ns, np, m, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a hypergeometric
    distribution.

    ``int_hypergeom`` generates a vector of pseudorandom integers from
    the discrete hypergeometric distribution of the number of specified
    items in a sample of size l, taken from a population of size k with
    m specified items in it.

    For full information please refer to the NAG Library document for
    g05te

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tef.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_hypergeom``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    ns : int
        l, the sample size of the hypergeometric distribution.

    np : int
        k, the population size of the hypergeometric distribution.

    m : int
        m, the number of specified items of the hypergeometric
        distribution.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_hypergeom``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the pseudorandom numbers from
        the specified hypergeometric distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `ns` = *<value>* and `np` = *<value>*.

            Constraint: 0 <= `ns` <= `np`.

        (`errno` 4)
            On entry, `np` = *<value>*.

            Constraint: `np` >= 0.

        (`errno` 5)
            On entry, `m` = *<value>* and `np` = *<value>*.

            Constraint: 0 <= `m` <= `np`.

        (`errno` 6)
            The value of `ns`, `np` or `m` is not the same as when
            `comm`['r'] was set up in a previous call with `mode` = 0 or
            2.

        (`errno` 6)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_hypergeom`` generates n integers x_i from a discrete
    hypergeometric distribution, where the probability of x_i = I is

        [table omitted]

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_hypergeom`` with the
    same parameter values can then use this reference vector to generate
    further variates.
    The reference array is generated by a recurrence relation if
    lm(k-l)(k-m) < 50k^3, otherwise Stirling's approximation is used.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_hypergeom``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_hypergeom'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05teft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_INT, # Input, np.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _np = utils._EngineIntScalarType(
        np, {'fun_name': fun_name, 'entity': 'np'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _np_py = utils._EngineIntScalarType.to_py(_np)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 28+20*int(ceil(sqrt(abs(float((_ns_py*_m_py*(_np_py-_m_py)*(_np_py-_ns_py)))/float(_np_py**3)))))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _ns,
        _np,
        _m,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s} and np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= ns <= np.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: np >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= m <= np.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of ns, np or m is not the same as when comm[\'r\'] was set up',
                    'in a previous call with mode = 0 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_log(mode, n, a, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a logarithmic
    distribution.

    ``int_log`` generates a vector of pseudorandom integers from the
    discrete logarithmic distribution with parameter a.

    For full information please refer to the NAG Library document for
    g05tf

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tff.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_log``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    a : float
        a, the parameter of the logarithmic distribution.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_log``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the n pseudorandom numbers from
        the specified logarithmic distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>*.

            Constraint: 0.0 < `a` < 1.0.

        (`errno` 4)
            The value of `a` is not the same as when `comm`['r'] was set
            up in a previous call.

            Previous value of `a` = *<value>* and `a` = *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_log`` generates n integers x_i from a discrete logarithmic
    distribution, where the probability of x_i = I is

        P(x_i = I) = -(a^I)/(I*log(1-a)), I = 1,2,...,

    where 0 < a < 1.

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_log`` with the same
    parameter value can then use this reference vector to generate
    further variates.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_log``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_log'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tfft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, a.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineFloat64ScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _a_py = utils._EngineFloat64ScalarType.to_py(_a)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                if _a_py != 1:
                    _r_dim1 = 18+int(ceil(float(40)/float(1-_a_py)))
                else:
                    _r_dim1 = 100
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _a,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < a < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of a is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of a = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'a = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_multinomial(mode, n, m, k, p, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a multinomial
    distribution.

    ``int_multinomial`` generates a sequence of n variates, each
    consisting of k pseudorandom integers, from the discrete multinomial
    distribution with k outcomes and m trials, where the outcomes have
    probabilities p_1,p_2,...,p_k respectively.

    For full information please refer to the NAG Library document for
    g05tg

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tgf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_multinomial``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of trials of the multinomial distribution.

    k : int
        k, the number of possible outcomes of the multinomial
        distribution.

    p : float, shape (`k`)
        Contains the probabilities p_j, for j = 1,2,...,k, of the k
        possible outcomes of the multinomial distribution.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_multinomial``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`, :), modified in place
        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `mode`=0: 0; otherwise: `k`.

        `On exit`, if not None on entry: the first n rows of
        `x`[i-1,j-1] each contain k pseudorandom numbers representing a
        k-dimensional variate from the specified multinomial
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `k` = *<value>*.

            Constraint: `k` >= 2.

        (`errno` 5)
            On entry, at least one element of the vector `p` is less
            than 0.0 or greater than 1.0.

        (`errno` 5)
            On entry, the sum of the elements of `p` do not equal one.

        (`errno` 6)
            The value of `m` or `k` is not the same as when `comm`['r']
            was set up in a previous call.

            Previous value of `m` = *<value>* and `m` = *<value>*.

            Previous value of `k` = *<value>* and `k` = *<value>*.

        (`errno` 6)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_multinomial`` generates a sequence of n groups of k integers
    x_{i,j}, for i = 1,2,...,n, for j = 1,2,...,k, from a multinomial
    distribution with m trials and k outcomes, where the probability of
    x_{i,j} = I_j for each j = 1,2,...,k is

        P(i_1 = I_1,...,i_k = I_k) = (m!)/(prod_j = 1^kI_j!)prod_j =
        1^kp_j^I_j = (m!)/(I_1!I_2!...I_k!)p_1^I_1p_2^I_2...p_k^I_k,

    where

        sum_j = 1^kp_j = 1 and sum_j = 1^kI_j = m.

    A single trial can have several outcomes (k) and the probability of
    achieving each outcome is known (p_j).
    After m trials each outcome will have occurred a certain number of
    times.
    The k numbers representing the numbers of occurrences for each
    outcome after m trials is then a single sample from the multinomial
    distribution defined by the parameters k, m and p_j, for j =
    1,2,...,k.
    This function returns n such samples.

    When k = 2 this distribution is equivalent to the binomial
    distribution with parameters m and p = p_1 (see ``int_binomial``).

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_multinomial`` with
    the same parameter values can then use this reference vector to
    generate further variates.
    The reference array is generated only for the outcome with greatest
    probability.
    The number of successes for the outcome with greatest probability is
    calculated first as for the binomial distribution (see
    ``int_binomial``); the number of successes for other outcomes are
    calculated in turn for the remaining reduced multinomial
    distribution; the number of successes for the final outcome is
    simply calculated to ensure that the total number of successes is m.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_multinomial``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_multinomial'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tgft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_INT, # Input, k.
        utils._EngineFloat64ArrayType.c_type, # Input, p[k].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[ldx, tdx].
        utils._BYREF_INT, # Input, ldx.
        utils._BYREF_INT, # Input, tdx.
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((x, 'x'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _k = utils._EngineIntScalarType(
        k, {'fun_name': fun_name, 'entity': 'k'},
    ).to_ctypes()
    _k_py = utils._EngineIntScalarType.to_py(_k)
    _p_cls = utils._EngineFloat64ArrayType(
        p, {'fun_name': fun_name, 'entity': 'p'},
        exp_shape=(_k_py,),
    )
    _p = _p_cls.to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 30+20*int(ceil(sqrt(abs(_m_py*_p_cls.get_max(0, _k_py-1)*(1-_p_cls.get_max(0, _k_py-1))))))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ldx_py = utils._EngineIntArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='l', allow_none=True)
    _ldx = utils._EngineIntScalarType(
        _ldx_py, {'fun_name': fun_name, 'entity': '_ldx'},
    ).to_ctypes()
    _tdx_py = utils._EngineIntArrayType.get_nd(x, {'fun_name': fun_name, 'entity': 'x'}, exp_rank=2, sorder=_m_sorder, dim_type='s', allow_none=True)
    _tdx = utils._EngineIntScalarType(
        _tdx_py, {'fun_name': fun_name, 'entity': '_tdx'},
    ).to_ctypes()
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    if _mode_py == 0:
        _x_dim2 = 0
    else:
        _x_dim2 = _k_py
    exp_shape_x = (_n_py, _x_dim2)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=exp_shape_x,
        mutual_sorder=_m_sorder,
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _mode,
        _n,
        _m,
        _k,
        _p,
        _r,
        _lr,
        _state,
        _x,
        _ldx,
        _tdx,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: k >= 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of the vector p is less than 0.0 or',
                    'greater than 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the sum of the elements of p do not equal one.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'The value of m or k is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of k = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'k = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 26 and _ifmt == 26:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_negbin(mode, n, m, p, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a negative binomial
    distribution.

    ``int_negbin`` generates a vector of pseudorandom integers from the
    discrete negative binomial distribution with parameter m and
    probability p of success at a trial.

    For full information please refer to the NAG Library document for
    g05th

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05thf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_negbin``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    m : int
        m, the number of failures of the distribution.

    p : float
        p, the parameter of the negative binomial distribution
        representing the probability of success at a single trial.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_negbin``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the n pseudorandom numbers from
        the specified negative binomial distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 0.

        (`errno` 4)
            On entry, `p` = *<value>*.

            Constraint: 0.0 <= `p` < 1.0.

        (`errno` 5)
            `p` or `m` is not the same as when `comm`['r'] was set up in
            a previous call.

            Previous value of `p` = *<value>* and `p` = *<value>*.

            Previous value of `m` = *<value>* and `m` = *<value>*.

        (`errno` 5)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 7)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_negbin`` generates n integers x_i from a discrete negative
    binomial distribution, where the probability of x_i = I (I successes
    before m failures) is

        P(x_i = I) = ((m+I-1)!)/(I!(m-1)!)*p^I*(1-p)^m, I = 0,1,....

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_negbin`` with the
    same parameter value can then use this reference vector to generate
    further variates.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_negbin``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_negbin'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05thft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, p.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _p = utils._EngineFloat64ScalarType(
        p, {'fun_name': fun_name, 'entity': 'p'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _p_py = utils._EngineFloat64ScalarType.to_py(_p)
    _m_py = utils._EngineIntScalarType.to_py(_m)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                if _p_py >= 1 or _p_py < 0:
                    _r_dim1 = 0
                else:
                    if _p_py != 1:
                        _r_dim1 = 28+int(ceil(float(20*sqrt(abs(_m_py*_p_py))+30*_p_py)/float(1-_p_py)))
                    else:
                        _r_dim1 = 100
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _m,
        _p,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= p < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'p or m is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of p = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'p = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Previous value of m = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 25 and _ifmt == 25:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _ierr_fl = 7
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_poisson(mode, n, lamda, comm, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a Poisson
    distribution.

    ``int_poisson`` generates a vector of pseudorandom integers from the
    discrete Poisson distribution with mean lambda.

    For full information please refer to the NAG Library document for
    g05tj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tjf.html

    Parameters
    ----------
    mode : int
        A code for selecting the operation to be performed by the
        function.

        `mode` = 0

            Set up reference vector only.

        `mode` = 1

            Generate variates using reference vector set up in a prior
            call to ``int_poisson``.

        `mode` = 2

            Set up reference vector and generate variates.

        `mode` = 3

            Generate variates without using the reference vector.

    n : int
        n, the number of pseudorandom numbers to be generated.

    lamda : float
        lambda, the mean of the Poisson distribution.

    comm : None or dict, communication object, modified in place
        Communication structure for the reference vector.

        If `mode` = 1, this argument must have been initialized by a
        prior call to ``int_poisson``.

        If `mode` = 3, `comm` is not referenced and may be None.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : None or int, shape (`n`), modified in place
        `On exit`, if not None on entry: the n pseudorandom numbers from
        the specified Poisson distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `mode` = *<value>*.

            Constraint: `mode` = 0, 1, 2 or 3.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `lamda` = *<value>*.

            Constraint: `lamda` >= 0.0.

        (`errno` 3)
            `lamda` is such that lr would have to be larger than the
            largest representable integer. Use `mode` = 3 instead.
            `lamda` = *<value>*.

        (`errno` 4)
            `lamda` is not the same as when `comm`['r'] was set up in a
            previous call.

            Previous value of `lamda` = *<value>* and `lamda` =
            *<value>*.

        (`errno` 4)
            On entry, some of the elements of the array `comm`['r'] have
            been corrupted or have not been initialized.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_poisson`` generates n integers x_i from a discrete Poisson
    distribution with mean lambda, where the probability of x_i = I is

        P(x_i = I) = lambda^I*e^(-lambda)/I!, I = 0,1,...,

    where lambda >= 0.

    The variates can be generated with or without using a search table
    and index.
    If a search table is used then it is stored with the index in a
    reference vector and subsequent calls to ``int_poisson`` with the
    same parameter values can then use this reference vector to generate
    further variates.
    The reference array is found using a recurrence relation if lambda
    is less than 50 and by Stirling's formula otherwise.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_poisson``.

    References
    ----------
    Kendall, M G and Stuart, A, 1969, `The Advanced Theory of Statistics
    (Volume 1)`, (3rd Edition), Griffin

    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_poisson'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, mode.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_FLOAT64, # Input, lamda.
        utils._EngineFloat64ArrayType.c_type, # Input/Output, r[lr].
        utils._BYREF_INT, # Input, lr.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _mode = utils._EngineIntScalarType(
        mode, {'fun_name': fun_name, 'entity': 'mode'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _lamda = utils._EngineFloat64ScalarType(
        lamda, {'fun_name': fun_name, 'entity': 'lamda'},
    ).to_ctypes()
    _vl_dummy_r = utils._EngineFloat64ArrayType.empty_ctypes(0)
    _mode_py = utils._EngineIntScalarType.to_py(_mode)
    if _mode_py == 1:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode = 1',
            )
    if _mode_py == 1:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            req_keys=('r',),
        )
    else:
        _comm_needs_init = utils._check_comm(
            comm, {'fun_name': fun_name, 'entity': 'comm'},
            opt_keys=('r',),
            allow_none=True,
        )
    if _mode_py != 3:
        if comm is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'comm'},
                'must not be None when mode != 3',
            )
    _lamda_py = utils._EngineFloat64ScalarType.to_py(_lamda)
    if comm is None:
        _r = _vl_dummy_r
    else:
        if _comm_needs_init:
            if _mode_py in (0, 2):
                _r_dim1 = 30+int(ceil(20*sqrt(abs(_lamda_py))+_lamda_py))
            elif _mode_py == 1:
                _r_dim1 = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
            else:
                _r_dim1 = 0
            _r = utils._EngineFloat64ArrayType.empty_ctypes(_r_dim1)
            comm['r'] = _r
        else:
            _r = comm['r']
    if comm is None:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_vl_dummy_r, {'fun_name': fun_name, 'entity': '_vl_dummy_r'}, exp_rank=1)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    else:
        _lr_py = utils._EngineFloat64ArrayType.get_shape(_r, {'fun_name': fun_name, 'entity': '_r'}, exp_rank=1, validate=False)[0]
        _lr = utils._EngineIntScalarType(
            _lr_py, {'fun_name': fun_name, 'entity': '_lr'},
        ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    if _mode_py != 0:
        if x is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'x'},
                'must not be None when mode != 0',
            )
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
        allow_none=True,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _mode,
        _n,
        _lamda,
        _r,
        _lr,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, mode = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: mode = 0, 1, 2 or 3.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, lamda = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: lamda >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 23 and _ifmt == 23:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'lamda is such that lr would have to be larger than the largest',
                    'representable integer. Use mode = 3 instead.',
                    'lamda = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'lamda is not the same as when comm[\'r\'] was set up in a previous call.',
                    'Previous value of lamda = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'lamda = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 24 and _ifmt == 24:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, some of the elements of the array comm[\'r\'] have been corrupted or',
                    'have not been initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        if x is not None:
            _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_poisson_varmean(m, vlamda, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a Poisson
    distribution with varying mean.

    ``int_poisson_varmean`` generates a vector of pseudorandom integers,
    each from a discrete Poisson distribution with differing parameter.

    For full information please refer to the NAG Library document for
    g05tk

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tkf.html

    Parameters
    ----------
    m : int
        m, the number of Poisson distributions for which pseudorandom
        variates are required.

    vlamda : float, shape (`m`)
        The means, lambda_j, for j = 1,2,...,`m`, of the Poisson
        distributions.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : int, shape (`m`), modified in place
        `On exit`: the m pseudorandom numbers from the specified Poisson
        distributions.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `m` = *<value>*.

            Constraint: `m` >= 1.

        (`errno` 2)
            On entry, at least one element of `vlamda` is less than
            zero.

        (`errno` 2)
            On entry, at least one element of `vlamda` is too large.

        (`errno` 3)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_poisson_varmean`` generates m integers x_j, each from a
    discrete Poisson distribution with mean lambda_j, where the
    probability of x_j = I is

        P(x_j = I) = lambda_j^I*e^(-lambda_j)/I!, I = 0,1,...,

    where

        lambda_j >= 0, j = 1,2,...,m.

    The methods used by this function have low set up times and are
    designed for efficient use when the value of the parameter lambda
    changes during the simulation.
    For large samples from a distribution with fixed lambda using
    ``int_poisson`` to set up and use a reference vector may be more
    efficient.

    When lambda < 7.5 the product of uniforms method is used, see for
    example Dagpunar (1988).
    For larger values of lambda an envelope rejection method is used
    with a target distribution:

        [table omitted]

    This distribution is generated using a ratio of uniforms method.
    A similar approach has also been suggested by Ahrens and Dieter
    (1989).
    The basic method is combined with quick acceptance and rejection
    tests given by Maclaren (1990).
    For values of lambda >= 87 Stirling's approximation is used in the
    computation of the Poisson distribution function, otherwise tables
    of factorials are used as suggested by Maclaren (1990).

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_poisson_varmean``.

    References
    ----------
    Ahrens, J H and Dieter, U, 1989, `A convenient sampling method with
    bounded computation times for Poisson distributions`, Amer. J. Math.
    Management Sci., 1--13

    Dagpunar, J, 1988, `Principles of Random Variate Generation`, Oxford
    University Press

    Maclaren, N M, 1990, `A Poisson random number generator`, Personal
    Communication
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_poisson_varmean'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tkft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, vlamda[m].
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[m].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _vlamda = utils._EngineFloat64ArrayType(
        vlamda, {'fun_name': fun_name, 'entity': 'vlamda'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_m_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _m,
        _vlamda,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: m >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of vlamda is less than zero.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of vlamda is too large.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def int_uniform(n, a, b, statecomm, x):
    u"""
    Generates a vector of pseudorandom integers from a uniform
    distribution.

    ``int_uniform`` generates a vector of pseudorandom integers
    uniformly distributed over the interval [a,b].

    For full information please refer to the NAG Library document for
    g05tl

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05tlf.html

    Parameters
    ----------
    n : int
        n, the number of pseudorandom numbers to be generated.

    a : int
        The end points a and b of the uniform distribution.

    b : int
        The end points a and b of the uniform distribution.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    x : int, shape (`n`), modified in place
        `On exit`: the n pseudorandom numbers from the specified uniform
        distribution.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, `a` = *<value>* and `b` = *<value>*.

            Constraint: `b` >= `a`.

        (`errno` 4)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``int_uniform`` generates the next n values y_i from a uniform (0,1]
    generator (see ``dist_uniform01`` for details) and applies the
    transformation

        x_i = a+floor((b-a+1)y_i),

    where floor(z) is the integer part of the real value z.
    The function ensures that the values x_i lie in the closed interval
    [a,b].

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``int_uniform``.

    References
    ----------
    Knuth, D E, 1981, `The Art of Computer Programming (Volume 2)`, (2nd
    Edition), Addison--Wesley
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.int_uniform'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05tlft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, a.
        utils._BYREF_INT, # Input, b.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._EngineIntArrayType.c_type, # Output, x[n].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _a = utils._EngineIntScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _b = utils._EngineIntScalarType(
        b, {'fun_name': fun_name, 'entity': 'b'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _x_cls = utils._EngineIntArrayType(
        x, {'fun_name': fun_name, 'entity': 'x'},
        exp_shape=(_n_py,),
        check_contents=False,
    )
    _x = _x_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _a,
        _b,
        _state,
        _x,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s} and b = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: b >= a.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _x_cls.refresh_py(_x)
        return_data = None
    _err_class.manage()
    return return_data

def bb_init(t0, tend, times, ntimes):
    u"""
    Initializes the Brownian bridge generator.

    ``bb_init`` initializes the Brownian bridge generator ``bb``.
    It must be called before any calls to ``bb``.

    For full information please refer to the NAG Library document for
    g05xa

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05xaf.html

    Parameters
    ----------
    t0 : float
        The starting value t_0 of the time interval.

    tend : float
        The end value T of the time interval.

    times : float, shape (`ntimes`)
        The points in the time interval (t_0,T) at which the Wiener
        process is to be constructed. The order in which points are
        listed in `times` determines the bridge construction order. The
        function ``bb_make_bridge_order`` can be used to create
        predefined bridge construction orders from a set of input times.

    ntimes : int
        The length of `times`, denoted by N in [Brownian Bridge
        Algorithm].

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_init``. Please contact NAG with the following error
            message: error in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `tend` = *<value>* and `t0` = *<value>*.

            Constraint: `tend` > `t0`.

        (`errno` 2)
            On entry, `ntimes` = *<value>*.

            Constraint: `ntimes` >= 1.

        (`errno` 3)
            On entry, `times`[*<value>*] = *<value>*, `t0` = *<value>*
            and `tend` = *<value>*.

            Constraint: `t0` < `times`[i-1] < `tend` for all i.

        (`errno` 4)
            On entry, `times`[*<value>*] and `times`[*<value>*] both
            equal *<value>*.

            Constraint: all elements of `times` must be unique.

    Notes
    -----

    Brownian Bridge Algorithm
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    Details on the Brownian bridge algorithm and the Brownian bridge
    process (sometimes also called a non-free Wiener process) can be
    found in [the G05 Introduction].
    We briefly recall some notation and definitions.

    Fix two times t_0 < T and let (t_i)_{1 <= i <= N} be any set of time
    points satisfying t_0 < t_1 < t_2<⋯<t_N < T.
    Let (X_t_i)_{1 <= i <= N} denote a d-dimensional Wiener sample path
    at these time points, and let C be any d*d matrix such that CC^T is
    the desired covariance structure for the Wiener process.
    Each point X_t_i of the sample path is constructed according to the
    Brownian bridge interpolation algorithm (see Glasserman (2004) or
    [the G05 Introduction]).
    We always start at some fixed point X_t_0 = x in R^d.
    If we set X_T = x+Csqrt(T-t_0)Z where Z is any d-dimensional
    standard Normal random variable, then X will behave like a normal
    (free) Wiener process.
    However if we fix the terminal value X_T = w in R^d, then X will
    behave like a non-free Wiener process.

    Implementation
    ~~~~~~~~~~~~~~
    Given the start and end points of the process, the order in which
    successive interpolation times t_j are chosen is called the `bridge
    construction order`.
    The construction order is given by the array `times`.
    Further information on construction orders is given in [the G05
    Introduction].
    For clarity we consider here the common scenario where the Brownian
    bridge algorithm is used with quasi-random points.
    If pseudorandom numbers are used instead, these details can be
    ignored.

    Suppose we require P Wiener sample paths each of dimension d.
    The main input to the Brownian bridge algorithm is then an array of
    quasi-random points Z^1,Z^2,…,Z^P where each point Z^p =
    (Z_1^p,Z_2^p,...,Z_D^p) has dimension D = d(N+1) or D = dN
    respectively, depending on whether a free or non-free Wiener process
    is required.
    When ``bb`` is called, the pth sample path for 1 <= p <= P is
    constructed as follows: if a non-free Wiener process is required set
    X_T equal to the terminal value w, otherwise construct X_T as

        X_T = X_t_0+Csqrt(T-t_0)[Z_1^p;...;Z_d^p]

    where C is the matrix described in [Brownian Bridge Algorithm].
    The array `times` holds the remaining time points t_1,t_2,…t_N in
    the order in which the bridge is to be constructed.
    For each j = 1,…,N set r = `times`[j-1], find

        q = max{t_0,`times`[i-1]:1 <= i < j,`times`[i-1] < r}

    and

        s = min{T,`times`[i-1]:1 <= i < j,`times`[i-1] > r}

    and construct the point X_r as

        X_r =
        (X_q(s-r)+X_s(r-q))/(s-q)+Csqrt((s-r)(r-q)/(s-q))[Z_jd-ad+1^p;...;Z_jd-ad+d^p]

    where a = 0 or a = 1 respectively depending on whether a free or
    non-free Wiener process is required.
    Note that in our discussion j is indexed from 1, and so X_r is
    interpolated between the nearest (in time) Wiener points which have
    already been constructed.
    The function ``bb_make_bridge_order`` can be used to initialize the
    `times` array for several predefined bridge construction orders.

    References
    ----------
    Glasserman, P, 2004, `Monte Carlo Methods in Financial Engineering`,
    Springer
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.bb_init'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05xaft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t0.
        utils._BYREF_FLOAT64, # Input, tend.
        utils._EngineFloat64ArrayType.c_type, # Input, times[ntimes].
        utils._BYREF_INT, # Input, ntimes.
        utils._EngineFloat64ArrayType.c_type, # Output, rcomm[12*(ntimes+1)].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t0 = utils._EngineFloat64ScalarType(
        t0, {'fun_name': fun_name, 'entity': 't0'},
    ).to_ctypes()
    _tend = utils._EngineFloat64ScalarType(
        tend, {'fun_name': fun_name, 'entity': 'tend'},
    ).to_ctypes()
    _ntimes = utils._EngineIntScalarType(
        ntimes, {'fun_name': fun_name, 'entity': 'ntimes'},
    ).to_ctypes()
    _ntimes_py = utils._EngineIntScalarType.to_py(_ntimes)
    _times = utils._EngineFloat64ArrayType(
        times, {'fun_name': fun_name, 'entity': 'times'},
        exp_shape=(_ntimes_py,),
    ).to_ctypes()
    _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(12*(_ntimes_py+1))
    _comm_py = {}
    _comm_py['rcomm'] = _rcomm
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t0,
        _tend,
        _times,
        _ntimes,
        _rcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_p_consts.IERR_UNEXPECTED and _ifmt == _nag_p_consts.IERR_UNEXPECTED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected error occurred during execution of bb_init.',
                    'Please contact NAG with the following error message:',
                    'error in {:s}, ASSERT = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tend = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    't0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: tend > t0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntimes = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ntimes >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 31:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, times[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't0 = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'tend = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: t0 < times[i-1] < tend',
                    'for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, times[{:s}] and times[{:s}]'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'both equal {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: all elements of times must be unique.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def bb(npaths, rcord, d, start, a, term, z, c, b, comm):
    u"""
    Generate paths for a free or non-free Wiener process using the
    Brownian bridge algorithm.

    ``bb`` uses a Brownian bridge algorithm to construct sample paths
    for a free or non-free Wiener process.
    The initialization function ``bb_init`` must be called prior to the
    first call to ``bb``.

    For full information please refer to the NAG Library document for
    g05xb

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05xbf.html

    Parameters
    ----------
    npaths : int
        The number of Wiener sample paths to create.

    rcord : int
        The order in which Normal random numbers are stored in `z` and
        in which the generated values are returned in `b`.

    d : int
        The dimension of each Wiener sample path.

    start : float, shape (`d`)
        The starting value of the Wiener process.

    a : int
        If `a` = 0, a free Wiener process is created beginning at
        `start` and `term` is ignored.

        If `a` = 1, a non-free Wiener process is created beginning at
        `start` and ending at `term`.

    term : float, shape (`d`)
        The terminal value at which the non-free Wiener process should
        end. If `a` = 0, `term` is ignored.

    z : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: `d`*(N+1-`a`); if
        `rcord`=2: `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        `d`*(N+1-`a`); otherwise: 0.

        `On entry`: the Normal random numbers used to construct the
        sample paths.

        If quasi-random numbers are used, the `d`*(N+1-`a`)-dimensional
        quasi-random points should be stored in successive rows of Z.

        `On exit`: the Normal random numbers premultiplied by C.

    c : float, shape (`d`, `d`)
        The lower triangular Cholesky factorization C such that CC^T
        gives the covariance matrix of the Wiener process. Elements of C
        above the diagonal are not referenced.

    b : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: `d`*(N+1); if `rcord`=2:
        `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        `d`*(N+1); otherwise: 0.

        `On exit`: the values of the Wiener sample paths.

        Let X_p,i^k denote the kth dimension of the ith point of the pth
        sample path where 1 <= k <= `d`, 1 <= i <= N+1 and 1 <= p <=
        `npaths`.

        The point X_p,i^k is stored at B(p,k+(i-1)*`d`).

        The starting value `start` is never stored, whereas the terminal
        value is always stored.

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bb_init``.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of ``bb``.
            Please contact NAG with the following error message: error
            in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `comm`['rcomm'] was not initialized or has been
            corrupted.

        (`errno` 2)
            On entry, `npaths` = *<value>*.

            Constraint: `npaths` >= 1.

        (`errno` 3)
            On entry, the value of `rcord` is invalid.

        (`errno` 4)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 1.

        (`errno` 5)
            On entry, `a` = *<value>*.

            Constraint: `a` = 0 or 1.

        (`errno` 6)
            On entry, ldz = *<value>* and `d`*(ntimes+1-`a`) =
            *<value>*.

            Constraint: ldz >= `d`*(ntimes+1-`a`).

    Notes
    -----
    For details on the Brownian bridge algorithm and the bridge
    construction order see [the G05 Introduction] and [Notes for
    bb_init].
    Recall that the terms Wiener process (or free Wiener process) and
    Brownian motion are often used interchangeably, while a non-free
    Wiener process (also known as a Brownian bridge process) refers to a
    process which is forced to terminate at a given point.

    References
    ----------
    Glasserman, P, 2004, `Monte Carlo Methods in Financial Engineering`,
    Springer
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.bb'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05xbft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, npaths.
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, d.
        utils._EngineFloat64ArrayType.c_type, # Input, start[d].
        utils._BYREF_INT, # Input, a.
        utils._EngineFloat64ArrayType.c_type, # Input, term[d].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, z[ldz, *].
        utils._BYREF_INT, # Input, ldz.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, d].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, ntimes.
        utils._EngineFloat64ArrayType.c_type, # Input, rcomm[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (c, 'c'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _npaths = utils._EngineIntScalarType(
        npaths, {'fun_name': fun_name, 'entity': 'npaths'},
    ).to_ctypes()
    _rcord = utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _rcord_py = utils._EngineIntScalarType.to_py(_rcord)
    if _rcord_py == 1:
        _sorder = utils._EngineIntScalarType.c_type(2)
    elif _rcord_py == 2:
        _sorder = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, the value of rcord is invalid.',
            ],
            errcodes=(3,),
        )
        _err_class.manage()
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _start = utils._EngineFloat64ArrayType(
        start, {'fun_name': fun_name, 'entity': 'start'},
        exp_shape=(_d_py,),
    ).to_ctypes()
    _a = utils._EngineIntScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _term = utils._EngineFloat64ArrayType(
        term, {'fun_name': fun_name, 'entity': 'term'},
        exp_shape=(_d_py,),
    ).to_ctypes()
    _ldz_py = utils._EngineFloat64ArrayType.get_nd(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldz = utils._EngineIntScalarType(
        _ldz_py, {'fun_name': fun_name, 'entity': '_ldz'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('rcomm',),
    )
    _vu_n_py = (int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _a_py = utils._EngineIntScalarType.to_py(_a)
    _npaths_py = utils._EngineIntScalarType.to_py(_npaths)
    if _sorder_py == 2:
        _z_dim1 = _d_py*(_vu_n_py+1-_a_py)
    elif _sorder_py == 1:
        _z_dim1 = _npaths_py
    else:
        _z_dim1 = 0
    if _sorder_py == 2:
        _z_dim2 = _npaths_py
    elif _sorder_py == 1:
        _z_dim2 = _d_py*(_vu_n_py+1-_a_py)
    else:
        _z_dim2 = 0
    exp_shape_z = (_z_dim1, _z_dim2)
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=exp_shape_z,
        mutual_sorder=_m_sorder,
    )
    _z = _z_cls.to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_d_py, _d_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _sorder_py == 2:
        _b_dim1 = _d_py*(_vu_n_py+1)
    elif _sorder_py == 1:
        _b_dim1 = _npaths_py
    else:
        _b_dim1 = 0
    if _sorder_py == 2:
        _b_dim2 = _npaths_py
    elif _sorder_py == 1:
        _b_dim2 = _d_py*(_vu_n_py+1)
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _vl_ntimes_py = (int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _ntimes = utils._EngineIntScalarType(
        _vl_ntimes_py, {'fun_name': fun_name, 'entity': '_ntimes'},
    ).to_ctypes()
    _rcomm = comm['rcomm']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _npaths,
        _sorder,
        _d,
        _start,
        _a,
        _term,
        _z,
        _ldz,
        _c,
        _ldc,
        _b,
        _ldb,
        _ntimes,
        _rcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_p_consts.IERR_UNEXPECTED and _ifmt == _nag_p_consts.IERR_UNEXPECTED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected error occurred during execution of bb.',
                    'Please contact NAG with the following error message:',
                    'error in {:s}, ASSERT = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 13:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npaths = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npaths >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 51:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a = 0 or 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 61:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldz = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'd*(ntimes+1-a) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldz >= d*(ntimes+1-a).',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def bb_inc_init(t0, tend, times, ntimes):
    u"""
    Initializes the generator which backs out the increments of sample
    paths generated by a Brownian bridge algorithm.

    ``bb_inc_init`` initializes the Brownian bridge increments generator
    ``bb_inc``.
    It must be called before any calls to ``bb_inc``.

    For full information please refer to the NAG Library document for
    g05xc

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05xcf.html

    Parameters
    ----------
    t0 : float
        The starting value t_0 of the time interval.

    tend : float
        The end value T of the time interval.

    times : float, shape (`ntimes`)
        The points in the time interval (t_0,T) at which the Wiener
        process is to be constructed. The order in which points are
        listed in `times` determines the bridge construction order. The
        function ``bb_make_bridge_order`` can be used to create
        predefined bridge construction orders from a set of input times.

    ntimes : int
        The length of `times`, denoted by N in [Brownian Bridge
        Algorithm].

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_inc_init``. Please contact NAG with the following error
            message: error in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `tend` = *<value>* and `t0` = *<value>*.

            Constraint: `tend` > `t0`.

        (`errno` 2)
            On entry, `ntimes` = *<value>*.

            Constraint: `ntimes` >= 1.

        (`errno` 3)
            On entry, `times`[*<value>*] = *<value>*, `t0` = *<value>*
            and `tend` = *<value>*.

            Constraint: `t0` < `times`[i-1] < `tend` for all i.

        (`errno` 4)
            On entry, `times`[i-1] = `times`[j-1] = *<value>*, for i =
            *<value>* and j = *<value>*.

            Constraint: all elements of `times` must be unique.

    Notes
    -----

    Brownian Bridge Algorithm
    ~~~~~~~~~~~~~~~~~~~~~~~~~
    Details on the Brownian bridge algorithm and the Brownian bridge
    process (sometimes also called a non-free Wiener process) can be
    found in [the G05 Introduction].
    We briefly recall some notation and definitions.

    Fix two times t_0 < T and let (t_i)_{1 <= i <= N} be any set of time
    points satisfying t_0 < t_1 < t_2<⋯<t_N < T.
    Let (X_t_i)_{1 <= i <= N} denote a d-dimensional Wiener sample path
    at these time points, and let C be any d*d matrix such that CC^T is
    the desired covariance structure for the Wiener process.
    Each point X_t_i of the sample path is constructed according to the
    Brownian bridge interpolation algorithm (see Glasserman (2004) or
    [the G05 Introduction]).
    We always start at some fixed point X_t_0 = x in R^d.
    If we set X_T = x+Csqrt(T-t_0)Z where Z is any d-dimensional
    standard Normal random variable, then X will behave like a normal
    (free) Wiener process.
    However if we fix the terminal value X_T = w in R^d, then X will
    behave like a non-free Wiener process.

    The Brownian bridge increments generator uses the Brownian bridge
    algorithm to construct sample paths for the (free or non-free)
    Wiener process X, and then uses this to compute the `scaled Wiener
    increments`

(X_t_1-X_t_0)/(t_1-t_0),(X_t_2-X_t_1)/(t_2-t_1),...,(X_t_N-X_t_{N-1})/(t_N-t_{N-1}),(X_T-X_t_N)/(T-t_N).

    Such increments can be useful in computing numerical solutions to
    stochastic differential equations driven by (free or non-free)
    Wiener processes.

    Implementation
    ~~~~~~~~~~~~~~
    Conceptually, the output of the Wiener increments generator is the
    same as if ``bb_init`` and ``bb`` were called first, and the scaled
    increments then constructed from their output.
    The implementation adopts a much more efficient approach whereby the
    scaled increments are computed directly without first constructing
    the Wiener sample path.

    Given the start and end points of the process, the order in which
    successive interpolation times t_j are chosen is called the `bridge
    construction order`.
    The construction order is given by the array `times`.
    Further information on construction orders is given in [the G05
    Introduction].
    For clarity we consider here the common scenario where the Brownian
    bridge algorithm is used with quasi-random points.
    If pseudorandom numbers are used instead, these details can be
    ignored.

    Suppose we require the increments of P Wiener sample paths each of
    dimension d.
    The main input to the Brownian bridge increments generator is then
    an array of quasi-random points Z^1,Z^2,…,Z^P where each point Z^p =
    (Z_1^p,Z_2^p,...,Z_D^p) has dimension D = d(N+1) or D = dN depending
    on whether a free or non-free Wiener process is required.
    When ``bb_inc`` is called, the pth sample path for 1 <= p <= P is
    constructed as follows: if a non-free Wiener process is required set
    X_T equal to the terminal value w, otherwise construct X_T as

        X_T = X_t_0+Csqrt(T-t_0)[Z_1^p;...;Z_d^p]

    where C is the matrix described in [Brownian Bridge Algorithm].
    The array `times` holds the remaining time points t_1,t_2,…t_N in
    the order in which the bridge is to be constructed.
    For each j = 1,…,N set r = `times`[j-1], find

        q = max{t_0,`times`[i-1]:1 <= i < j,`times`[i-1] < r}

    and

        s = min{T,`times`[i-1]:1 <= i < j,`times`[i-1] > r}

    and construct the point X_r as

        X_r =
        (X_q(s-r)+X_s(r-q))/(s-q)+Csqrt((s-r)(r-q)/(s-q))[Z_jd-ad+1^p;...;Z_jd-ad+d^p]

    where a = 0 or a = 1 depending on whether a free or non-free Wiener
    process is required.
    The function ``bb_make_bridge_order`` can be used to initialize the
    `times` array for several predefined bridge construction orders.
    Lastly, the scaled Wiener increments

(X_t_1-X_t_0)/(t_1-t_0),(X_t_2-X_t_1)/(t_2-t_1),…,(X_t_N-X_t_{N-1})/(t_N-t_{N-1}),(X_T-X_t_N)/(T-t_N)

    are computed.

    References
    ----------
    Glasserman, P, 2004, `Monte Carlo Methods in Financial Engineering`,
    Springer

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.bb_inc_init'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05xcft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, t0.
        utils._BYREF_FLOAT64, # Input, tend.
        utils._EngineFloat64ArrayType.c_type, # Input, times[ntimes].
        utils._BYREF_INT, # Input, ntimes.
        utils._EngineFloat64ArrayType.c_type, # Output, rcomm[12*(ntimes+1)].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _t0 = utils._EngineFloat64ScalarType(
        t0, {'fun_name': fun_name, 'entity': 't0'},
    ).to_ctypes()
    _tend = utils._EngineFloat64ScalarType(
        tend, {'fun_name': fun_name, 'entity': 'tend'},
    ).to_ctypes()
    _ntimes = utils._EngineIntScalarType(
        ntimes, {'fun_name': fun_name, 'entity': 'ntimes'},
    ).to_ctypes()
    _ntimes_py = utils._EngineIntScalarType.to_py(_ntimes)
    _times = utils._EngineFloat64ArrayType(
        times, {'fun_name': fun_name, 'entity': 'times'},
        exp_shape=(_ntimes_py,),
    ).to_ctypes()
    _rcomm = utils._EngineFloat64ArrayType.empty_ctypes(12*(_ntimes_py+1))
    _comm_py = {}
    _comm_py['rcomm'] = _rcomm
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _t0,
        _tend,
        _times,
        _ntimes,
        _rcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_p_consts.IERR_UNEXPECTED and _ifmt == _nag_p_consts.IERR_UNEXPECTED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected error occurred during execution of bb_inc_init.',
                    'Please contact NAG with the following error message:',
                    'error in {:s}, ASSERT = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, tend = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    't0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: tend > t0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntimes = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ntimes >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 31:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, times[{:s}] = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    't0 = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'tend = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: t0 < times[i-1] < tend',
                    'for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, times[i-1] = times[j-1] = {:s},'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                    ),
                    'for i = {:s} and j = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[2]),
                        utils._STR_FMT_STR.format(_eb_data[3]),
                    ),
                    'Constraint: all elements of times must be unique.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def bb_inc(npaths, rcord, d, a, diff, z, c, b, comm):
    u"""
    Backs out the increments from sample paths generated by a Brownian
    bridge algorithm.

    ``bb_inc`` computes scaled increments of sample paths of a free or
    non-free Wiener process, where the sample paths are constructed by a
    Brownian bridge algorithm.
    The initialization function ``bb_inc_init`` must be called prior to
    the first call to ``bb_inc``.

    For full information please refer to the NAG Library document for
    g05xd

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05xdf.html

    Parameters
    ----------
    npaths : int
        The number of Wiener sample paths.

    rcord : int
        The order in which Normal random numbers are stored in `z` and
        in which the generated values are returned in `b`.

    d : int
        The dimension of each Wiener sample path.

    a : int
        If `a` = 0, a free Wiener process is created and `diff` is
        ignored.

        If `a` = 1, a non-free Wiener process is created where `diff` is
        the difference between the terminal value and the starting value
        of the process.

    diff : float, shape (`d`)
        The difference between the terminal value and starting value of
        the Wiener process. If `a` = 0, `diff` is ignored.

    z : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: `d`*(N+1-`a`); if
        `rcord`=2: `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        `d`*(N+1-`a`); otherwise: 0.

        `On entry`: the Normal random numbers used to construct the
        sample paths.

        If quasi-random numbers are used, the `d`*(N+1-`a`)-dimensional
        quasi-random points should be stored in successive rows of Z.

        `On exit`: the Normal random numbers premultiplied by `c`.

    c : float, shape (`d`, `d`)
        The lower triangular Cholesky factorization C such that CC^T
        gives the covariance matrix of the Wiener process. Elements of C
        above the diagonal are not referenced.

    b : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: `d`*(N+1); if `rcord`=2:
        `npaths`; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `npaths`; if `rcord`=2:
        `d`*(N+1); otherwise: 0.

        `On exit`: the scaled Wiener increments.

        Let X_p,i^k denote the kth dimension of the ith point of the pth
        sample path where 1 <= k <= `d`, 1 <= i <= N+1 and 1 <= p <=
        `npaths`.

        The increment ((X_p,i^k-X_p,i-1^k))/((t_i-t_{i-1})) is stored at
        B(p,k+(i-1)*`d`).

    comm : dict, communication object
        Communication structure.

        This argument must have been initialized by a prior call to
        ``bb_inc_init``.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of ``bb_inc``.
            Please contact NAG with the following error message: error
            in *<value>*, ASSERT = *<value>*.

        (`errno` 1)
            On entry, `comm`['rcomm'] was not initialized or has been
            corrupted.

        (`errno` 2)
            On entry, `npaths` = *<value>*.

            Constraint: `npaths` >= 1.

        (`errno` 3)
            On entry, the value of `rcord` is invalid.

        (`errno` 4)
            On entry, `d` = *<value>*.

            Constraint: `d` >= 1.

        (`errno` 5)
            On entry, `a` = *<value>*.

            Constraint: `a` = 0 or 1.

        (`errno` 6)
            On entry, ldz = *<value>* and `d`*(ntimes+1-`a`) =
            *<value>*.

            Constraint: ldz >= `d`*(ntimes+1-`a`).

    Notes
    -----
    For details on the Brownian bridge algorithm and the bridge
    construction order see [the G05 Introduction] and [Notes for
    bb_inc_init].
    Recall that the terms Wiener process (or free Wiener process) and
    Brownian motion are often used interchangeably, while a non-free
    Wiener process (also known as a Brownian bridge process) refers to a
    process which is forced to terminate at a given point.

    Fix two times t_0 < T, let (t_i)_{1 <= i <= N} be any set of time
    points satisfying t_0 < t_1 < t_2 < ... < t_N < T, and let X_t_0,
    (X_t_i)_{1 <= i <= N}, X_T denote a d-dimensional Wiener sample path
    at these time points.

    The Brownian bridge increments generator uses the Brownian bridge
    algorithm to construct sample paths for the (free or non-free)
    Wiener process X, and then uses this to compute the `scaled Wiener
    increments`

(X_t_1-X_t_0)/(t_1-t_0),(X_t_2-X_t_1)/(t_2-t_1),...,(X_t_N-X_t_{N-1})/(t_N-t_{N-1}),(X_T-X_t_N)/(T-t_N)

    References
    ----------
    Glasserman, P, 2004, `Monte Carlo Methods in Financial Engineering`,
    Springer

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.bb_inc'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05xdft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, npaths.
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, d.
        utils._BYREF_INT, # Input, a.
        utils._EngineFloat64ArrayType.c_type, # Input, diff[d].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, z[ldz, *].
        utils._BYREF_INT, # Input, ldz.
        utils._EngineFloat64ArrayType.c_type, # Input, c[ldc, d].
        utils._BYREF_INT, # Input, ldc.
        utils._EngineFloat64ArrayType.c_type, # Output, b[ldb, *].
        utils._BYREF_INT, # Input, ldb.
        utils._BYREF_INT, # Input, ntimes.
        utils._EngineFloat64ArrayType.c_type, # Input, rcomm[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'), (c, 'c'), (b, 'b')),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _npaths = utils._EngineIntScalarType(
        npaths, {'fun_name': fun_name, 'entity': 'npaths'},
    ).to_ctypes()
    _rcord = utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _rcord_py = utils._EngineIntScalarType.to_py(_rcord)
    if _rcord_py == 1:
        _sorder = utils._EngineIntScalarType.c_type(2)
    elif _rcord_py == 2:
        _sorder = utils._EngineIntScalarType.c_type(1)
    else:
        _err_class = utils._EngineErrorExit(
            fun_name=fun_name,
            msg_lines=[
                'On entry, the value of rcord is invalid.',
            ],
            errcodes=(3,),
        )
        _err_class.manage()
    _d = utils._EngineIntScalarType(
        d, {'fun_name': fun_name, 'entity': 'd'},
    ).to_ctypes()
    _a = utils._EngineIntScalarType(
        a, {'fun_name': fun_name, 'entity': 'a'},
    ).to_ctypes()
    _d_py = utils._EngineIntScalarType.to_py(_d)
    _diff = utils._EngineFloat64ArrayType(
        diff, {'fun_name': fun_name, 'entity': 'diff'},
        exp_shape=(_d_py,),
    ).to_ctypes()
    _ldz_py = utils._EngineFloat64ArrayType.get_nd(z, {'fun_name': fun_name, 'entity': 'z'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldz = utils._EngineIntScalarType(
        _ldz_py, {'fun_name': fun_name, 'entity': '_ldz'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('rcomm',),
    )
    _vu_n_py = (int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _a_py = utils._EngineIntScalarType.to_py(_a)
    _npaths_py = utils._EngineIntScalarType.to_py(_npaths)
    if _sorder_py == 2:
        _z_dim1 = _d_py*(_vu_n_py+1-_a_py)
    elif _sorder_py == 1:
        _z_dim1 = _npaths_py
    else:
        _z_dim1 = 0
    if _sorder_py == 2:
        _z_dim2 = _npaths_py
    elif _sorder_py == 1:
        _z_dim2 = _d_py*(_vu_n_py+1-_a_py)
    else:
        _z_dim2 = 0
    exp_shape_z = (_z_dim1, _z_dim2)
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=exp_shape_z,
        mutual_sorder=_m_sorder,
    )
    _z = _z_cls.to_ctypes()
    _ldc_py = utils._EngineFloat64ArrayType.get_nd(c, {'fun_name': fun_name, 'entity': 'c'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldc = utils._EngineIntScalarType(
        _ldc_py, {'fun_name': fun_name, 'entity': '_ldc'},
    ).to_ctypes()
    _c = utils._EngineFloat64ArrayType(
        c, {'fun_name': fun_name, 'entity': 'c'},
        exp_shape=(_d_py, _d_py),
        mutual_sorder=_m_sorder,
    ).to_ctypes()
    _ldb_py = utils._EngineFloat64ArrayType.get_nd(b, {'fun_name': fun_name, 'entity': 'b'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldb = utils._EngineIntScalarType(
        _ldb_py, {'fun_name': fun_name, 'entity': '_ldb'},
    ).to_ctypes()
    if _sorder_py == 2:
        _b_dim1 = _d_py*(_vu_n_py+1)
    elif _sorder_py == 1:
        _b_dim1 = _npaths_py
    else:
        _b_dim1 = 0
    if _sorder_py == 2:
        _b_dim2 = _npaths_py
    elif _sorder_py == 1:
        _b_dim2 = _d_py*(_vu_n_py+1)
    else:
        _b_dim2 = 0
    exp_shape_b = (_b_dim1, _b_dim2)
    _b_cls = utils._EngineFloat64ArrayType(
        b, {'fun_name': fun_name, 'entity': 'b'},
        exp_shape=exp_shape_b,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _b = _b_cls.to_ctypes(can_reference_contents=False)
    _vl_ntimes_py = (int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)+0.5) if utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False) > 0 else int(utils._EngineFloat64ArrayType.get_element(comm['rcomm'], {'fun_name': fun_name, 'entity': 'comm[\'rcomm\']'}, (1,), validate=False)-0.5))
    _ntimes = utils._EngineIntScalarType(
        _vl_ntimes_py, {'fun_name': fun_name, 'entity': '_ntimes'},
    ).to_ctypes()
    _rcomm = comm['rcomm']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _npaths,
        _sorder,
        _d,
        _a,
        _diff,
        _z,
        _ldz,
        _c,
        _ldc,
        _b,
        _ldb,
        _ntimes,
        _rcomm,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_p_consts.IERR_UNEXPECTED and _ifmt == _nag_p_consts.IERR_UNEXPECTED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected error occurred during execution of bb_inc.',
                    'Please contact NAG with the following error message:',
                    'error in {:s}, ASSERT = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 12:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 13:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'rcomm\'] was not initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, npaths = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: npaths >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, d = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: d >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 51:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, a = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: a = 0 or 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 61:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldz = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'd*(ntimes+1-a) = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldz >= d*(ntimes+1-a).',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        _b_cls.refresh_py(_b)
        return_data = None
    _err_class.manage()
    return return_data

def bb_make_bridge_order(bgord, t0, tend, ntimes, intime, nmove, move, times):
    u"""
    Creates a Brownian bridge construction order out of a set of input
    times.

    ``bb_make_bridge_order`` takes a set of input times and permutes
    them to specify one of several predefined Brownian bridge
    construction orders.
    The permuted times can be passed to ``bb_init`` or ``bb_inc_init``
    to initialize the Brownian bridge generators with the chosen bridge
    construction order.

    For full information please refer to the NAG Library document for
    g05xe

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05xef.html

    Parameters
    ----------
    bgord : int
        The bridge construction order to use.

    t0 : float
        t_0, the start value of the time interval on which the Wiener
        process is to be constructed.

    tend : float
        T, the largest time at which the Wiener process is to be
        constructed.

    ntimes : int
        N, the number of time points in the Wiener process, excluding
        t_0 and T.

    intime : float, shape (`ntimes`)
        The time points, t_1,t_2,...,t_N, at which the Wiener process is
        to be constructed. Note that the final time T is not included in
        this array.

    nmove : int
        The number of elements in the array `move`.

    move : None or int, shape (`nmove`)
        The indices of the entries in `intime` which should be moved to
        the front of the `times` array, with `move`[j-1] = i setting the
        jth element of `times` to t_i. Note that i ranges from 1 to
        `ntimes`. When `nmove` = 0, `move` is not referenced.

    times : float, shape (`ntimes`), modified in place
        `On exit`: the output bridge construction order. This should be
        passed to ``bb_init`` or ``bb_inc_init``.

    Raises
    ------
    NagValueError
        (`errno` -99)
            An unexpected error occurred during execution of
            ``bb_make_bridge_order``. Please contact NAG with the
            following error message: error in *<value>*, ASSERT =
            *<value>*.

        (`errno` 1)
            On entry, `bgord` = *<value>*.

            Constraint: `bgord` = 1, 2, 3 or 4

        (`errno` 2)
            On entry, `ntimes` = *<value>*.

            Constraint: `ntimes` >= 1.

        (`errno` 3)
            On entry, `nmove` = *<value>* and `ntimes` = *<value>*.

            Constraint: 0 <= `nmove` <= `ntimes`.

        (`errno` 4)
            On entry, `intime`[0] = *<value>* and `t0` = *<value>*.

            Constraint: `intime`[0] > `t0`.

        (`errno` 4)
            On entry, `intime`[*<value>*] = *<value>* and
            `intime`[*<value>*] = *<value>*.

            Constraint: the elements in `intime` must be in increasing
            order.

        (`errno` 4)
            On entry, `ntimes` = *<value>*, `intime`[`ntimes`-1] =
            *<value>* and `tend` = *<value>*.

            Constraint: `intime`[`ntimes`-1] < `tend`.

        (`errno` 5)
            On entry, `move`[*<value>*] = *<value>*.

            Constraint: `move`[i] >= 1 for all i.

        (`errno` 5)
            On entry, `move`[*<value>*] = *<value>* and `ntimes` =
            *<value>*.

            Constraint: `move`[i] <= `ntimes` for all i.

        (`errno` 6)
            On entry, `move`[*<value>*] and `move`[*<value>*] both equal
            *<value>*.

            Constraint: all elements in `move` must be unique.

    Notes
    -----
    The Brownian bridge algorithm (see Glasserman (2004)) is a popular
    method for constructing a Wiener process at a set of discrete times,
    t_0 < t_1 < t_2 <,..., < t_N < T, for N >= 1.
    To ease notation we assume that T has the index N+1 so that T =
    t_{N+1}.
    Inherent in the algorithm is the notion of a `bridge construction
    order` which specifies the order in which the N+2 points of the
    Wiener process, X_t_0,X_T and X_t_i, for i = 1,2,...,N, are
    generated.
    The value of X_t_0 is always assumed known, and the first point to
    be generated is always the final time X_T.
    Thereafter, successive points are generated iteratively by an
    interpolation formula, using points which were computed at previous
    iterations.
    In many cases the bridge construction order is not important, since
    any construction order will yield a correct process.
    However, in certain cases, for example when using quasi-random
    variates to construct the sample paths, the bridge construction
    order can be important.

    Supported Bridge Construction Orders
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ``bb_make_bridge_order`` accepts as input an array of time points
    t_1,t_2,...,t_N,T at which the Wiener process is to be sampled.
    These time points are then permuted to construct the bridge.
    In all of the supported construction orders the first construction
    point is T which has index N+1.
    The remaining points are constructed by iteratively bisecting
    (sub-intervals of) the `time indices` interval [0,N+1], as Figure
    [label omitted] illustrates:

    [figure omitted]

    The time indices interval is processed in levels L^i, for i =
    1,2,...,.
    Each level L^i contains n_i points L_1^i,...,L_n_i^i where n_i <=
    2^(i-1).
    The number of points at each level depends on the value of N.
    The points L_j^i for i >= 1 and j = 1,2,...n_i are computed as
    follows: define L_0^0 = N+1 and set

        [table omitted]

    By convention the maximum of the empty set is taken to be to be
    zero. Figure [label omitted] illustrates the algorithm when N+1 is a
    power of two.
    When N+1 is not a power of two, one must decide how to round the
    divisions by 2.
    For example, if one rounds down to the nearest integer, then one
    could get the following:

    [figure omitted]

    From the series of bisections outlined above, two ways of ordering
    the time indices L_j^i are supported.
    In both cases, levels are always processed from coarsest to finest
    (i.e., increasing i).
    Within a level, the time indices can either be processed left to
    right (i.e., increasing j) or right to left (i.e., decreasing j).
    For example, when processing left to right, the sequence of time
    indices could be generated as:

        N+1 L_1^1 L_1^2 L_2^2 L_1^3 L_2^3 L_3^3 L_4^3 ...

    while when processing right to left, the same sequence would be
    generated as:

        N+1 L_1^1 L_2^2 L_1^2 L_4^3 L_3^3 L_2^3 L_1^3 ...

    ``bb_make_bridge_order``, therefore, offers four bridge construction
    methods; processing either left to right or right to left, with
    rounding either up or down.
    Which method is used is controlled by the `bgord` argument.
    For example, on the set of times

        t_1 t_2 t_3 t_4 t_5 t_6 t_7 t_8 t_9 t_10 t_11 t_12 T

    the Brownian bridge would be constructed in the following orders:

        `bgord` = 1 (processing left to right, rounding down)

                T t_6 t_3 t_9 t_1 t_4 t_7 t_11 t_2 t_5 t_8 t_10 t_12

        `bgord` = 2 (processing left to right, rounding up)

                T t_7 t_4 t_10 t_2 t_6 t_9 t_12 t_1 t_3 t_5 t_8 t_11

        `bgord` = 3 (processing right to left, rounding down)

                T t_6 t_9 t_3 t_11 t_7 t_4 t_1 t_12 t_10 t_8 t_5 t_2

        `bgord` = 4 (processing right to left, rounding up)

                T t_7 t_10 t_4 t_12 t_9 t_6 t_2 t_11 t_8 t_5 t_3 t_1.

    The four construction methods described above can be further
    modified through the use of the input array `move`.
    To see the effect of this argument, suppose that an array A holds
    the output of ``bb_make_bridge_order`` when `nmove` = 0 (i.e., the
    bridge construction order as specified by `bgord` only).
    Let

        B = {t_j:j = `move`[i-1],i = 1,2,...,`nmove`}

    be the array of all times identified by `move`, and let C be the
    array A with all the elements in B removed, i.e.,

        C = {A(i):A(i) != B(j),i = 1,2,...,`ntimes`,j =
        1,2,...,`nmove`}.

    Then the output of ``bb_make_bridge_order`` when `nmove` > 0 is
    given by

        B(1) B(2) ... B(`nmove`) C(1) C(2) ... C(`ntimes`-`nmove`)

    When the Brownian bridge is used with quasi-random variates, this
    functionality can be used to allow specific sections of the bridge
    to be constructed using the lowest dimensions of the quasi-random
    points.

    References
    ----------
    Glasserman, P, 2004, `Monte Carlo Methods in Financial Engineering`,
    Springer

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.bb_make_bridge_order'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05xeft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, bgord.
        utils._BYREF_FLOAT64, # Input, t0.
        utils._BYREF_FLOAT64, # Input, tend.
        utils._BYREF_INT, # Input, ntimes.
        utils._EngineFloat64ArrayType.c_type, # Input, intime[ntimes].
        utils._BYREF_INT, # Input, nmove.
        utils._EngineIntArrayType.c_type, # Input, move[nmove].
        utils._EngineFloat64ArrayType.c_type, # Output, times[ntimes].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _bgord = utils._EngineIntScalarType(
        bgord, {'fun_name': fun_name, 'entity': 'bgord'},
    ).to_ctypes()
    _t0 = utils._EngineFloat64ScalarType(
        t0, {'fun_name': fun_name, 'entity': 't0'},
    ).to_ctypes()
    _tend = utils._EngineFloat64ScalarType(
        tend, {'fun_name': fun_name, 'entity': 'tend'},
    ).to_ctypes()
    _ntimes = utils._EngineIntScalarType(
        ntimes, {'fun_name': fun_name, 'entity': 'ntimes'},
    ).to_ctypes()
    _ntimes_py = utils._EngineIntScalarType.to_py(_ntimes)
    _intime = utils._EngineFloat64ArrayType(
        intime, {'fun_name': fun_name, 'entity': 'intime'},
        exp_shape=(_ntimes_py,),
    ).to_ctypes()
    _nmove = utils._EngineIntScalarType(
        nmove, {'fun_name': fun_name, 'entity': 'nmove'},
    ).to_ctypes()
    _nmove_py = utils._EngineIntScalarType.to_py(_nmove)
    if _nmove_py != 0:
        if move is None:
            raise utils.NagTypeError(
                {'fun_name': fun_name, 'entity': 'move'},
                'must not be None when nmove != 0',
            )
    _move = utils._EngineIntArrayType(
        move, {'fun_name': fun_name, 'entity': 'move'},
        exp_shape=(_nmove_py,),
        check_contents=_nmove_py != 0,
        allow_none=True,
    ).to_ctypes(can_reference_contents=_nmove_py != 0)
    _times_cls = utils._EngineFloat64ArrayType(
        times, {'fun_name': fun_name, 'entity': 'times'},
        exp_shape=(_ntimes_py,),
        check_contents=False,
    )
    _times = _times_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _bgord,
        _t0,
        _tend,
        _ntimes,
        _intime,
        _nmove,
        _move,
        _times,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == _nag_p_consts.IERR_UNEXPECTED and _ifmt == _nag_p_consts.IERR_UNEXPECTED:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'An unexpected error occurred during execution of bb_make_bridge_order.',
                    'Please contact NAG with the following error message:',
                    'error in {:s}, ASSERT = {:s}.'.format(
                        utils._STR_FMT_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 11:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, bgord = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: bgord = 1, 2, 3 or 4',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntimes = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ntimes >= 1.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 31:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, nmove = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ntimes = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 0 <= nmove <= ntimes.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 41:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, intime[0] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    't0 = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: intime[0] > t0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 42:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, intime[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'intime[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraint: the elements in intime must be in increasing order.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 43:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ntimes = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'intime[ntimes-1] = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'tend = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: intime[ntimes-1] < tend.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 51:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, move[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: move[i] >= 1 for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 52:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, move[{:s}] = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ntimes = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'Constraint: move[i] <= ntimes for all i.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 61:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, move[{:s}] and move[{:s}]'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                    'both equal {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: all elements in move must be unique.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _times_cls.refresh_py(_times)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_normal(xmean, std, n, quas, comm):
    u"""
    Generates a Normal quasi-random number sequence.

    ``quasi_normal`` generates a quasi-random sequence from a Normal
    (Gaussian) distribution.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yj

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05yjf.html

    Parameters
    ----------
    xmean : float, shape (idim)
        Specifies, for each dimension, the mean of the Normal
        distribution.

    std : float, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        Normal distribution.

    n : int
        The number of quasi-random numbers required.

    quas : float, shape (`n`, idim), modified in place
        `On exit`: contains the `n` quasi-random numbers of dimension
        idim

        If sorder = 2, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If sorder = 1, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.0.

        (`errno` 4)
            There have been too many calls to the generator.

    Notes
    -----
    ``quasi_normal`` generates a quasi-random sequence from a Normal
    distribution by first generating a uniform quasi-random sequence
    which is then transformed into a Normal sequence using the inverse
    of the Normal CDF.
    The type of uniform sequence used depends on the initialization
    function called and can include the low-discrepancy sequences
    proposed by Sobol, Faure or Niederreiter.
    If the initialization function ``quasi_init_scrambled`` was used
    then the underlying uniform sequence is first scrambled prior to
    being transformed (see [Notes for quasi_init_scrambled] for
    details).

    References
    ----------
    Bratley, P and Fox, B L, 1988, `Algorithm 659: implementing Sobol's
    quasirandom sequence generator`, ACM Trans. Math. Software (14(1)),
    88--100

    Fox, B L, 1986, `Algorithm 647: implementation and relative
    efficiency of quasirandom sequence generators`, ACM Trans. Math.
    Software (12(4)), 362--376

    Wichura, 1988, `Algorithm AS 241: the percentage points of the
    Normal distribution`, Appl. Statist. (37), 477--484

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_normal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05yjft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._EngineFloat64ArrayType.c_type, # Input, xmean[idim].
        utils._EngineFloat64ArrayType.c_type, # Input, std[idim].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._BYREF_INT, # Input, idim.
        utils._BYREF_INT, # Input, ldquas.
        utils._BYREF_INT, # Input, tdquas.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _idim_py = utils._EngineFloat64ArrayType.get_shape(xmean, {'fun_name': fun_name, 'entity': 'xmean'}, exp_rank=1)[0]
    _idim = utils._EngineIntScalarType(
        _idim_py, {'fun_name': fun_name, 'entity': '_idim'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _xmean = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _std = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=(_n_py, _vl_idim_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _xmean,
        _std,
        _n,
        _quas,
        _iref,
        _liref,
        _idim,
        _ldquas,
        _tdquas,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 111 and _ifmt == 111:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and std[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: std[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There have been too many calls to the generator.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_lognormal(xmean, std, n, quas, comm):
    u"""
    Generates a log-normal quasi-random number sequence.

    ``quasi_lognormal`` generates a quasi-random sequence from a
    log-normal distribution.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yk

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ykf.html

    Parameters
    ----------
    xmean : float, shape (idim)
        Specifies, for each dimension, the mean of the underlying Normal
        distribution.

    std : float, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        underlying Normal distribution.

    n : int
        The number of quasi-random numbers required.

    quas : float, shape (`n`, idim), modified in place
        `On exit`: contains the `n` quasi-random numbers of dimension
        idim.

        If sorder = 2, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If sorder = 1, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 2)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 3)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.

        (`errno` 4)
            There have been too many calls to the generator.

        (`errno` 5)
            On entry, i = *<value>* and `xmean`[i-1] = *<value>*.

            Constraint: abs(`xmean`[i-1]) <= *<value>*.

    Notes
    -----
    ``quasi_lognormal`` generates a quasi-random sequence from a
    log-normal distribution by first generating a uniform quasi-random
    sequence which is then transformed into a log-normal sequence using
    the exponential of the inverse of the Normal CDF.
    The type of uniform sequence used depends on the initialization
    function called and can include the low-discrepancy sequences
    proposed by Sobol, Faure or Niederreiter.
    If the initialization function ``quasi_init_scrambled`` was used
    then the underlying uniform sequence is first scrambled prior to
    being transformed (see [Notes for quasi_init_scrambled] for
    details).

    References
    ----------
    Bratley, P and Fox, B L, 1988, `Algorithm 659: implementing Sobol's
    quasirandom sequence generator`, ACM Trans. Math. Software (14(1)),
    88--100

    Fox, B L, 1986, `Algorithm 647: implementation and relative
    efficiency of quasirandom sequence generators`, ACM Trans. Math.
    Software (12(4)), 362--376

    Wichura, 1988, `Algorithm AS 241: the percentage points of the
    Normal distribution`, Appl. Statist. (37), 477--484
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_lognormal'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ykft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._EngineFloat64ArrayType.c_type, # Input, xmean[idim].
        utils._EngineFloat64ArrayType.c_type, # Input, std[idim].
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._BYREF_INT, # Input, idim.
        utils._BYREF_INT, # Input, ldquas.
        utils._BYREF_INT, # Input, tdquas.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType.c_type(_nag_enum_consts.nag_col_major)
    _idim_py = utils._EngineFloat64ArrayType.get_shape(xmean, {'fun_name': fun_name, 'entity': 'xmean'}, exp_rank=1)[0]
    _idim = utils._EngineIntScalarType(
        _idim_py, {'fun_name': fun_name, 'entity': '_idim'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _xmean = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _std = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=(_n_py, _vl_idim_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _xmean,
        _std,
        _n,
        _quas,
        _iref,
        _liref,
        _idim,
        _ldquas,
        _tdquas,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 111 and _ifmt == 111:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and std[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: std[i-1] >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'There have been too many calls to the generator.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and xmean[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(xmean[i-1]) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_init(genid, idim, iskip):
    u"""
    Initializes a quasi-random number generator.

    ``quasi_init`` initializes a quasi-random generator prior to calling
    ``quasi_uniform``, ``quasi_normal`` or ``quasi_lognormal``.

    For full information please refer to the NAG Library document for
    g05yl

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ylf.html

    Parameters
    ----------
    genid : int
        Must identify the quasi-random generator to use.

        `genid` = 1

            Sobol generator.

        `genid` = 2

            Sobol (A659) generator.

        `genid` = 3

            Niederreiter generator.

        `genid` = 4

            Faure generator.

    idim : int
        The number of dimensions required.

    iskip : int
        The number of terms of the sequence to skip on initialization
        for the Sobol and Niederreiter generators. If `genid` = 4,
        `iskip` is ignored.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: `genid` = 1, 2, 3 or 4.

        (`errno` 2)
            On entry, `idim` = *<value>*.

            Constraint: 1 <= `idim` <= *<value>*.

        (`errno` 5)
            On entry, `iskip` < 0 or `iskip` is too large: `iskip` =
            *<value>*, maximum value is *<value>*.

    Notes
    -----
    ``quasi_init`` selects a quasi-random number generator through the
    input value of `genid` and initializes the `comm`['iref']
    communication array for use by the functions ``quasi_uniform``,
    ``quasi_normal`` or ``quasi_lognormal``.

    One of three types of quasi-random generator may be chosen, allowing
    the low-discrepancy sequences proposed by Sobol, Faure or
    Niederreiter to be generated.

    Two sets of Sobol sequences are supplied, the first, is based on the
    work of Joe and Kuo (2008).
    The second, referred to in the documentation as 'Sobol (A659)', is
    based on Algorithm 659 of Bratley and Fox (1988) with the extension
    to 1111 dimensions proposed by Joe and Kuo (2003).
    Both sets of Sobol sequences should satisfy the so-called Property
    A, up to 1111 dimensions, but the first set should have better
    two-dimensional projections than those produced using Algorithm 659.

    References
    ----------
    Bratley, P and Fox, B L, 1988, `Algorithm 659: implementing Sobol's
    quasirandom sequence generator`, ACM Trans. Math. Software (14(1)),
    88--100

    Fox, B L, 1986, `Algorithm 647: implementation and relative
    efficiency of quasirandom sequence generators`, ACM Trans. Math.
    Software (12(4)), 362--376

    Joe, S and Kuo, F Y, 2003, `Remark on Algorithm 659: implementing
    Sobol's quasirandom sequence generator`, ACM Trans. Math. Software
    (TOMS) (29), 49--57

    Joe, S and Kuo, F Y, 2008, `Constructing Sobol sequences with better
    two-dimensional projections`, SIAM J. Sci. Comput. (30), 2635--2654
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_init'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ylft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, genid.
        utils._BYREF_INT, # Input, idim.
        utils._EngineIntArrayType.c_type, # Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._BYREF_INT, # Input, iskip.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _genid = utils._EngineIntScalarType(
        genid, {'fun_name': fun_name, 'entity': 'genid'},
    ).to_ctypes()
    _idim = utils._EngineIntScalarType(
        idim, {'fun_name': fun_name, 'entity': 'idim'},
    ).to_ctypes()
    _genid_py = utils._EngineIntScalarType.to_py(_genid)
    _idim_py = utils._EngineIntScalarType.to_py(_idim)
    if _genid_py in (1, 2, 3):
        _liref = utils._EngineIntScalarType(
            33*_idim_py+7, {'fun_name': fun_name, 'entity': '_liref'},
        ).to_ctypes()
    elif _genid_py == 4:
        _liref = utils._EngineIntScalarType.c_type(407)
    else:
        _liref = utils._EngineIntScalarType.c_type(0) # Added by code generator.
        _liref_py = 0 # Added by code generator.
    if _genid_py in (1, 2, 3):
        _iref_dim1 = 33*_idim_py+7
    elif _genid_py == 4:
        _iref_dim1 = 407
    else:
        _iref_dim1 = 0
    _iref = utils._EngineIntArrayType.empty_ctypes(_iref_dim1)
    _comm_py = {}
    _comm_py['iref'] = _iref
    _iskip = utils._EngineIntScalarType(
        iskip, {'fun_name': fun_name, 'entity': 'iskip'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _genid,
        _idim,
        _iref,
        _liref,
        _iskip,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, genid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: genid = 1, 2, 3 or 4.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= idim <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iskip < 0 or iskip is too large: iskip = {:s},'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'maximum value is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def quasi_uniform(n, rcord, quas, comm):
    u"""
    Generates a uniform quasi-random number sequence.

    ``quasi_uniform`` generates a uniformly distributed low-discrepancy
    sequence as proposed by Sobol, Faure or Niederreiter.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05ym

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ymf.html

    Parameters
    ----------
    n : int
        The number of quasi-random numbers required.

    rcord : int
        The order in which the generated values are returned.

    quas : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `rcord`=1: idim; if `rcord`=2: `n`;
        otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `rcord`=1: `n`; if `rcord`=2: idim;
        otherwise: 0.

        `On exit`: contains the `n` quasi-random numbers of dimension
        idim.

        If `rcord` = 1, `quas`[i-1,j-1] holds the jth value for the ith
        dimension.

        If `rcord` = 2, `quas`[i-1,j-1] holds the ith value for the jth
        dimension.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 1)
            On entry, value of `n` would result in too many calls to the
            generator: `n` = *<value>*, generator has previously been
            called *<value>* times.

        (`errno` 2)
            On entry, `rcord` = *<value>*.

            Constraint: `rcord` = 1 or 2.

        (`errno` 4)
            On entry, ldquas = *<value>* and `n` = *<value>*.

            Constraint: if `rcord` = 2, ldquas >= `n`.

        (`errno` 5)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 84)
            On entry, the specified dimensions are out of sync.

            A different number of values have been generated from at
            least one of the specified dimensions.

    Notes
    -----
    Low discrepancy (quasi-random) sequences are used in numerical
    integration, simulation and optimization.
    Like pseudorandom numbers they are uniformly distributed but they
    are not statistically independent, rather they are designed to give
    more even distribution in multidimensional space (uniformity).
    Therefore, they are often more efficient than pseudorandom numbers
    in multidimensional Monte Carlo methods.

    ``quasi_uniform`` generates a set of points x^1,x^2,...,x^N with
    high uniformity in the S-dimensional unit cube I^S = [0,1]^S.

    Let G be a subset of I^S and define the counting function S_N(G) as
    the number of points x^i in G.
    For each x = (x_1,x_2,...,x_S) in I^S, let G_x be the rectangular
    S-dimensional region

        G_x = [0,x_1)*[0,x_2)*...*[0,x_S)

    with volume x_1,x_2,...,x_S.
    Then one measure of the uniformity of the points x^1,x^2,...,x^N is
    the discrepancy:

        D_N^*(x^1,x^2,...,x^N) = sup_{x in
        I^S}(|S_N(G_x)-Nx_1,x_2,...,x_S|).

    which has the form

        D_N^*(x^1,x^2,...,x^N) <= C_S(log(N))^S+O((log(N))^(S-1)) for
        all N >= 2.

    The principal aim in the construction of low-discrepancy sequences
    is to find sequences of points in I^S with a bound of this form
    where the constant C_S is as small as possible.

    The type of low-discrepancy sequence generated by ``quasi_uniform``
    depends on the initialization function called and can include those
    proposed by Sobol, Faure or Niederreiter.
    If the initialization function ``quasi_init_scrambled`` was used
    then the sequence will be scrambled (see [Notes for
    quasi_init_scrambled] for details).

    References
    ----------
    Bratley, P and Fox, B L, 1988, `Algorithm 659: implementing Sobol's
    quasirandom sequence generator`, ACM Trans. Math. Software (14(1)),
    88--100

    Fox, B L, 1986, `Algorithm 647: implementation and relative
    efficiency of quasirandom sequence generators`, ACM Trans. Math.
    Software (12(4)), 362--376
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_uniform'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ymft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, rcord.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._BYREF_INT, # Input, ldquas.
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, idim.
        utils._BYREF_INT, # Input, tdquas.
        utils._BYREF_INT, # Input, liref.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _rcord = utils._EngineIntScalarType(
        rcord, {'fun_name': fun_name, 'entity': 'rcord'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _rcord_py = utils._EngineIntScalarType.to_py(_rcord)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    if _rcord_py == 1:
        _quas_dim1 = _vl_idim_py
    elif _rcord_py == 2:
        _quas_dim1 = _n_py
    else:
        _quas_dim1 = 0
    if _rcord_py == 1:
        _quas_dim2 = _n_py
    elif _rcord_py == 2:
        _quas_dim2 = _vl_idim_py
    else:
        _quas_dim2 = 0
    exp_shape_quas = (_quas_dim1, _quas_dim2)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=exp_shape_quas,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _idim = utils._EngineIntScalarType(
        _vl_idim_py, {'fun_name': fun_name, 'entity': '_idim'},
    ).to_ctypes()
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _n,
        _rcord,
        _quas,
        _ldquas,
        _iref,
        _idim,
        _tdquas,
        _liref,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 21 and _ifmt == 21:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, value of n would result in too many calls to the generator:',
                    'n = {:s}, generator has previously been called'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'times.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rcord = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: rcord = 1 or 2.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _ierr_fl = 4
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldquas = {:s} and n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: if rcord = 2, ldquas >= n.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 5
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 84 and _ifmt == 84:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the specified dimensions are out of sync.',
                    'A different number of values have been generated from',
                    'at least one of the specified dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_init_scrambled(genid, stype, idim, iskip, nsdigi, statecomm):
    u"""
    Initializes a scrambled quasi-random number generator.

    ``quasi_init_scrambled`` initializes a scrambled quasi-random
    generator prior to calling ``quasi_uniform``, ``quasi_normal`` or
    ``quasi_lognormal``.
    It must be preceded by a call to one of the pseudorandom
    initialization functions ``init_repeat`` or ``init_nonrepeat``.

    For full information please refer to the NAG Library document for
    g05yn

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ynf.html

    Parameters
    ----------
    genid : int
        Must identify the quasi-random generator to use.

        `genid` = 1

            Sobol generator.

        `genid` = 2

            Sobol (A659) generator.

        `genid` = 3

            Niederreiter generator.

    stype : int
        Must identify the scrambling method to use.

        `stype` = 0

            No scrambling. This is equivalent to calling ``quasi_init``.

        `stype` = 1

            Owen like scrambling.

        `stype` = 2

            Faure--Tezuka scrambling.

        `stype` = 3

            Owen and Faure--Tezuka scrambling.

    idim : int
        The number of dimensions required.

    iskip : int
        The number of terms of the sequence to skip on initialization
        for the Sobol and Niederreiter generators.

    nsdigi : int
        Controls the number of digits (bits) to scramble when `genid` =
        1 or 2, otherwise `nsdigi` is ignored. If `nsdigi` < 1 or
        `nsdigi` > 30 then all the digits are scrambled.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    Returns
    -------
    comm : dict, communication object
        Communication structure.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `genid` = *<value>*.

            Constraint: 1 <= `genid` <= *<value>*.

        (`errno` 2)
            On entry, `stype` = *<value>*.

            Constraint: 0 <= `stype` <= 3.

        (`errno` 3)
            On entry, `idim` = *<value>*.

            Constraint: 1 <= `idim` <= *<value>*.

        (`errno` 6)
            On entry, `iskip` = *<value>*.

            Constraint: 0 <= `iskip` <= 2^30.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    ``quasi_init_scrambled`` selects a quasi-random number generator
    through the input value of `genid`, a method of scrambling through
    the input value of `stype` and initializes the `comm`['iref']
    communication array for use in the functions ``quasi_uniform``,
    ``quasi_normal`` or ``quasi_lognormal``.

    Scrambled quasi-random sequences are an extension of standard
    quasi-random sequences that attempt to eliminate the bias inherent
    in a quasi-random sequence whilst retaining the low-discrepancy
    properties.
    The use of a scrambled sequence allows error estimation of Monte
    Carlo results by performing a number of iterates and computing the
    variance of the results.

    This implementation of scrambled quasi-random sequences is based on
    TOMS Algorithm 823 and details can be found in the accompanying
    paper, Hong and Hickernell (2003).
    Three methods of scrambling are supplied; the first a restricted
    form of Owen's scrambling (Owen (1995)), the second based on the
    method of Faure and Tezuka (2000) and the last method combines the
    first two.

    Scrambled versions of the Niederreiter sequence and two sets of
    Sobol sequences are provided.
    The first Sobol sequence is obtained using `genid` = 1.
    The first 10000 direction numbers for this sequence are based on the
    work of Joe and Kuo (2008).
    For dimensions greater than 10000 the direction numbers are randomly
    generated using the pseudorandom generator specified in
    `statecomm`['state'] (see Jäckel (2002) for details).
    The second Sobol sequence is obtained using `genid` = 2 and referred
    to in the documentation as 'Sobol (A659)'.
    The first 1111 direction numbers for this sequence are based on
    Algorithm 659 of Bratley and Fox (1988) with the extension proposed
    by Joe and Kuo (2003).
    For dimensions greater than 1111 the direction numbers are once
    again randomly generated.
    The Niederreiter sequence is obtained by setting `genid` = 3.

    References
    ----------
    Bratley, P and Fox, B L, 1988, `Algorithm 659: implementing Sobol's
    quasirandom sequence generator`, ACM Trans. Math. Software (14(1)),
    88--100

    Faure, H and Tezuka, S, 2000, `Another random scrambling of digital
    (t,s)-sequences`, Monte Carlo and Quasi-Monte Carlo Methods,
    Springer-Verlag, Berlin, Germany, (eds K T Fang, F J Hickernell and
    H Niederreiter)

    Hong, H S and Hickernell, F J, 2003, `Algorithm 823: implementing
    scrambled digital sequences`, ACM Trans. Math. Software (29:2),
    95--109

    Jäckel, P, 2002, `Monte Carlo Methods in Finance`, Wiley Finance
    Series, John Wiley and Sons, England

    Joe, S and Kuo, F Y, 2003, `Remark on Algorithm 659: implementing
    Sobol's quasirandom sequence generator`, ACM Trans. Math. Software
    (TOMS) (29), 49--57

    Joe, S and Kuo, F Y, 2008, `Constructing Sobol sequences with better
    two-dimensional projections`, SIAM J. Sci. Comput. (30), 2635--2654

    Niederreiter, H, 1988, `Low-discrepancy and low dispersion
    sequences`, Journal of Number Theory (30), 51--70

    Owen, A B, 1995, `Randomly permuted (t,m,s)-nets and
    (t,s)-sequences`, Monte Carlo and Quasi-Monte Carlo Methods in
    Scientific Computing, Lecture Notes in Statistics (106),
    Springer-Verlag, New York, NY, 299--317, (eds H Niederreiter and P
    J-S Shiue)

    See Also
    --------
    :meth:`naginterfaces.library.examples.rand.bb_inc_ex.main`
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_init_scrambled'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ynft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, genid.
        utils._BYREF_INT, # Input, stype.
        utils._BYREF_INT, # Input, idim.
        utils._EngineIntArrayType.c_type, # Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._BYREF_INT, # Input, iskip.
        utils._BYREF_INT, # Input, nsdigi.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _genid = utils._EngineIntScalarType(
        genid, {'fun_name': fun_name, 'entity': 'genid'},
    ).to_ctypes()
    _stype = utils._EngineIntScalarType(
        stype, {'fun_name': fun_name, 'entity': 'stype'},
    ).to_ctypes()
    _idim = utils._EngineIntScalarType(
        idim, {'fun_name': fun_name, 'entity': 'idim'},
    ).to_ctypes()
    _idim_py = utils._EngineIntScalarType.to_py(_idim)
    _liref = utils._EngineIntScalarType(
        33*_idim_py+7, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = utils._EngineIntArrayType.empty_ctypes(33*_idim_py+7)
    _comm_py = {}
    _comm_py['iref'] = _iref
    _iskip = utils._EngineIntScalarType(
        iskip, {'fun_name': fun_name, 'entity': 'iskip'},
    ).to_ctypes()
    _nsdigi = utils._EngineIntScalarType(
        nsdigi, {'fun_name': fun_name, 'entity': 'nsdigi'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _genid,
        _stype,
        _idim,
        _iref,
        _liref,
        _iskip,
        _nsdigi,
        _state,
        _lstate,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 10 and _ifmt == 10:
            _ierr_fl = 1
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, genid = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= genid <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 1 and _ifmt == 1:
            _ierr_fl = 2
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, stype = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= stype <= 3.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _ierr_fl = 3
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, idim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 1 <= idim <= {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _ierr_fl = 6
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, iskip = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0 <= iskip <= 2^30.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _ierr_fl = 8
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        return_data = _comm_py
    _err_class.manage()
    return return_data

def quasi_uniform_bydim(sorder, n, fdim, ldim, quas, comm):
    u"""
    Generates a uniform quasi-random number sequence, for a subset of
    dimensions.

    ``quasi_uniform_bydim`` generates a uniformly distributed
    low-discrepancy sequence as proposed by Sobol or Niederreiter.
    Values are generated for a subset of dimensions.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yp

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ypf.html

    Parameters
    ----------
    sorder : int
        The order in which the generated values are returned.

    n : int
        The number of quasi-random numbers required.

    fdim : int
        The first dimension to return.

    ldim : int
        The last dimension to return.

    quas : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2:
        `ldim`-`fdim`+1; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `ldim`-`fdim`+1; if
        `sorder`=2: `n`; otherwise: 0.

        `On exit`: contains the `n` quasi-random numbers for the
        required dimensions of a sequence with idim dimensions.

        For i = 1,2,...,`n`, j = `fdim`,`fdim`+1,...,`ldim` and k =
        j-`fdim`+1, the ith value for the jth dimension is held in

            `quas`[i-1,k-1], if `sorder` = 1,

            `quas`[k-1,i-1], if `sorder` = 2.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 22)
            On entry, value of `n` would result in too many calls to the
            generator: `n` = *<value>*, generator has previously been
            called *<value>* times.

        (`errno` 31)
            On entry, `fdim` = *<value>* and `ldim` = *<value>*.

            Constraint: 1 <= `fdim` <= `ldim`

        (`errno` 41)
            On entry, `ldim` = *<value>* and idim = *<value>*.

            Constraint: `ldim` <= idim

        (`errno` 62)
            On entry, ldquas = *<value>* and `ldim`-`fdim`+1 =
            *<value>*.

            Constraint: ldquas >= `ldim`-`fdim`+1.

        (`errno` 81)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 83)
            On entry, `comm`['iref'] is too short to use with
            ``quasi_uniform_bydim``.

        (`errno` 84)
            On entry, the specified dimensions are out of sync.

            A different number of values have been generated from at
            least one of the specified dimensions.

        (`errno` 85)
            ``quasi_uniform_bydim`` can not be used with the Faure
            generator.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quasi_uniform_bydim`` generates a sequence of low discrepancy
    (quasi-random) values for a specified subset of dimensions.
    See [the G05 Introduction] for more details.

    Prior to calling this function one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled`` must be called and the
    type of sequence returned by ``quasi_uniform_bydim`` can include the
    low-discrepancy sequences proposed by Sobol or Niederreiter,
    depending on the arguments supplied to the initialization routine
    used.
    If the initialization function ``quasi_init_scrambled`` was used
    then the returned sequence is scrambled (see [Notes for
    quasi_init_scrambled] for details).
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_uniform_bydim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ypft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._BYREF_INT, # Input, fdim.
        utils._BYREF_INT, # Input, ldim.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._BYREF_INT, # Input, ldquas.
        utils._BYREF_INT, # Input, tdquas.
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _fdim = utils._EngineIntScalarType(
        fdim, {'fun_name': fun_name, 'entity': 'fdim'},
    ).to_ctypes()
    _ldim = utils._EngineIntScalarType(
        ldim, {'fun_name': fun_name, 'entity': 'ldim'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ldim_py = utils._EngineIntScalarType.to_py(_ldim)
    _fdim_py = utils._EngineIntScalarType.to_py(_fdim)
    if _sorder_py == 1:
        _quas_dim1 = _n_py
    elif _sorder_py == 2:
        _quas_dim1 = _ldim_py-_fdim_py+1
    else:
        _quas_dim1 = 0
    if _sorder_py == 1:
        _quas_dim2 = _ldim_py-_fdim_py+1
    elif _sorder_py == 2:
        _quas_dim2 = _n_py
    else:
        _quas_dim2 = 0
    exp_shape_quas = (_quas_dim1, _quas_dim2)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=exp_shape_quas,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _fdim,
        _ldim,
        _quas,
        _ldquas,
        _tdquas,
        _iref,
        _liref,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, value of n would result in too many calls to the generator:',
                    'n = {:s}, generator has previously been called'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'times.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fdim = {:s} and ldim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= fdim <= ldim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldim = {:s} and idim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldim <= idim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 62 and _ifmt == 62:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldquas = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ldim-fdim+1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldquas >= ldim-fdim+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 81 and _ifmt == 81:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 82 and _ifmt == 82:
            _ierr_fl = 81
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 81
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 83 and _ifmt == 83:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] is too short to use with quasi_uniform_bydim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 84 and _ifmt == 84:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the specified dimensions are out of sync.',
                    'A different number of values have been generated from',
                    'at least one of the specified dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 85 and _ifmt == 85:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'quasi_uniform_bydim can not be used with the Faure generator.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_normal_bydim(sorder, n, xmean, std, fdim, ldim, quas, comm):
    u"""
    Generates a Normal quasi-random number sequence, for a subset of
    dimensions.

    ``quasi_normal_bydim`` generates a quasi-random sequence from a
    Normal (Gaussian) distribution.
    Values are generated for a subset of dimensions.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yq

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05yqf.html

    Parameters
    ----------
    sorder : int
        The order in which the generated values are returned.

    n : int
        The number of quasi-random numbers required.

    xmean : float, shape (idim)
        Specifies, for each dimension, the mean of the Normal
        distribution.

    std : float, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        Normal distribution.

    fdim : int
        The first dimension to return.

    ldim : int
        The last dimension to return.

    quas : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2:
        `ldim`-`fdim`+1; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `ldim`-`fdim`+1; if
        `sorder`=2: `n`; otherwise: 0.

        `On exit`: contains the `n` quasi-random numbers for the
        required dimensions of a sequence with idim dimensions.

        For i = 1,2,...,`n`, j = `fdim`,`fdim`+1,...,`ldim` and k =
        j-`fdim`+1, the ith value for the jth dimension is held in

            `quas`[i-1,k-1], if `sorder` = 1,

            `quas`[k-1,i-1], if `sorder` = 2.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 22)
            On entry, value of `n` would result in too many calls to the
            generator: `n` = *<value>*, generator has previously been
            called *<value>* times.

        (`errno` 41)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.0.

        (`errno` 61)
            On entry, `fdim` = *<value>* and `ldim` = *<value>*.

            Constraint: 1 <= `fdim` <= `ldim`

        (`errno` 71)
            On entry, `ldim` = *<value>* and idim = *<value>*.

            Constraint: `ldim` <= idim

        (`errno` 92)
            On entry, ldquas = *<value>* and `ldim`-`fdim`+1 =
            *<value>*.

            Constraint: ldquas >= `ldim`-`fdim`+1.

        (`errno` 111)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 113)
            On entry, `comm`['iref'] is too short to use with
            ``quasi_normal_bydim``.

        (`errno` 114)
            On entry, the specified dimensions are out of sync.

            A different number of values have been generated from at
            least one of the specified dimensions.

        (`errno` 115)
            ``quasi_normal_bydim`` can not be used with the Faure
            generator.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quasi_normal_bydim`` generates a quasi-random sequence, for a
    specified subset of dimensions, from a Normal distribution by first
    generating a uniform quasi-random sequence, for the specified subset
    of dimensions, which is then transformed into a Normal sequence
    using the inverse of the Normal CDF.
    The type of uniform sequence used depends on the initialization
    function called and can include the low-discrepancy sequences
    proposed by Sobol, Faure or Niederreiter.
    If the initialization function ``quasi_init_scrambled`` was used
    then the underlying uniform sequence is first scrambled prior to
    being transformed (see [Notes for quasi_init_scrambled] for
    details).
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_normal_bydim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05yqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xmean[idim].
        utils._EngineFloat64ArrayType.c_type, # Input, std[idim].
        utils._BYREF_INT, # Input, idim.
        utils._BYREF_INT, # Input, fdim.
        utils._BYREF_INT, # Input, ldim.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._BYREF_INT, # Input, ldquas.
        utils._BYREF_INT, # Input, tdquas.
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _idim_py = utils._EngineFloat64ArrayType.get_shape(xmean, {'fun_name': fun_name, 'entity': 'xmean'}, exp_rank=1)[0]
    _idim = utils._EngineIntScalarType(
        _idim_py, {'fun_name': fun_name, 'entity': '_idim'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _xmean = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _std = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _fdim = utils._EngineIntScalarType(
        fdim, {'fun_name': fun_name, 'entity': 'fdim'},
    ).to_ctypes()
    _ldim = utils._EngineIntScalarType(
        ldim, {'fun_name': fun_name, 'entity': 'ldim'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ldim_py = utils._EngineIntScalarType.to_py(_ldim)
    _fdim_py = utils._EngineIntScalarType.to_py(_fdim)
    if _sorder_py == 1:
        _quas_dim1 = _n_py
    elif _sorder_py == 2:
        _quas_dim1 = _ldim_py-_fdim_py+1
    else:
        _quas_dim1 = 0
    if _sorder_py == 1:
        _quas_dim2 = _ldim_py-_fdim_py+1
    elif _sorder_py == 2:
        _quas_dim2 = _n_py
    else:
        _quas_dim2 = 0
    exp_shape_quas = (_quas_dim1, _quas_dim2)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=exp_shape_quas,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _xmean,
        _std,
        _idim,
        _fdim,
        _ldim,
        _quas,
        _ldquas,
        _tdquas,
        _iref,
        _liref,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, value of n would result in too many calls to the generator:',
                    'n = {:s}, generator has previously been called'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'times.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and std[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: std[i-1] >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fdim = {:s} and ldim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= fdim <= ldim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldim = {:s} and idim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldim <= idim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldquas = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ldim-fdim+1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldquas >= ldim-fdim+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 113 and _ifmt == 113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] is too short to use with quasi_normal_bydim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 114 and _ifmt == 114:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the specified dimensions are out of sync.',
                    'A different number of values have been generated from',
                    'at least one of the specified dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 115 and _ifmt == 115:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'quasi_normal_bydim can not be used with the Faure generator.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def quasi_lognormal_bydim(sorder, n, xmean, std, fdim, ldim, quas, comm):
    u"""
    Generates a log-normal quasi-random number sequence, for a subset of
    dimensions.

    ``quasi_lognormal_bydim`` generates a quasi-random sequence from a
    log-normal distribution.
    It must be preceded by a call to one of the initialization functions
    ``quasi_init`` or ``quasi_init_scrambled``.

    For full information please refer to the NAG Library document for
    g05yr

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05yrf.html

    Parameters
    ----------
    sorder : int
        The order in which the generated values are returned.

    n : int
        The number of quasi-random numbers required.

    xmean : float, shape (idim)
        Specifies, for each dimension, the mean of the underlying Normal
        distribution.

    std : float, shape (idim)
        Specifies, for each dimension, the standard deviation of the
        underlying Normal distribution.

    fdim : int
        The first dimension to return.

    ldim : int
        The last dimension to return.

    quas : float, shape (:, :), modified in place
        Note: the required extent for this argument in dimension 1 is
        determined as follows: if `sorder`=1: `n`; if `sorder`=2:
        `ldim`-`fdim`+1; otherwise: 0.

        Note: the required extent for this argument in dimension 2 is
        determined as follows: if `sorder`=1: `ldim`-`fdim`+1; if
        `sorder`=2: `n`; otherwise: 0.

        `On exit`: contains the `n` quasi-random numbers for the
        required dimensions of a sequence with idim dimensions.

        For i = 1,2,...,`n`, j = `fdim`,`fdim`+1,...,`ldim` and k =
        j-`fdim`+1, the ith value for the jth dimension is held in

            `quas`[i-1,k-1], if `sorder` = 1,

            `quas`[k-1,i-1], if `sorder` = 2.

    comm : dict, communication object, modified in place
        Communication structure.

        This argument must have been initialized by a prior call to
        ``quasi_init`` or ``quasi_init_scrambled``.

    Raises
    ------
    NagValueError
        (`errno` 11)
            On entry, `sorder` = *<value>*.

            Constraint: `sorder` = 1 or 2.

        (`errno` 21)
            On entry, `n` = *<value>*.

            Constraint: `n` >= 0.

        (`errno` 22)
            On entry, value of `n` would result in too many calls to the
            generator: `n` = *<value>*, generator has previously been
            called *<value>* times.

        (`errno` 31)
            On entry, i = *<value>* and `xmean`[i-1] = *<value>*.

            Constraint: abs(`xmean`[i-1]) <= *<value>*.

        (`errno` 41)
            On entry, i = *<value>* and `std`[i-1] = *<value>*.

            Constraint: `std`[i-1] >= 0.

        (`errno` 61)
            On entry, `fdim` = *<value>* and `ldim` = *<value>*.

            Constraint: 1 <= `fdim` <= `ldim`

        (`errno` 71)
            On entry, `ldim` = *<value>* and idim = *<value>*.

            Constraint: `ldim` <= idim

        (`errno` 92)
            On entry, ldquas = *<value>* and `ldim`-`fdim`+1 =
            *<value>*.

            Constraint: ldquas >= `ldim`-`fdim`+1.

        (`errno` 111)
            On entry, `comm`['iref'] has either not been initialized or
            has been corrupted.

        (`errno` 113)
            On entry, `comm`['iref'] is too short to use with
            ``quasi_lognormal_bydim``.

        (`errno` 114)
            On entry, the specified dimensions are out of sync.

            A different number of values have been generated from at
            least one of the specified dimensions.

        (`errno` 115)
            ``quasi_lognormal_bydim`` can not be used with the Faure
            generator.

    Notes
    -----
    No equivalent traditional C interface for this routine exists in the
    NAG Library.

    ``quasi_lognormal_bydim`` generates a quasi-random sequence, for a
    specified subset of dimensions, from a log-normal distribution by
    first generating a uniform quasi-random sequence, for the specified
    subset of dimensions, which is then transformed into a log-normal
    sequence using the exponential of the inverse of the Normal CDF.
    The type of uniform sequence used depends on the initialization
    function called and can include the low-discrepancy sequences
    proposed by Sobol, Faure or Niederreiter.
    If the initialization function ``quasi_init_scrambled`` was used
    then the underlying uniform sequence is first scrambled prior to
    being transformed (see [Notes for quasi_init_scrambled] for
    details).
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.quasi_lognormal_bydim'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05yrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, sorder.
        utils._BYREF_INT, # Input, n.
        utils._EngineFloat64ArrayType.c_type, # Input, xmean[idim].
        utils._EngineFloat64ArrayType.c_type, # Input, std[idim].
        utils._BYREF_INT, # Input, idim.
        utils._BYREF_INT, # Input, fdim.
        utils._BYREF_INT, # Input, ldim.
        utils._EngineFloat64ArrayType.c_type, # Output, quas[ldquas, tdquas].
        utils._BYREF_INT, # Input, ldquas.
        utils._BYREF_INT, # Input, tdquas.
        utils._EngineIntArrayType.c_type, # Input/Output, iref[liref].
        utils._BYREF_INT, # Input, liref.
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((quas, 'quas'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _sorder = utils._EngineIntScalarType(
        sorder, {'fun_name': fun_name, 'entity': 'sorder'},
    ).to_ctypes()
    _n = utils._EngineIntScalarType(
        n, {'fun_name': fun_name, 'entity': 'n'},
    ).to_ctypes()
    _idim_py = utils._EngineFloat64ArrayType.get_shape(xmean, {'fun_name': fun_name, 'entity': 'xmean'}, exp_rank=1)[0]
    _idim = utils._EngineIntScalarType(
        _idim_py, {'fun_name': fun_name, 'entity': '_idim'},
    ).to_ctypes()
    utils._check_comm(
        comm, {'fun_name': fun_name, 'entity': 'comm'},
        req_keys=('iref',),
    )
    _vl_idim_py = utils._EngineIntArrayType.get_element(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, (1,), validate=False)
    _xmean = utils._EngineFloat64ArrayType(
        xmean, {'fun_name': fun_name, 'entity': 'xmean'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _std = utils._EngineFloat64ArrayType(
        std, {'fun_name': fun_name, 'entity': 'std'},
        exp_shape=(_vl_idim_py,),
    ).to_ctypes()
    _fdim = utils._EngineIntScalarType(
        fdim, {'fun_name': fun_name, 'entity': 'fdim'},
    ).to_ctypes()
    _ldim = utils._EngineIntScalarType(
        ldim, {'fun_name': fun_name, 'entity': 'ldim'},
    ).to_ctypes()
    _ldquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='l')
    _ldquas = utils._EngineIntScalarType(
        _ldquas_py, {'fun_name': fun_name, 'entity': '_ldquas'},
    ).to_ctypes()
    _tdquas_py = utils._EngineFloat64ArrayType.get_nd(quas, {'fun_name': fun_name, 'entity': 'quas'}, exp_rank=2, sorder=_m_sorder, dim_type='s')
    _tdquas = utils._EngineIntScalarType(
        _tdquas_py, {'fun_name': fun_name, 'entity': '_tdquas'},
    ).to_ctypes()
    _sorder_py = utils._EngineIntScalarType.to_py(_sorder)
    _n_py = utils._EngineIntScalarType.to_py(_n)
    _ldim_py = utils._EngineIntScalarType.to_py(_ldim)
    _fdim_py = utils._EngineIntScalarType.to_py(_fdim)
    if _sorder_py == 1:
        _quas_dim1 = _n_py
    elif _sorder_py == 2:
        _quas_dim1 = _ldim_py-_fdim_py+1
    else:
        _quas_dim1 = 0
    if _sorder_py == 1:
        _quas_dim2 = _ldim_py-_fdim_py+1
    elif _sorder_py == 2:
        _quas_dim2 = _n_py
    else:
        _quas_dim2 = 0
    exp_shape_quas = (_quas_dim1, _quas_dim2)
    _quas_cls = utils._EngineFloat64ArrayType(
        quas, {'fun_name': fun_name, 'entity': 'quas'},
        exp_shape=exp_shape_quas,
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _quas = _quas_cls.to_ctypes(can_reference_contents=False)
    _liref_py = utils._EngineIntArrayType.get_shape(comm['iref'], {'fun_name': fun_name, 'entity': 'comm[\'iref\']'}, exp_rank=1, validate=False)[0]
    _liref = utils._EngineIntScalarType(
        _liref_py, {'fun_name': fun_name, 'entity': '_liref'},
    ).to_ctypes()
    _iref = comm['iref']
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _sorder,
        _n,
        _xmean,
        _std,
        _idim,
        _fdim,
        _ldim,
        _quas,
        _ldquas,
        _tdquas,
        _iref,
        _liref,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, sorder = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: sorder = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 21 and _ifmt == 21:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, n = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: n >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 22 and _ifmt == 22:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, value of n would result in too many calls to the generator:',
                    'n = {:s}, generator has previously been called'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    '{:s}'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'times.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 31 and _ifmt == 31:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and xmean[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: abs(xmean[i-1]) <= {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 41 and _ifmt == 41:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, i = {:s} and std[i-1] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: std[i-1] >= 0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 61 and _ifmt == 61:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, fdim = {:s} and ldim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: 1 <= fdim <= ldim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 71 and _ifmt == 71:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldim = {:s} and idim = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldim <= idim',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 92 and _ifmt == 92:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ldquas = {:s} and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'ldim-fdim+1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ldquas >= ldim-fdim+1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 111 and _ifmt == 111:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 112 and _ifmt == 112:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 201 and _ifmt == 201:
            _ierr_fl = 111
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] has either not been initialized or has been corrupted.',
                ],
                errcodes=(_ierr_fl, _ierr_py, _ifmt),
            )
        elif _ierr_py == 113 and _ifmt == 113:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, comm[\'iref\'] is too short to use with quasi_lognormal_bydim.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 114 and _ifmt == 114:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, the specified dimensions are out of sync.',
                    'A different number of values have been generated from',
                    'at least one of the specified dimensions.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 115 and _ifmt == 115:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'quasi_lognormal_bydim can not be used with the Faure generator.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _quas_cls.refresh_py(_quas)
        return_data = None
    _err_class.manage()
    return return_data

def field_1d_user_setup(ns, xmin, xmax, maxm, var, cov1, pad, icorr, lam, xx, eig, data=None):
    u"""
    Setup for simulating one-dimensional random fields, user-defined
    variogram.

    ``field_1d_user_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in one dimension, for a
    user-defined variogram, using the `circulant embedding method`.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_1d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zm

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05zmf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field.

    xmin : float
        The lower bound for the interval over which the random field is
        to be simulated.

    xmax : float
        The upper bound for the interval over which the random field is
        to be simulated.

    maxm : int
        The maximum size of the circulant matrix to use. For example, if
        the embedding matrix is to be allowed to double in size three
        times before the approximation procedure is used, then choose
        `maxm` = 2^(k+2) where k = 1+ceiling(log2(`ns`-1)).

        Suggested value: 2^(k+2) where k = 1+ceiling(log2(`ns`-1)).

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    cov1 : callable gamma = cov1(x, data=None)
        `cov1` must evaluate the variogram gamma(x), without the
        multiplicative factor sigma^2, for all x >= 0.

        The value returned in `gamma` is multiplied internally by `var`.

        Parameters
        ~~~~~~~~~~
        x : float
            The value x at which the variogram gamma(x) is to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gamma : float
            The value of the variogram gamma(x)/sigma^2.

    pad : int
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

            The embedding matrix is padded with zeros.

        `pad` = 1

            The embedding matrix is padded with values of the variogram.

        Suggested value: `pad` = 1.

    icorr : int
        Determines which approximation to implement if required, as
        described in [Notes].

        Suggested value: `icorr` = 0.

    lam : float, shape (`maxm`), modified in place
        `On exit`: contains the square roots of the eigenvalues of the
        embedding matrix.

    xx : float, shape (`ns`), modified in place
        `On exit`: the points at which values of the random field will
        be output.

    eig : float, shape (3), modified in place
        `On exit`: indicates information about the negative eigenvalues
        in the embedding matrix which have had to be set to zero.
        `eig`[0] contains the smallest eigenvalue, `eig`[1] contains the
        sum of the squares of the negative eigenvalues, and `eig`[2]
        contains the sum of the absolute values of the negative
        eigenvalues.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    m : int
        The size of the embedding matrix.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

            No approximation was used.

        `approx` = 1

            Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `maxm` = *<value>*.

            Constraint: the minimum calculated value for `maxm` is
            *<value>*.

        (`errno` 5)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 7)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 8)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.

    Notes
    -----
    A one-dimensional random field Z(x) in R is a function which is
    random at every point x in R, so Z(x) is a random variable for each
    x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    positive semidefinite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R and C(x,y) = C(x+a,y+a) for all x,y,a in R and hence we
    can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_1d_user_setup`` and ``field_1d_generate`` are
    used to simulate a one-dimensional stationary Gaussian random field,
    with mean function zero and variogram gamma(x), over an interval
    [x_min,x_max], using an equally spaced set of N points on the
    interval.
    The problem reduces to sampling a Normal random vector X of size N,
    with mean vector zero and a symmetric Toeplitz covariance matrix A.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric circulant matrix B of size M >= 2*(N-1), which can
    now be factorized as B = W Lambda W* = R*R, where W is the Fourier
    matrix (W* is the complex conjugate of W), Lambda is the diagonal
    matrix containing the eigenvalues of B and R = Lambda^(1/2)W*. B is
    known as the embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by M,
    and so only the first row (or column) of B is needed -- the whole
    matrix does not need to be formed.

    As long as all of the values of Lambda are non-negative (i.e., B is
    positive semidefinite), B is a covariance matrix for a random vector
    Y, two samples of which can now be simulated from the real and
    imaginary parts of R*(U+iV), where U and V have elements from the
    standard Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N elements of each sample of Y -- because the original
    covariance matrix A is embedded in B, X will have the correct
    distribution.

    If B is not positive semidefinite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than `maxm`, an approximation procedure is used.
    We write Lambda = Lambda_++Lambda_-, where Lambda_+ and Lambda_-
    contain the non-negative and negative eigenvalues of B respectively.
    Then B is replaced by rho B_+ where B_+ = W Lambda_+W* and rho in
    (0,1] is a scaling factor.
    The error epsilon in approximating the distribution of the random
    field is given by

        epsilon = sqrt(((1-rho)^2trace(Lambda)+rho^2trace(Lambda_-))/M).

    Three choices for rho are available, and are determined by the input
    argument `icorr`:

        setting `icorr` = 0 sets

                rho = (trace(Lambda))/(trace(Lambda_+)),

        setting `icorr` = 1 sets

                rho = sqrt((trace(Lambda))/(trace(Lambda_+))),

        setting `icorr` = 2 sets rho = 1.

    ``field_1d_user_setup`` finds a suitable positive semidefinite
    embedding matrix B and outputs its size, `m`, and the square roots
    of its eigenvalues in `lam`.
    If approximation is used, information regarding the accuracy of the
    approximation is output.
    Note that only the first row (or column) of B is actually formed and
    stored.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1994, `Simulation of stationary Gaussian
    processes in [0,1]^d`, Journal of Computational and Graphical
    Statistics (3(4)), 409--432
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_1d_user_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_cov1_name = fun_name + ':cov1'

    cov1h = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_INT, # Input, np.
        utils._EngineFloat64ArrayType.c_type, # Input, params[np].
        utils._BYREF_FLOAT64, # Output, gamma.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cov1h(_cov1, en_data, x, np, params, gamma, _iuser, _ruser):
        """Helper for cov1."""
        try:
            _x_cov1h = utils._EngineFloat64ScalarType.to_py(x.contents)
            if data is not None:
                _gamma_cov1h = cov1(
                    _x_cov1h,
                    data,
                )
            else:
                _gamma_cov1h = cov1(
                    _x_cov1h,
                )
            utils._EngineFloat64ScalarType(
                _gamma_cov1h, {'fun_name': cb_cov1_name, 'returned': 'gamma'},
            ).refresh_ctypes_buffer(gamma)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g05zmft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_INT, # Input, maxm.
        utils._BYREF_FLOAT64, # Input, var.
        utils._UNUSED_FUNCTYPE,
        cov1h,
        utils._BYREF_INT, # Input, pad.
        utils._BYREF_INT, # Input, icorr.
        utils._EngineFloat64ArrayType.c_type, # Output, lam[maxm].
        utils._EngineFloat64ArrayType.c_type, # Output, xx[ns].
        utils._BYREF_INT, # Output, m.
        utils._BYREF_INT, # Output, approx.
        utils._BYREF_FLOAT64, # Output, rho.
        utils._BYREF_INT, # Output, icount.
        utils._EngineFloat64ArrayType.c_type, # Output, eig[3].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _maxm = utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    utils._check_callable(cov1, {'fun_name': fun_name, 'entity': 'cov1'})
    _cov1 = utils._cast_to_unused_callback(cov1)
    _cov1h = utils._cast_callback(
        cov1h, cb_cov1h,
        {'fun_name': fun_name, 'entity': 'cov1'},
    )
    _pad = utils._EngineIntScalarType(
        pad, {'fun_name': fun_name, 'entity': 'pad'},
    ).to_ctypes()
    _icorr = utils._EngineIntScalarType(
        icorr, {'fun_name': fun_name, 'entity': 'icorr'},
    ).to_ctypes()
    _maxm_py = utils._EngineIntScalarType.to_py(_maxm)
    _lam_cls = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(_maxm_py,),
        check_contents=False,
    )
    _lam = _lam_cls.to_ctypes(can_reference_contents=False)
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _xx_cls = utils._EngineFloat64ArrayType(
        xx, {'fun_name': fun_name, 'entity': 'xx'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _xx = _xx_cls.to_ctypes(can_reference_contents=False)
    _m = utils._EngineIntScalarType.empty_ctypes()
    _approx = utils._EngineIntScalarType.empty_ctypes()
    _rho = utils._EngineFloat64ScalarType.empty_ctypes()
    _icount = utils._EngineIntScalarType.empty_ctypes()
    _eig_cls = utils._EngineFloat64ArrayType(
        eig, {'fun_name': fun_name, 'entity': 'eig'},
        exp_shape=(3,),
        check_contents=False,
    )
    _eig = _eig_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _xmin,
        _xmax,
        _maxm,
        _var,
        _cov1,
        _cov1h,
        _pad,
        _icorr,
        _lam,
        _xx,
        _m,
        _approx,
        _rho,
        _icount,
        _eig,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: the minimum calculated value for maxm is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pad = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pad = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icorr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icorr = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lam_cls.refresh_py(_lam)
        _xx_cls.refresh_py(_xx)
        _eig_cls.refresh_py(_eig)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _approx_py = utils._EngineIntScalarType.to_py(_approx)
        _rho_py = utils._EngineFloat64ScalarType.to_py(_rho)
        _icount_py = utils._EngineIntScalarType.to_py(_icount)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['m', 'approx', 'rho', 'icount'],
        )(
            _m_py,
            _approx_py,
            _rho_py,
            _icount_py,
        )
    _err_class.manage()
    return return_data

def field_1d_predef_setup(ns, xmin, xmax, maxm, var, icov1, np, params, pad, icorr, lam, xx, eig):
    u"""
    Setup for simulating one-dimensional random fields.

    ``field_1d_predef_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in one dimension, for a
    preset variogram, using the `circulant embedding method`.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_1d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zn

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05znf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field.

    xmin : float
        The lower bound for the interval over which the random field is
        to be simulated. Note that if `icov1` = 14 (for simulating
        fractional Brownian motion), `xmin` is not referenced and the
        lower bound for the interval is set to zero.

    xmax : float
        The upper bound for the interval over which the random field is
        to be simulated. Note that if `icov1` = 14 (for simulating
        fractional Brownian motion), the lower bound for the interval is
        set to zero and so `xmax` is required to be greater than zero.

    maxm : int
        The maximum size of the circulant matrix to use. For example, if
        the embedding matrix is to be allowed to double in size three
        times before the approximation procedure is used, then choose
        `maxm` = 2^(k+2) where k = 1+ceiling(log2(`ns`-1)).

        Suggested value: 2^(k+2) where k = 1+ceiling(log2(`ns`-1)).

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    icov1 : int
        Determines which of the preset variograms to use. The choices
        are given below. Note that x' = (|x|)/l, where l is the
        correlation length and is a parameter for most of the
        variograms, and sigma^2 is the variance specified by `var`.

        `icov1` = 1

            Symmetric stable variogram

                gamma(x) = sigma^2exp(-(x')^nu),

            where

                l = `params`[0], l > 0,

                nu = `params`[1], 0 <= nu <= 2.

        `icov1` = 2

            Cauchy variogram

                gamma(x) = sigma^2(1+(x')^2)^(-nu),

            where

                l = `params`[0], l > 0,

                nu = `params`[1], nu > 0.

        `icov1` = 3

            Differential variogram with compact support

                gamma(x) = {[table omitted]

            where

                l = `params`[0], l > 0.

        `icov1` = 4

            Exponential variogram

                gamma(x) = sigma^2exp(-x'),

            where

                l = `params`[0], l > 0.

        `icov1` = 5

            Gaussian variogram

                gamma(x) = sigma^2exp(-(x')^2),

            where

                l = `params`[0], l > 0.

        `icov1` = 6

            Nugget variogram

                gamma(x) = {[table omitted]

            No parameters need be set for this value of `icov1`.

        `icov1` = 7

            Spherical variogram

                gamma(x) = {[table omitted]

            where

                l = `params`[0], l > 0.

        `icov1` = 8

            Bessel variogram

                gamma(x) = sigma^22^nuGamma(nu+1)J_nu(x')/((x')^nu),

            where

                J_nu(·) is the Bessel function of the first kind,

                l = `params`[0], l > 0,

                nu = `params`[1], nu >= -0.5.

        `icov1` = 9

            Hole effect variogram

                gamma(x) = sigma^2(sin(x'))/(x'),

            where

                l = `params`[0], l > 0.

        `icov1` = 10

            Whittle-Matérn variogram

                gamma(x) = sigma^22^(1-nu)(x')^nu K_nu(x')/Gamma(nu),

            where

                K_nu(·) is the modified Bessel function of the second
                kind,

                l = `params`[0], l > 0,

                nu = `params`[1], nu > 0.

        `icov1` = 11

            Continuously parameterised variogram with compact support

                gamma(x) = {[table omitted]

            where

                x'' = (x')/s,

                K_nu(·) is the modified Bessel function of the second
                kind,

                l = `params`[0], l > 0,

                s = `params`[1], s > 0 (second correlation length),

                nu = `params`[2], nu > 0.

        `icov1` = 12

            Generalized hyperbolic distribution variogram

                gamma(x) =
                sigma^2((delta^2+(x')^2)^(lambda/2))/(delta^lambda
                K_lambda(kappadelta))K_lambda(kappa(delta^2+(x')^2)^(1/2)),

            where

                K_lambda(·) is the modified Bessel function of the
                second kind,

                l = `params`[0], l > 0,

                lambda = `params`[1], no constraint on lambda

                delta = `params`[2], delta > 0,

                kappa = `params`[3], kappa > 0.

        `icov1` = 13

            Cosine variogram

                gamma(x) = sigma^2cos(x'),

            where

                l = `params`[0], l > 0.

        `icov1` = 14

            Used for simulating fractional Brownian motion B^H(t).
            Fractional Brownian motion itself is not a stationary
            Gaussian random field, but its increments X~(i) =
            B^H(t_i)-B^H(t_{i-1}) can be simulated in the same way as a
            stationary random field. The variogram for the so-called
            'increment process' is

                C(X~(t_i),X~(t_j)) = gamma~(x) =
                (delta^(2H))/2(|x/delta-1|^(2H)+|x/delta+1|^(2H)-2|x/delta|^(2H)),

            where

                x = t_j-t_i,

                H = `params`[0], 0 < H < 1, H is the Hurst parameter,

                delta = `params`[1], delta > 0, normally delta =
                t_i-t_{i-1} is the (fixed) step size.

            We scale the increments to set gamma(0) = 1; let X(i) =
            X~(i)/(delta^(-H)), then

                C(X(t_i),X(t_j)) = gamma(x) =
                1/2(|x/delta-1|^(2H)+|x/delta+1|^(2H)-2|x/delta|^(2H)).

            The increments X(i) can then be simulated using
            ``field_1d_generate``, then multiplied by delta^H to obtain
            the original increments X~(i) for the fractional Brownian
            motion.

    np : int
        The number of parameters to be set. Different variograms need a
        different number of parameters.

        `icov1` = 6

            `np` must be set to 0.

        `icov1` = 3, 4, 5, 7, 9 or 13

            `np` must be set to 1.

        `icov1` = 1, 2, 8, 10 or 14

            `np` must be set to 2.

        `icov1` = 11

            `np` must be set to 3.

        `icov1` = 12

            `np` must be set to 4.

    params : float, shape (`np`)
        The parameters set for the variogram.

    pad : int
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

            The embedding matrix is padded with zeros.

        `pad` = 1

            The embedding matrix is padded with values of the variogram.

        Suggested value: `pad` = 1.

    icorr : int
        Determines which approximation to implement if required, as
        described in [Notes].

        Suggested value: `icorr` = 0.

    lam : float, shape (`maxm`), modified in place
        `On exit`: contains the square roots of the eigenvalues of the
        embedding matrix.

    xx : float, shape (`ns`), modified in place
        `On exit`: the points at which values of the random field will
        be output.

    eig : float, shape (3), modified in place
        `On exit`: indicates information about the negative eigenvalues
        in the embedding matrix which have had to be set to zero.
        `eig`[0] contains the smallest eigenvalue, `eig`[1] contains the
        sum of the squares of the negative eigenvalues, and `eig`[2]
        contains the sum of the absolute values of the negative
        eigenvalues.

    Returns
    -------
    m : int
        The size of the embedding matrix.

    approx : int
        Indicates whether approximation was used.

        `approx` = 0

            No approximation was used.

        `approx` = 1

            Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `icov1` != 14, `xmin` = *<value>* and `xmax` =
            *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 3)
            On entry, `icov1` = 14 and `xmax` = *<value>*.

            Constraint: `xmax` > 0.0.

        (`errno` 4)
            On entry, `maxm` = *<value>*.

            Constraint: the minimum calculated value for `maxm` is
            *<value>*.

        (`errno` 5)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 6)
            On entry, `icov1` = *<value>*.

            Constraint: `icov1` >= 1 and `icov1` <= 14.

        (`errno` 7)
            On entry, `np` = *<value>*.

            Constraint: for `icov1` = *<value>*, `np` = *<value>*.

        (`errno` 8)
            On entry, `params`[*<value>*] = *<value>*.

            Constraint: dependent on `icov1`.

        (`errno` 9)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 10)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.

    Notes
    -----
    A one-dimensional random field Z(x) in R is a function which is
    random at every point x in R, so Z(x) is a random variable for each
    x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    positive semidefinite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R and C(x,y) = C(x+a,y+a) for all x,y,a in R and hence we
    can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_1d_predef_setup`` and ``field_1d_generate``
    are used to simulate a one-dimensional stationary Gaussian random
    field, with mean function zero and variogram gamma(x), over an
    interval [x_min,x_max], using an equally spaced set of N points.
    The problem reduces to sampling a Normal random vector X of size N,
    with mean vector zero and a symmetric Toeplitz covariance matrix A.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric circulant matrix B of size M >= 2*(N-1), which can
    now be factorized as B = W Lambda W* = R*R, where W is the Fourier
    matrix (W* is the complex conjugate of W), Lambda is the diagonal
    matrix containing the eigenvalues of B and R = Lambda^(1/2)W*. B is
    known as the embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by M,
    and so only the first row (or column) of B is needed -- the whole
    matrix does not need to be formed.

    As long as all of the values of Lambda are non-negative (i.e., B is
    positive semidefinite), B is a covariance matrix for a random vector
    Y, two samples of which can now be simulated from the real and
    imaginary parts of R*(U+iV), where U and V have elements from the
    standard Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N elements of each sample of Y -- because the original
    covariance matrix A is embedded in B, X will have the correct
    distribution.

    If B is not positive semidefinite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than `maxm`, an approximation procedure is used.
    We write Lambda = Lambda_++Lambda_-, where Lambda_+ and Lambda_-
    contain the non-negative and negative eigenvalues of B respectively.
    Then B is replaced by rho B_+ where B_+ = W Lambda_+W* and rho in
    (0,1] is a scaling factor.
    The error epsilon in approximating the distribution of the random
    field is given by

        epsilon = sqrt(((1-rho)^2trace(Lambda)+rho^2trace(Lambda_-))/M).

    Three choices for rho are available, and are determined by the input
    argument `icorr`:

        setting `icorr` = 0 sets

                rho = (trace(Lambda))/(trace(Lambda_+)),

        setting `icorr` = 1 sets

                rho = sqrt((trace(Lambda))/(trace(Lambda_+))),

        setting `icorr` = 2 sets rho = 1.

    ``field_1d_predef_setup`` finds a suitable positive semidefinite
    embedding matrix B and outputs its size, `m`, and the square roots
    of its eigenvalues in `lam`.
    If approximation is used, information regarding the accuracy of the
    approximation is output.
    Note that only the first row (or column) of B is actually formed and
    stored.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1997, `Algorithm AS 312: An Algorithm for
    Simulating Stationary Gaussian Random Fields`, Journal of the Royal
    Statistical Society, Series C (Applied Statistics) (Volume 46) (1),
    171--181
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_1d_predef_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05znft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_INT, # Input, maxm.
        utils._BYREF_FLOAT64, # Input, var.
        utils._BYREF_INT, # Input, icov1.
        utils._BYREF_INT, # Input, np.
        utils._EngineFloat64ArrayType.c_type, # Input, params[np].
        utils._BYREF_INT, # Input, pad.
        utils._BYREF_INT, # Input, icorr.
        utils._EngineFloat64ArrayType.c_type, # Output, lam[maxm].
        utils._EngineFloat64ArrayType.c_type, # Output, xx[ns].
        utils._BYREF_INT, # Output, m.
        utils._BYREF_INT, # Output, approx.
        utils._BYREF_FLOAT64, # Output, rho.
        utils._BYREF_INT, # Output, icount.
        utils._EngineFloat64ArrayType.c_type, # Output, eig[3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _maxm = utils._EngineIntScalarType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    _icov1 = utils._EngineIntScalarType(
        icov1, {'fun_name': fun_name, 'entity': 'icov1'},
    ).to_ctypes()
    _np = utils._EngineIntScalarType(
        np, {'fun_name': fun_name, 'entity': 'np'},
    ).to_ctypes()
    _np_py = utils._EngineIntScalarType.to_py(_np)
    _params = utils._EngineFloat64ArrayType(
        params, {'fun_name': fun_name, 'entity': 'params'},
        exp_shape=(_np_py,),
    ).to_ctypes()
    _pad = utils._EngineIntScalarType(
        pad, {'fun_name': fun_name, 'entity': 'pad'},
    ).to_ctypes()
    _icorr = utils._EngineIntScalarType(
        icorr, {'fun_name': fun_name, 'entity': 'icorr'},
    ).to_ctypes()
    _maxm_py = utils._EngineIntScalarType.to_py(_maxm)
    _lam_cls = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(_maxm_py,),
        check_contents=False,
    )
    _lam = _lam_cls.to_ctypes(can_reference_contents=False)
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _xx_cls = utils._EngineFloat64ArrayType(
        xx, {'fun_name': fun_name, 'entity': 'xx'},
        exp_shape=(_ns_py,),
        check_contents=False,
    )
    _xx = _xx_cls.to_ctypes(can_reference_contents=False)
    _m = utils._EngineIntScalarType.empty_ctypes()
    _approx = utils._EngineIntScalarType.empty_ctypes()
    _rho = utils._EngineFloat64ScalarType.empty_ctypes()
    _icount = utils._EngineIntScalarType.empty_ctypes()
    _eig_cls = utils._EngineFloat64ArrayType(
        eig, {'fun_name': fun_name, 'entity': 'eig'},
        exp_shape=(3,),
        check_contents=False,
    )
    _eig = _eig_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _xmin,
        _xmax,
        _maxm,
        _var,
        _icov1,
        _np,
        _params,
        _pad,
        _icorr,
        _lam,
        _xx,
        _m,
        _approx,
        _rho,
        _icount,
        _eig,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icov1 != 14, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icov1 = 14 and xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: the minimum calculated value for maxm is {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icov1 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icov1 >= 1 and icov1 <= 14.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: for icov1 = {:s}, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, params[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dependent on icov1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pad = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pad = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icorr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icorr = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lam_cls.refresh_py(_lam)
        _xx_cls.refresh_py(_xx)
        _eig_cls.refresh_py(_eig)
        _m_py = utils._EngineIntScalarType.to_py(_m)
        _approx_py = utils._EngineIntScalarType.to_py(_approx)
        _rho_py = utils._EngineFloat64ScalarType.to_py(_rho)
        _icount_py = utils._EngineIntScalarType.to_py(_icount)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['m', 'approx', 'rho', 'icount'],
        )(
            _m_py,
            _approx_py,
            _rho_py,
            _icount_py,
        )
    _err_class.manage()
    return return_data

def field_1d_generate(ns, s, m, lam, rho, statecomm, z):
    u"""
    Generates realizations of a one-dimensional random field.

    ``field_1d_generate`` produces realizations of a stationary Gaussian
    random field in one dimension, using the circulant embedding method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zp

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05zpf.html

    Parameters
    ----------
    ns : int
        The number of sample points to be generated in realizations of
        the random field. This must be the same value as supplied to
        ``field_1d_predef_setup`` or ``field_1d_user_setup`` when
        calculating the eigenvalues of the embedding matrix.

    s : int
        S, the number of realizations of the random field to simulate.

    m : int
        M, the size of the embedding matrix, as returned by
        ``field_1d_predef_setup`` or ``field_1d_user_setup``.

    lam : float, shape (`m`)
        Must contain the square roots of the eigenvalues of the
        embedding matrix, as returned by ``field_1d_predef_setup`` or
        ``field_1d_user_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_1d_predef_setup`` or ``field_1d_user_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    z : float, shape (`ns`, `s`), modified in place
        `On exit`: contains the realizations of the random field. The
        jth realization, for the `ns` sample points, is stored in
        `z`[i-1,j-1], for i = 1,2,...,`ns`. The sample points are as
        returned in xx by ``field_1d_predef_setup`` or
        ``field_1d_user_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, `m` = *<value>* and `ns` = *<value>*.

            Constraint: `m` >= max(1,2*(`ns`-1)).

        (`errno` 4)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 <= `rho` <= 1.0.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    A one-dimensional random field Z(x) in R is a function which is
    random at every point x in R, so Z(x) is a random variable for each
    x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    non-negative definite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R and C(x,y) = C(x+a,y+a) for all x,y,a in R and hence we
    can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_1d_user_setup`` or ``field_1d_predef_setup``,
    along with ``field_1d_generate``, are used to simulate a
    one-dimensional stationary Gaussian random field, with mean function
    zero and variogram gamma(x), over an interval [x_min,x_max], using
    an equally spaced set of N points.
    The problem reduces to sampling a Normal random vector X of size N,
    with mean vector zero and a symmetric Toeplitz covariance matrix A.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric circulant matrix B of size M >= 2*(N-1), which can
    now be factorized as B = W Lambda W* = R*R, where W is the Fourier
    matrix (W* is the complex conjugate of W), Lambda is the diagonal
    matrix containing the eigenvalues of B and R = Lambda^(1/2)W*. B is
    known as the embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by M,
    and so only the first row (or column) of B is needed -- the whole
    matrix does not need to be formed.

    As long as all of the values of Lambda are non-negative (i.e., B is
    non-negative definite), B is a covariance matrix for a random vector
    Y, two samples of which can now be simulated from the real and
    imaginary parts of R*(U+iV), where U and V have elements from the
    standard Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N elements of each sample of Y -- because the original
    covariance matrix A is embedded in B, X will have the correct
    distribution.

    If B is not non-negative definite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than maxm, an approximation procedure is used.
    See the documentation of ``field_1d_user_setup`` or
    ``field_1d_predef_setup`` for details of the approximation
    procedure.

    ``field_1d_generate`` takes the square roots of the eigenvalues of
    the embedding matrix B, and its size M, as input and outputs S
    realizations of the random field in Z.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``field_1d_generate``.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1994, `Simulation of stationary Gaussian
    processes in [0,1]^d`, Journal of Computational and Graphical
    Statistics (3(4)), 409--432
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_1d_generate'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05zpft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_INT, # Input, s.
        utils._BYREF_INT, # Input, m.
        utils._EngineFloat64ArrayType.c_type, # Input, lam[m].
        utils._BYREF_FLOAT64, # Input, rho.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineFloat64ArrayType.c_type, # Output, z[ns, s].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _s = utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _lam = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _rho = utils._EngineFloat64ScalarType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _s_py = utils._EngineIntScalarType.to_py(_s)
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_ns_py, _s_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _s,
        _m,
        _lam,
        _rho,
        _state,
        _lstate,
        _z,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s} and ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= max(1,2*(ns-1)).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of lam was negative.',
                    'Constraint: all elements of lam must be non-negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rho = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 <= rho <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        return_data = None
    _err_class.manage()
    return return_data

def field_2d_user_setup(ns, xmin, xmax, ymin, ymax, maxm, var, cov2, even, pad, icorr, lam, xx, yy, m, eig, data=None):
    u"""
    Setup for simulating two-dimensional random fields, user-defined
    variogram.

    ``field_2d_user_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in two dimensions, for a
    user-defined variogram, using the `circulant embedding method`.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_2d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zq

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05zqf.html

    Parameters
    ----------
    ns : int, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction, N_1 and `ns`[1] sample
        points in the y-direction, N_2. The total number of sample
        points on the grid is, therefore, `ns`[0]*`ns`[1].

    xmin : float
        The lower bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    xmax : float
        The upper bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    ymin : float
        The lower bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    ymax : float
        The upper bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    maxm : int, shape (2)
        Determines the maximum size of the circulant matrix to use -- a
        maximum of `maxm`[0] elements in the x-direction, and a maximum
        of `maxm`[1] elements in the y-direction. The maximum size of
        the circulant matrix is thus `maxm`[0]*`maxm`[1].

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    cov2 : callable gamma = cov2(x, y, data=None)
        `cov2` must evaluate the variogram gamma(x) for all x if `even`
        = 0, and for all x with non-negative entries if `even` = 1.

        The value returned in `gamma` is multiplied internally by `var`.

        Parameters
        ~~~~~~~~~~
        x : float
            The coordinate x at which the variogram gamma(x) is to be
            evaluated.

        y : float
            The coordinate y at which the variogram gamma(x) is to be
            evaluated.

        data : arbitrary, optional, modifiable in place
            User-communication data for callback functions.

        Returns
        ~~~~~~~
        gamma : float
            The value of the variogram gamma(x).

    even : int
        Indicates whether the covariance function supplied is even or
        uneven.

        `even` = 0

            The covariance function is uneven.

        `even` = 1

            The covariance function is even.

    pad : int
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

            The embedding matrix is padded with zeros.

        `pad` = 1

            The embedding matrix is padded with values of the variogram.

        Suggested value: `pad` = 1.

    icorr : int
        Determines which approximation to implement if required, as
        described in [Notes].

        Suggested value: `icorr` = 0.

    lam : float, shape (`maxm`[0]*`maxm`[1]), modified in place
        `On exit`: contains the square roots of the eigenvalues of the
        embedding matrix.

    xx : float, shape (`ns`[0]), modified in place
        `On exit`: the points of the x-coordinates at which values of
        the random field will be output.

    yy : float, shape (`ns`[1]), modified in place
        `On exit`: the points of the y-coordinates at which values of
        the random field will be output.

    m : int, shape (2), modified in place
        `On exit`: `m`[0] contains M_1, the size of the circulant blocks
        and `m`[1] contains M_2, the number of blocks, resulting in a
        final square matrix of size M_1*M_2.

    eig : float, shape (3), modified in place
        `On exit`: indicates information about the negative eigenvalues
        in the embedding matrix which have had to be set to zero.
        `eig`[0] contains the smallest eigenvalue, `eig`[1] contains the
        sum of the squares of the negative eigenvalues, and `eig`[2]
        contains the sum of the absolute values of the negative
        eigenvalues.

    data : arbitrary, optional
        User-communication data for callback functions.

    Returns
    -------
    approx : int
        Indicates whether approximation was used.

        `approx` = 0

            No approximation was used.

        `approx` = 1

            Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 6)
            On entry, `maxm` = [*<value>*,*<value>*].

            Constraint: the minima for `maxm` are [*<value>*,*<value>*].

        (`errno` 7)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 9)
            On entry, `even` = *<value>*.

            Constraint: `even` = 0 or 1.

        (`errno` 10)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 11)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.

    Notes
    -----
    A two-dimensional random field Z(x) in R^2 is a function which is
    random at every point x in R^2, so Z(x) is a random variable for
    each x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    positive semidefinite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R^2, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R^2 and C(x,y) = C(x+a,y+a) for all x,y,a in R^2 and hence
    we can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_2d_user_setup`` and ``field_2d_generate`` are
    used to simulate a two-dimensional stationary Gaussian random field,
    with mean function zero and variogram gamma(x), over a domain
    [x_min,x_max]*[y_min,y_max], using an equally spaced set of N_1*N_2
    points; N_1 points in the x-direction and N_2 points in the
    y-direction.
    The problem reduces to sampling a Normal random vector X of size
    N_1*N_2, with mean vector zero and a symmetric covariance matrix A,
    which is an N_2*N_2 block Toeplitz matrix with Toeplitz blocks of
    size N_1*N_1.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric matrix B, which is an M_2*M_2 block circulant
    matrix with circulant blocks of size M_1*M_1, where M_1 >= 2*(N_1-1)
    and M_2 >= 2*(N_2-1). B can now be factorized as B = W Lambda W* =
    R*R, where W is the two-dimensional Fourier matrix (W* is the
    complex conjugate of W), Lambda is the diagonal matrix containing
    the eigenvalues of B and R = Lambda^(1/2)W*. B is known as the
    embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by
    M_1*M_2, and so only the first row (or column) of B is needed -- the
    whole matrix does not need to be formed.

    The symmetry of A as a block matrix, and the symmetry of each block
    of A, depends on whether the variogram gamma is even or not. gamma
    is even in its first coordinate if gamma([-x_1,x_2]^T) =
    gamma([x_1,x_2]^T), even in its second coordinate if
    gamma([x_1,-x_2]^T) = gamma([x_1,x_2]^T), and even if it is even in
    both coordinates (in two dimensions it is impossible for gamma to be
    even in one coordinate and uneven in the other).
    If gamma is even then A is a symmetric block matrix and has
    symmetric blocks; if gamma is uneven then A is not a symmetric block
    matrix and has non-symmetric blocks.
    In the uneven case, M_1 and M_2 are set to be odd in order to
    guarantee symmetry in B.

    As long as all of the values of Lambda are non-negative (i.e., B is
    positive semidefinite), B is a covariance matrix for a random vector
    Y which has M_2 blocks of size M_1.
    Two samples of Y can now be simulated from the real and imaginary
    parts of R*(U+iV), where U and V have elements from the standard
    Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N_1 elements of the first N_2 blocks of each sample of Y
    -- because the original covariance matrix A is embedded in B, X will
    have the correct distribution.

    If B is not positive semidefinite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than `maxm`, an approximation procedure is used.
    We write Lambda = Lambda_++Lambda_-, where Lambda_+ and Lambda_-
    contain the non-negative and negative eigenvalues of B respectively.
    Then B is replaced by rho B_+ where B_+ = W Lambda_+W* and rho in
    (0,1] is a scaling factor.
    The error epsilon in approximating the distribution of the random
    field is given by

        epsilon = sqrt(((1-rho)^2trace(Lambda)+rho^2trace(Lambda_-))/M).

    Three choices for rho are available, and are determined by the input
    argument `icorr`:

        setting `icorr` = 0 sets

                rho = (trace(Lambda))/(trace(Lambda_+)),

        setting `icorr` = 1 sets

                rho = sqrt((trace(Lambda))/(trace(Lambda_+))),

        setting `icorr` = 2 sets rho = 1.

    ``field_2d_user_setup`` finds a suitable positive semidefinite
    embedding matrix B and outputs its sizes in the vector `m` and the
    square roots of its eigenvalues in `lam`.
    If approximation is used, information regarding the accuracy of the
    approximation is output.
    Note that only the first row (or column) of B is actually formed and
    stored.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1994, `Simulation of stationary Gaussian
    processes in [0,1]^d`, Journal of Computational and Graphical
    Statistics (3(4)), 409--432
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_2d_user_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    cb_cov2_name = fun_name + ':cov2'

    cov2h = utils._EngineState._sys_functype(
        None,
        utils._UNUSED_FUNCTYPE,
        utils._BYREF_ENGINEDATA,
        utils._BYREF_FLOAT64, # Input, x.
        utils._BYREF_FLOAT64, # Input, y.
        utils._BYREF_INT, # Input, norm.
        utils._BYREF_INT, # Input, np.
        utils._EngineFloat64ArrayType.c_type, # Input, params[np].
        utils._BYREF_FLOAT64, # Output, gamma.
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
    )

    def cb_cov2h(_cov2, en_data, x, y, norm, np, params, gamma, _iuser, _ruser):
        """Helper for cov2."""
        try:
            _x_cov2h = utils._EngineFloat64ScalarType.to_py(x.contents)
            _y_cov2h = utils._EngineFloat64ScalarType.to_py(y.contents)
            if data is not None:
                _gamma_cov2h = cov2(
                    _x_cov2h,
                    _y_cov2h,
                    data,
                )
            else:
                _gamma_cov2h = cov2(
                    _x_cov2h,
                    _y_cov2h,
                )
            utils._EngineFloat64ScalarType(
                _gamma_cov2h, {'fun_name': cb_cov2_name, 'returned': 'gamma'},
            ).refresh_ctypes_buffer(gamma)
        except:
            excs.append(_sys.exc_info())
            en_data.contents.hlperr = utils._IERR_HELPER

    engine_ctypes = utils._EngineState._get_symbol('g05zqft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, ns[2].
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_FLOAT64, # Input, ymin.
        utils._BYREF_FLOAT64, # Input, ymax.
        utils._EngineIntArrayType.c_type, # Input, maxm[2].
        utils._BYREF_FLOAT64, # Input, var.
        utils._UNUSED_FUNCTYPE,
        cov2h,
        utils._BYREF_INT, # Input, even.
        utils._BYREF_INT, # Input, pad.
        utils._BYREF_INT, # Input, icorr.
        utils._EngineFloat64ArrayType.c_type, # Output, lam[maxm[0]*maxm[1]].
        utils._EngineFloat64ArrayType.c_type, # Output, xx[ns[0]].
        utils._EngineFloat64ArrayType.c_type, # Output, yy[ns[1]].
        utils._EngineIntArrayType.c_type, # Output, m[2].
        utils._BYREF_INT, # Output, approx.
        utils._BYREF_FLOAT64, # Output, rho.
        utils._BYREF_INT, # Output, icount.
        utils._EngineFloat64ArrayType.c_type, # Output, eig[3].
        utils._EngineIntArrayType.c_type, # Input/Output, iuser[*].
        utils._EngineFloat64ArrayType.c_type, # Input/Output, ruser[*].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ns = utils._EngineIntArrayType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
        exp_shape=(2,),
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _ymin = utils._EngineFloat64ScalarType(
        ymin, {'fun_name': fun_name, 'entity': 'ymin'},
    ).to_ctypes()
    _ymax = utils._EngineFloat64ScalarType(
        ymax, {'fun_name': fun_name, 'entity': 'ymax'},
    ).to_ctypes()
    _maxm = utils._EngineIntArrayType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
        exp_shape=(2,),
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    utils._check_callable(cov2, {'fun_name': fun_name, 'entity': 'cov2'})
    _cov2 = utils._cast_to_unused_callback(cov2)
    _cov2h = utils._cast_callback(
        cov2h, cb_cov2h,
        {'fun_name': fun_name, 'entity': 'cov2'},
    )
    _even = utils._EngineIntScalarType(
        even, {'fun_name': fun_name, 'entity': 'even'},
    ).to_ctypes()
    _pad = utils._EngineIntScalarType(
        pad, {'fun_name': fun_name, 'entity': 'pad'},
    ).to_ctypes()
    _icorr = utils._EngineIntScalarType(
        icorr, {'fun_name': fun_name, 'entity': 'icorr'},
    ).to_ctypes()
    _lam_cls = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(utils._EngineIntArrayType.get_element(maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (0,))*utils._EngineIntArrayType.get_element(maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (1,)),),
        check_contents=False,
    )
    _lam = _lam_cls.to_ctypes(can_reference_contents=False)
    _xx_cls = utils._EngineFloat64ArrayType(
        xx, {'fun_name': fun_name, 'entity': 'xx'},
        exp_shape=(utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,)),),
        check_contents=False,
    )
    _xx = _xx_cls.to_ctypes(can_reference_contents=False)
    _yy_cls = utils._EngineFloat64ArrayType(
        yy, {'fun_name': fun_name, 'entity': 'yy'},
        exp_shape=(utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)),),
        check_contents=False,
    )
    _yy = _yy_cls.to_ctypes(can_reference_contents=False)
    _m_cls = utils._EngineIntArrayType(
        m, {'fun_name': fun_name, 'entity': 'm'},
        exp_shape=(2,),
        check_contents=False,
    )
    _m = _m_cls.to_ctypes(can_reference_contents=False)
    _approx = utils._EngineIntScalarType.empty_ctypes()
    _rho = utils._EngineFloat64ScalarType.empty_ctypes()
    _icount = utils._EngineIntScalarType.empty_ctypes()
    _eig_cls = utils._EngineFloat64ArrayType(
        eig, {'fun_name': fun_name, 'entity': 'eig'},
        exp_shape=(3,),
        check_contents=False,
    )
    _eig = _eig_cls.to_ctypes(can_reference_contents=False)
    _iuser = None
    _ruser = None
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _xmin,
        _xmax,
        _ymin,
        _ymax,
        _maxm,
        _var,
        _cov2,
        _cov2h,
        _even,
        _pad,
        _icorr,
        _lam,
        _xx,
        _yy,
        _m,
        _approx,
        _rho,
        _icount,
        _eig,
        _iuser,
        _ruser,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns[0] >= 1, ns[1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ymin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ymax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ymin < ymax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: the minima for maxm are',
                    '[{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, even = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: even = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pad = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pad = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icorr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icorr = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lam_cls.refresh_py(_lam)
        _xx_cls.refresh_py(_xx)
        _yy_cls.refresh_py(_yy)
        _m_cls.refresh_py(_m)
        _eig_cls.refresh_py(_eig)
        _approx_py = utils._EngineIntScalarType.to_py(_approx)
        _rho_py = utils._EngineFloat64ScalarType.to_py(_rho)
        _icount_py = utils._EngineIntScalarType.to_py(_icount)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['approx', 'rho', 'icount'],
        )(
            _approx_py,
            _rho_py,
            _icount_py,
        )
    _err_class.manage()
    return return_data

def field_2d_predef_setup(ns, xmin, xmax, ymin, ymax, maxm, var, icov2, norm, np, params, pad, icorr, lam, xx, yy, m, eig):
    u"""
    Setup for simulating two-dimensional random fields, preset
    variogram.

    ``field_2d_predef_setup`` performs the setup required in order to
    simulate stationary Gaussian random fields in two dimensions, for a
    preset variogram, using the `circulant embedding method`.
    Specifically, the eigenvalues of the extended covariance matrix (or
    embedding matrix) are calculated, and their square roots output, for
    use by ``field_2d_generate``, which simulates the random field.

    For full information please refer to the NAG Library document for
    g05zr

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05zrf.html

    Parameters
    ----------
    ns : int, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction, N_1 and `ns`[1] sample
        points in the y-direction, N_2. The total number of sample
        points on the grid is, therefore, `ns`[0]*`ns`[1].

    xmin : float
        The lower bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    xmax : float
        The upper bound for the x-coordinate, for the region in which
        the random field is to be simulated.

    ymin : float
        The lower bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    ymax : float
        The upper bound for the y-coordinate, for the region in which
        the random field is to be simulated.

    maxm : int, shape (2)
        Determines the maximum size of the circulant matrix to use -- a
        maximum of `maxm`[0] elements in the x-direction, and a maximum
        of `maxm`[1] elements in the y-direction. The maximum size of
        the circulant matrix is thus `maxm`[0]*`maxm`[1].

    var : float
        The multiplicative factor sigma^2 of the variogram gamma(x).

    icov2 : int
        Determines which of the preset variograms to use. The choices
        are given below. Note that x' = ||x/(l_1),y/(l_2)||, where l_1
        and l_2 are correlation lengths in the x and y directions
        respectively and are parameters for most of the variograms, and
        sigma^2 is the variance specified by `var`.

        `icov2` = 1

            Symmetric stable variogram

                gamma(x) = sigma^2exp(-(x')^nu),

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                nu = `params`[2], 0 < nu <= 2.

        `icov2` = 2

            Cauchy variogram

                gamma(x) = sigma^2(1+(x')^2)^(-nu),

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                nu = `params`[2], nu > 0.

        `icov2` = 3

            Differential variogram with compact support

                gamma(x) = {[table omitted]

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0.

        `icov2` = 4

            Exponential variogram

                gamma(x) = sigma^2exp(-x'),

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0.

        `icov2` = 5

            Gaussian variogram

                gamma(x) = sigma^2exp(-(x')^2),

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0.

        `icov2` = 6

            Nugget variogram

                gamma(x) = {[table omitted]

            No parameters need be set for this value of `icov2`.

        `icov2` = 7

            Spherical variogram

                gamma(x) = {[table omitted]

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0.

        `icov2` = 8

            Bessel variogram

                gamma(x) = sigma^22^nuGamma(nu+1)J_nu(x')/((x')^nu),

            where

                J_nu(·) is the Bessel function of the first kind,

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                nu = `params`[2], nu >= 0.

        `icov2` = 9

            Hole effect variogram

                gamma(x) = sigma^2(sin(x'))/(x'),

            where

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0.

        `icov2` = 10

            Whittle-Matérn variogram

                gamma(x) = sigma^22^(1-nu)(x')^nu K_nu(x')/Gamma(nu),

            where

                K_nu(·) is the modified Bessel function of the second
                kind,

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                nu = `params`[2], nu > 0.

        `icov2` = 11

            Continuously parameterised variogram with compact support

                gamma(x) = {[table omitted]

            where

                x′′ = ||(x')/l_1 s_1,(y')/l_2 s_2||,

                K_nu(·) is the modified Bessel function of the second
                kind,

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                s_1 = `params`[2], s_1 > 0,

                s_2 = `params`[3], s_2 > 0,

                nu = `params`[4], nu > 0.

        `icov2` = 12

            Generalized hyperbolic distribution variogram

                gamma(x) =
                sigma^2((delta^2+(x')^2)^(lambda/2))/(delta^lambda
                K_lambda(kappadelta))K_lambda(kappa(delta^2+(x')^2)^(1/2)),

            where

                K_lambda(·) is the modified Bessel function of the
                second kind,

                l_1 = `params`[0], l_1 > 0,

                l_2 = `params`[1], l_2 > 0,

                lambda = `params`[2], no constraint on lambda,

                delta = `params`[3], delta > 0,

                kappa = `params`[4], kappa > 0.

    norm : int
        Determines which norm to use when calculating the variogram.

        `norm` = 1

            The 1-norm is used, i.e., ||x,y|| = |x|+|y|.

        `norm` = 2

            The 2-norm (Euclidean norm) is used, i.e., ||x,y|| =
            sqrt(x^2+y^2).

        Suggested value: `norm` = 2.

    np : int
        The number of parameters to be set. Different covariance
        functions need a different number of parameters.

        `icov2` = 6

            `np` must be set to 0.

        `icov2` = 3, 4, 5, 7 or 9

            `np` must be set to 2.

        `icov2` = 1, 2, 8 or 10

            `np` must be set to 3.

        `icov2` = 11 or 12

            `np` must be set to 5.

    params : float, shape (`np`)
        The parameters for the variogram as detailed in the description
        of `icov2`.

    pad : int
        Determines whether the embedding matrix is padded with zeros, or
        padded with values of the variogram. The choice of padding may
        affect how big the embedding matrix must be in order to be
        positive semidefinite.

        `pad` = 0

            The embedding matrix is padded with zeros.

        `pad` = 1

            The embedding matrix is padded with values of the variogram.

        Suggested value: `pad` = 1.

    icorr : int
        Determines which approximation to implement if required, as
        described in [Notes].

        Suggested value: `icorr` = 0.

    lam : float, shape (`maxm`[0]*`maxm`[1]), modified in place
        `On exit`: contains the square roots of the eigenvalues of the
        embedding matrix.

    xx : float, shape (`ns`[0]), modified in place
        `On exit`: the points of the x-coordinates at which values of
        the random field will be output.

    yy : float, shape (`ns`[1]), modified in place
        `On exit`: the points of the y-coordinates at which values of
        the random field will be output.

    m : int, shape (2), modified in place
        `On exit`: `m`[0] contains M_1, the size of the circulant blocks
        and `m`[1] contains M_2, the number of blocks, resulting in a
        final square matrix of size M_1*M_2.

    eig : float, shape (3), modified in place
        `On exit`: indicates information about the negative eigenvalues
        in the embedding matrix which have had to be set to zero.
        `eig`[0] contains the smallest eigenvalue, `eig`[1] contains the
        sum of the squares of the negative eigenvalues, and `eig`[2]
        contains the sum of the absolute values of the negative
        eigenvalues.

    Returns
    -------
    approx : int
        Indicates whether approximation was used.

        `approx` = 0

            No approximation was used.

        `approx` = 1

            Approximation was used.

    rho : float
        Indicates the scaling of the covariance matrix. `rho` = 1.0
        unless approximation was used with `icorr` = 0 or 1.

    icount : int
        Indicates the number of negative eigenvalues in the embedding
        matrix which have had to be set to zero.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `xmin` = *<value>* and `xmax` = *<value>*.

            Constraint: `xmin` < `xmax`.

        (`errno` 4)
            On entry, `ymin` = *<value>* and `ymax` = *<value>*.

            Constraint: `ymin` < `ymax`.

        (`errno` 6)
            On entry, `maxm` = [*<value>*,*<value>*].

            Constraint: the minimum calculated value for `maxm` are
            [*<value>*,*<value>*].

        (`errno` 7)
            On entry, `var` = *<value>*.

            Constraint: `var` >= 0.0.

        (`errno` 8)
            On entry, `icov2` = *<value>*.

            Constraint: `icov2` >= 1 and `icov2` <= 12.

        (`errno` 9)
            On entry, `norm` = *<value>*.

            Constraint: `norm` = 1 or 2.

        (`errno` 10)
            On entry, `np` = *<value>*.

            Constraint: for `icov2` = *<value>*, `np` = *<value>*.

        (`errno` 11)
            On entry, `params`[*<value>*] = *<value>*.

            Constraint: dependent on `icov2`, see documentation.

        (`errno` 12)
            On entry, `pad` = *<value>*.

            Constraint: `pad` = 0 or 1.

        (`errno` 13)
            On entry, `icorr` = *<value>*.

            Constraint: `icorr` = 0, 1 or 2.

    Notes
    -----
    A two-dimensional random field Z(x) in R^2 is a function which is
    random at every point x in R^2, so Z(x) is a random variable for
    each x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    positive semidefinite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R^2, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R^2 and C(x,y) = C(x+a,y+a) for all x,y,a in R^2 and hence
    we can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_2d_predef_setup`` and ``field_2d_generate``
    are used to simulate a two-dimensional stationary Gaussian random
    field, with mean function zero and variogram gamma(x), over a domain
    [x_min,x_max]*[y_min,y_max], using an equally spaced set of N_1*N_2
    points; N_1 points in the x-direction and N_2 points in the
    y-direction.
    The problem reduces to sampling a Gaussian random vector X of size
    N_1*N_2, with mean vector zero and a symmetric covariance matrix A,
    which is an N_2*N_2 block Toeplitz matrix with Toeplitz blocks of
    size N_1*N_1.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric matrix B, which is an M_2*M_2 block circulant
    matrix with circulant blocks of size M_1*M_1, where M_1 >= 2*(N_1-1)
    and M_2 >= 2*(N_2-1). B can now be factorized as B = W Lambda W* =
    R*R, where W is the two-dimensional Fourier matrix (W* is the
    complex conjugate of W), Lambda is the diagonal matrix containing
    the eigenvalues of B and R = Lambda^(1/2)W*. B is known as the
    embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by
    M_1*M_2, and so only the first row (or column) of B is needed -- the
    whole matrix does not need to be formed.

    As long as all of the values of Lambda are non-negative (i.e., B is
    positive semidefinite), B is a covariance matrix for a random vector
    Y which has M_2 blocks of size M_1.
    Two samples of Y can now be simulated from the real and imaginary
    parts of R*(U+iV), where U and V have elements from the standard
    Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N_1 elements of the first N_2 blocks of each sample of Y
    -- because the original covariance matrix A is embedded in B, X will
    have the correct distribution.

    If B is not positive semidefinite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than `maxm`, an approximation procedure is used.
    We write Lambda = Lambda_++Lambda_-, where Lambda_+ and Lambda_-
    contain the non-negative and negative eigenvalues of B respectively.
    Then B is replaced by rho B_+ where B_+ = W Lambda_+W* and rho in
    (0,1] is a scaling factor.
    The error epsilon in approximating the distribution of the random
    field is given by

        epsilon = sqrt(((1-rho)^2trace(Lambda)+rho^2trace(Lambda_-))/M).

    Three choices for rho are available, and are determined by the input
    argument `icorr`:

        setting `icorr` = 0 sets

                rho = (trace(Lambda))/(trace(Lambda_+)),

        setting `icorr` = 1 sets

                rho = sqrt((trace(Lambda))/(trace(Lambda_+))),

        setting `icorr` = 2 sets rho = 1.

    ``field_2d_predef_setup`` finds a suitable positive semidefinite
    embedding matrix B and outputs its sizes in the vector `m` and the
    square roots of its eigenvalues in `lam`.
    If approximation is used, information regarding the accuracy of the
    approximation is output.
    Note that only the first row (or column) of B is actually formed and
    stored.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1997, `Algorithm AS 312: An Algorithm for
    Simulating Stationary Gaussian Random Fields`, Journal of the Royal
    Statistical Society, Series C (Applied Statistics) (Volume 46) (1),
    171--181
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_2d_predef_setup'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05zrft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, ns[2].
        utils._BYREF_FLOAT64, # Input, xmin.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_FLOAT64, # Input, ymin.
        utils._BYREF_FLOAT64, # Input, ymax.
        utils._EngineIntArrayType.c_type, # Input, maxm[2].
        utils._BYREF_FLOAT64, # Input, var.
        utils._BYREF_INT, # Input, icov2.
        utils._BYREF_INT, # Input, norm.
        utils._BYREF_INT, # Input, np.
        utils._EngineFloat64ArrayType.c_type, # Input, params[np].
        utils._BYREF_INT, # Input, pad.
        utils._BYREF_INT, # Input, icorr.
        utils._EngineFloat64ArrayType.c_type, # Output, lam[maxm[0]*maxm[1]].
        utils._EngineFloat64ArrayType.c_type, # Output, xx[ns[0]].
        utils._EngineFloat64ArrayType.c_type, # Output, yy[ns[1]].
        utils._EngineIntArrayType.c_type, # Output, m[2].
        utils._BYREF_INT, # Output, approx.
        utils._BYREF_FLOAT64, # Output, rho.
        utils._BYREF_INT, # Output, icount.
        utils._EngineFloat64ArrayType.c_type, # Output, eig[3].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec)
    _ns = utils._EngineIntArrayType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
        exp_shape=(2,),
    ).to_ctypes()
    _xmin = utils._EngineFloat64ScalarType(
        xmin, {'fun_name': fun_name, 'entity': 'xmin'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _ymin = utils._EngineFloat64ScalarType(
        ymin, {'fun_name': fun_name, 'entity': 'ymin'},
    ).to_ctypes()
    _ymax = utils._EngineFloat64ScalarType(
        ymax, {'fun_name': fun_name, 'entity': 'ymax'},
    ).to_ctypes()
    _maxm = utils._EngineIntArrayType(
        maxm, {'fun_name': fun_name, 'entity': 'maxm'},
        exp_shape=(2,),
    ).to_ctypes()
    _var = utils._EngineFloat64ScalarType(
        var, {'fun_name': fun_name, 'entity': 'var'},
    ).to_ctypes()
    _icov2 = utils._EngineIntScalarType(
        icov2, {'fun_name': fun_name, 'entity': 'icov2'},
    ).to_ctypes()
    _norm = utils._EngineIntScalarType(
        norm, {'fun_name': fun_name, 'entity': 'norm'},
    ).to_ctypes()
    _np = utils._EngineIntScalarType(
        np, {'fun_name': fun_name, 'entity': 'np'},
    ).to_ctypes()
    _np_py = utils._EngineIntScalarType.to_py(_np)
    _params = utils._EngineFloat64ArrayType(
        params, {'fun_name': fun_name, 'entity': 'params'},
        exp_shape=(_np_py,),
    ).to_ctypes()
    _pad = utils._EngineIntScalarType(
        pad, {'fun_name': fun_name, 'entity': 'pad'},
    ).to_ctypes()
    _icorr = utils._EngineIntScalarType(
        icorr, {'fun_name': fun_name, 'entity': 'icorr'},
    ).to_ctypes()
    _lam_cls = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(utils._EngineIntArrayType.get_element(maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (0,))*utils._EngineIntArrayType.get_element(maxm, {'fun_name': fun_name, 'entity': 'maxm'}, (1,)),),
        check_contents=False,
    )
    _lam = _lam_cls.to_ctypes(can_reference_contents=False)
    _xx_cls = utils._EngineFloat64ArrayType(
        xx, {'fun_name': fun_name, 'entity': 'xx'},
        exp_shape=(utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,)),),
        check_contents=False,
    )
    _xx = _xx_cls.to_ctypes(can_reference_contents=False)
    _yy_cls = utils._EngineFloat64ArrayType(
        yy, {'fun_name': fun_name, 'entity': 'yy'},
        exp_shape=(utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)),),
        check_contents=False,
    )
    _yy = _yy_cls.to_ctypes(can_reference_contents=False)
    _m_cls = utils._EngineIntArrayType(
        m, {'fun_name': fun_name, 'entity': 'm'},
        exp_shape=(2,),
        check_contents=False,
    )
    _m = _m_cls.to_ctypes(can_reference_contents=False)
    _approx = utils._EngineIntScalarType.empty_ctypes()
    _rho = utils._EngineFloat64ScalarType.empty_ctypes()
    _icount = utils._EngineIntScalarType.empty_ctypes()
    _eig_cls = utils._EngineFloat64ArrayType(
        eig, {'fun_name': fun_name, 'entity': 'eig'},
        exp_shape=(3,),
        check_contents=False,
    )
    _eig = _eig_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _xmin,
        _xmax,
        _ymin,
        _ymax,
        _maxm,
        _var,
        _icov2,
        _norm,
        _np,
        _params,
        _pad,
        _icorr,
        _lam,
        _xx,
        _yy,
        _m,
        _approx,
        _rho,
        _icount,
        _eig,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns[0] >= 1, ns[1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: xmin < xmax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ymin = {:s} and'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'ymax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ymin < ymax.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, maxm = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: the minimum calculated value for maxm are',
                    '[{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, var = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: var >= 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icov2 = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icov2 >= 1 and icov2 <= 12.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 9 and _ifmt == 9:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, norm = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: norm = 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 10 and _ifmt == 10:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: for icov2 = {:s}, np = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                    ),
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 11 and _ifmt == 11:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, params[{:s}] = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: dependent on icov2, see documentation.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 12 and _ifmt == 12:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, pad = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: pad = 0 or 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 13 and _ifmt == 13:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, icorr = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: icorr = 0, 1 or 2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _lam_cls.refresh_py(_lam)
        _xx_cls.refresh_py(_xx)
        _yy_cls.refresh_py(_yy)
        _m_cls.refresh_py(_m)
        _eig_cls.refresh_py(_eig)
        _approx_py = utils._EngineIntScalarType.to_py(_approx)
        _rho_py = utils._EngineFloat64ScalarType.to_py(_rho)
        _icount_py = utils._EngineIntScalarType.to_py(_icount)
        return_data = _collections.namedtuple(
            'nag_return_data',
            ['approx', 'rho', 'icount'],
        )(
            _approx_py,
            _rho_py,
            _icount_py,
        )
    _err_class.manage()
    return return_data

def field_2d_generate(ns, s, m, lam, rho, statecomm, z):
    u"""
    Generates realizations of a two-dimensional random field.

    ``field_2d_generate`` produces realizations of a stationary Gaussian
    random field in two dimensions, using the circulant embedding
    method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_2d_user_setup`` or ``field_2d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zs

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05zsf.html

    Parameters
    ----------
    ns : int, shape (2)
        The number of sample points to use in each direction, with
        `ns`[0] sample points in the x-direction and `ns`[1] sample
        points in the y-direction. The total number of sample points on
        the grid is, therefore, `ns`[0]*`ns`[1]. This must be the same
        value as supplied to ``field_2d_user_setup`` or
        ``field_2d_predef_setup`` when calculating the eigenvalues of
        the embedding matrix.

    s : int
        S, the number of realizations of the random field to simulate.

    m : int, shape (2)
        Indicates the size, M, of the embedding matrix as returned by
        ``field_2d_user_setup`` or ``field_2d_predef_setup``. The
        embedding matrix is a block circulant matrix with circulant
        blocks. `m`[0] is the size of each block, and `m`[1] is the
        number of blocks.

    lam : float, shape (`m`[0]*`m`[1])
        Contains the square roots of the eigenvalues of the embedding
        matrix, as returned by ``field_2d_user_setup`` or
        ``field_2d_predef_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_2d_user_setup`` or ``field_2d_predef_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    z : float, shape (`ns`[0]*`ns`[1], `s`), modified in place
        `On exit`: contains the realizations of the random field. The
        kth realization (where k = 1,2,...,`s`) of the random field on
        the two-dimensional grid (x_i,y_j) is stored in `z`[ (j-1)*
        `ns`[0] +i -1,k-1], for i = 1,2,...,`ns`[0] and for j =
        1,2,...,`ns`[1]. The points are returned in xx and yy by
        ``field_2d_user_setup`` or ``field_2d_predef_setup``.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = [*<value>*,*<value>*].

            Constraint: `ns`[0] >= 1, `ns`[1] >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, `m` = [*<value>*,*<value>*], and `ns` =
            [*<value>*,*<value>*].

            Constraints: `m`[i-1] >= max(1, 2*(`ns`[i-1])-1), for i =
            1,2.

        (`errno` 4)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 5)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 < `rho` <= 1.0.

        (`errno` 6)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    A two-dimensional random field Z(x) in R^2 is a function which is
    random at every point x in R^2, so Z(x) is a random variable for
    each x.
    The random field has a mean function mu(x) = 𝔼[Z(x)] and a symmetric
    positive semidefinite covariance function C(x,y) =
    𝔼[(Z(x)-mu(x))(Z(y)-mu(y))]. Z(x) is a Gaussian random field if for
    any choice of n in ℕ and x_1,...,x_n in R^2, the random vector
    [Z(x_1),...,Z(x_n)]^T follows a multivariate Normal distribution,
    which would have a mean vector mu~ with entries mu~_i = mu(x_i) and
    a covariance matrix C~ with entries C~_{ij} = C(x_i,x_j).
    A Gaussian random field Z(x) is stationary if mu(x) is constant for
    all x in R^2 and C(x,y) = C(x+a,y+a) for all x,y,a in R^2 and hence
    we can express the covariance function C(x,y) as a function gamma of
    one variable: C(x,y) = gamma(x-y). gamma is known as a variogram (or
    more correctly, a semivariogram) and includes the multiplicative
    factor sigma^2 representing the variance such that gamma(0) =
    sigma^2.

    The functions ``field_2d_user_setup`` or ``field_2d_predef_setup``
    along with ``field_2d_generate`` are used to simulate a
    two-dimensional stationary Gaussian random field, with mean function
    zero and variogram gamma(x), over a domain
    [x_min,x_max]*[y_min,y_max], using an equally spaced set of N_1*N_2
    points; N_1 points in the x-direction and N_2 points in the
    y-direction.
    The problem reduces to sampling a Gaussian random vector X of size
    N_1*N_2, with mean vector zero and a symmetric covariance matrix A,
    which is an N_2*N_2 block Toeplitz matrix with Toeplitz blocks of
    size N_1*N_1.
    Since A is in general expensive to factorize, a technique known as
    the `circulant embedding method` is used. A is embedded into a
    larger, symmetric matrix B, which is an M_2*M_2 block circulant
    matrix with circulant bocks of size M_1*M_1, where M_1 >= 2*(N_1-1)
    and M_2 >= 2*(N_2-1). B can now be factorized as B = W Lambda W* =
    R*R, where W is the two-dimensional Fourier matrix (W* is the
    complex conjugate of W), Lambda is the diagonal matrix containing
    the eigenvalues of B and R = Lambda^(1/2)W*. B is known as the
    embedding matrix.
    The eigenvalues can be calculated by performing a discrete Fourier
    transform of the first row (or column) of B and multiplying by
    M_1*M_2, and so only the first row (or column) of B is needed -- the
    whole matrix does not need to be formed.

    The symmetry of A as a block matrix, and the symmetry of each block
    of A, depends on whether the covariance function gamma is even or
    not. gamma is even if gamma(x) = gamma(-x) for all x in R^2, and
    uneven otherwise (in higher dimensions, gamma can be even in some
    coordinates and uneven in others, but in two dimensions gamma is
    either even in both coordinates or uneven in both coordinates).
    If gamma is even then A is a symmetric block matrix and has
    symmetric blocks; if gamma is uneven then A is not a symmetric block
    matrix and has non-symmetric blocks.
    In the uneven case, M_1 and M_2 are set to be odd in order to
    guarantee symmetry in B.

    As long as all of the values of Lambda are non-negative (i.e., B is
    positive semidefinite), B is a covariance matrix for a random vector
    Y which has M_2 'blocks' of size M_1.
    Two samples of Y can now be simulated from the real and imaginary
    parts of R*(U+iV), where U and V have elements from the standard
    Normal distribution.
    Since R*(U+iV) = W Lambda^(1/2)(U+iV), this calculation can be done
    using a discrete Fourier transform of the vector Lambda^(1/2)(U+iV).
    Two samples of the random vector X can now be recovered by taking
    the first N_1 elements of the first N_2 blocks of each sample of Y
    -- because the original covariance matrix A is embedded in B, X will
    have the correct distribution.

    If B is not positive semidefinite, larger embedding matrices B can
    be tried; however if the size of the matrix would have to be larger
    than maxm, an approximation procedure is used.
    See the documentation of ``field_2d_user_setup`` or
    ``field_2d_predef_setup`` for details of the approximation
    procedure.

    ``field_2d_generate`` takes the square roots of the eigenvalues of
    the embedding matrix B, and its size vector M, as input and outputs
    S realizations of the random field in Z.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``field_2d_generate``.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1994, `Simulation of stationary Gaussian
    processes in [0,1]^d`, Journal of Computational and Graphical
    Statistics (3(4)), 409--432
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_2d_generate'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05zsft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._EngineIntArrayType.c_type, # Input, ns[2].
        utils._BYREF_INT, # Input, s.
        utils._EngineIntArrayType.c_type, # Input, m[2].
        utils._EngineFloat64ArrayType.c_type, # Input, lam[m[0]*m[1]].
        utils._BYREF_FLOAT64, # Input, rho.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineFloat64ArrayType.c_type, # Output, z[ns[0]*ns[1], s].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ns = utils._EngineIntArrayType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
        exp_shape=(2,),
    ).to_ctypes()
    _s = utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _m = utils._EngineIntArrayType(
        m, {'fun_name': fun_name, 'entity': 'm'},
        exp_shape=(2,),
    ).to_ctypes()
    _lam = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(utils._EngineIntArrayType.get_element(m, {'fun_name': fun_name, 'entity': 'm'}, (0,))*utils._EngineIntArrayType.get_element(m, {'fun_name': fun_name, 'entity': 'm'}, (1,)),),
    ).to_ctypes()
    _rho = utils._EngineFloat64ScalarType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _s_py = utils._EngineIntScalarType.to_py(_s)
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (0,))*utils._EngineIntArrayType.get_element(ns, {'fun_name': fun_name, 'entity': 'ns'}, (1,)), _s_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _s,
        _m,
        _lam,
        _rho,
        _state,
        _lstate,
        _z,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: ns[0] >= 1, ns[1] >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = [{:s},{:s}], and'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'ns = [{:s},{:s}].'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[3]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[4]),
                    ),
                    'Constraints: m[i-1] >= max(1, 2*(ns[i-1])-1), for i = 1,2.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of lam was negative.',
                    'Constraint: all elements of lam must be non-negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rho = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < rho <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        return_data = None
    _err_class.manage()
    return return_data

def field_fracbm_generate(ns, s, m, xmax, h, lam, rho, statecomm, z, xx):
    u"""
    Generates realizations of fractional Brownian motion.

    ``field_fracbm_generate`` produces realizations of a fractional
    Brownian motion, using the circulant embedding method.
    The square roots of the eigenvalues of the extended covariance
    matrix (or embedding matrix) need to be input, and can be calculated
    using ``field_1d_predef_setup``.

    For full information please refer to the NAG Library document for
    g05zt

    https://www.nag.com/numeric/nl/nagdoc_28.5/flhtml/g05/g05ztf.html

    Parameters
    ----------
    ns : int
        The number of steps (points) to be generated in realizations of
        the increments of the fractional Brownian motion. This must be
        the same value as supplied to ``field_1d_predef_setup`` when
        calculating the eigenvalues of the embedding matrix.

        Note: in the context of fractional Brownian motion, `ns`
        represents the number of `steps` from a zero starting state.
        Realizations returned in `z` include this starting state and so
        `ns`+1 values are returned for each realization.

    s : int
        S, the number of realizations of the fractional Brownian motion
        to simulate.

    m : int
        The size, M, of the embedding matrix, as returned by
        ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    xmax : float
        The upper bound for the interval over which the fractional
        Brownian motion is to be simulated, as input to
        ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    h : float
        The Hurst parameter, H, for the fractional Brownian motion. This
        must be the same value as supplied to ``field_1d_predef_setup``
        in params[0], when the eigenvalues of the embedding matrix were
        calculated.

    lam : float, shape (`m`)
        Contains the square roots of the eigenvalues of the embedding
        matrix, as returned by ``field_1d_user_setup`` or
        ``field_1d_predef_setup``.

    rho : float
        Indicates the scaling of the covariance matrix, as returned by
        ``field_1d_user_setup`` or ``field_1d_predef_setup``.

    statecomm : dict, RNG communication object, modified in place
        RNG communication structure.

        This argument must have been initialized by a prior call to
        ``init_repeat`` or ``init_nonrepeat``.

    z : float, shape (`ns`+1, `s`), modified in place
        `On exit`: contains the realizations of the fractional Brownian
        motion, Z. The jth realization, for the ith point `xx`[i-1], is
        stored in `z`[i-1,j-1], for i = 1,2,...,`ns`+1, for j =
        1,2,...,`s`.

    xx : float, shape (`ns`+1), modified in place
        `On exit`: the points at which values of the fractional Brownian
        motion are output. The first point is always zero, and the
        subsequent `ns` points represent the equispaced steps towards
        the last point, `xmax`. Note that in ``field_1d_user_setup`` and
        ``field_1d_predef_setup``, the returned ns sample points are the
        mid-points of the grid returned in `xx` here.

    Raises
    ------
    NagValueError
        (`errno` 1)
            On entry, `ns` = *<value>*.

            Constraint: `ns` >= 1.

        (`errno` 2)
            On entry, `s` = *<value>*.

            Constraint: `s` >= 1.

        (`errno` 3)
            On entry, `m` = *<value>*, and `ns` = *<value>*.

            Constraint: `m` >= max(1, 2*(`ns`-1)).

        (`errno` 4)
            On entry, `xmax` = *<value>*.

            Constraint: `xmax` > 0.0.

        (`errno` 5)
            On entry, `h` = *<value>*.

            Constraint: 0.0 < `h` < 1.0.

        (`errno` 6)
            On entry, at least one element of `lam` was negative.

            Constraint: all elements of `lam` must be non-negative.

        (`errno` 7)
            On entry, `rho` = *<value>*.

            Constraint: 0.0 < `rho` <= 1.0.

        (`errno` 8)
            On entry, `statecomm`['state'] vector has been corrupted or
            not initialized.

    Notes
    -----
    The functions ``field_1d_predef_setup`` and
    ``field_fracbm_generate`` are used to simulate a fractional Brownian
    motion process with Hurst parameter H over an interval [0,x_max],
    using a set of equally spaced points.
    Fractional Brownian motion itself cannot be simulated directly using
    this method, since it is not a stationary Gaussian random field;
    however its increments can be simulated like a stationary Gaussian
    random field.
    The circulant embedding method is described in the documentation for
    ``field_1d_predef_setup``.

    ``field_fracbm_generate`` takes the square roots of the eigenvalues
    of the embedding matrix as returned by ``field_1d_predef_setup``
    when icov1 = 14, and its size M, as input and outputs S realizations
    of the fractional Brownian motion in Z.

    One of the initialization functions ``init_repeat`` (for a
    repeatable sequence if computed sequentially) or ``init_nonrepeat``
    (for a non-repeatable sequence) must be called prior to the first
    call to ``field_fracbm_generate``.

    References
    ----------
    Dietrich, C R and Newsam, G N, 1997, `Fast and exact simulation of
    stationary Gaussian processes through circulant embedding of the
    covariance matrix`, SIAM J. Sci. Comput. (18), 1088--1107

    Schlather, M, 1999, `Introduction to positive definite functions and
    to unconditional simulation of random fields`, Technical Report ST
    99--10, Lancaster University

    Wood, A T A and Chan, G, 1994, `Simulation of stationary Gaussian
    processes in [0,1]^d`, Journal of Computational and Graphical
    Statistics (3(4)), 409--432
    """

    excs = []
    fun_name = 'naginterfaces.base.rand.field_fracbm_generate'

    def cb_py_c_print_rec(en_data, _, rec, length_rec, ifail): # pragma: no cover
        # pylint: disable=missing-docstring
        try:
            utils._py_print_rec(
                en_data,
                -1,
                rec[:utils._EngineIntScalarType.to_py(
                    length_rec.contents
                )],
                None,
                excs,
                ifail.contents,
                logger=_logging.getLogger(fun_name),
            )
        except:
            utils._print_err(
                _sys.exc_info(),
                excs,
                en_data,
                ifail.contents,
            )

    engine_ctypes = utils._EngineState._get_symbol('g05ztft')
    engine_ctypes.restype = None
    engine_ctypes.argtypes = (
        utils._BYREF_ENGINEDATA,
        utils._BYREF_INT, # Input, ns.
        utils._BYREF_INT, # Input, s.
        utils._BYREF_INT, # Input, m.
        utils._BYREF_FLOAT64, # Input, xmax.
        utils._BYREF_FLOAT64, # Input, h.
        utils._EngineFloat64ArrayType.c_type, # Input, lam[m].
        utils._BYREF_FLOAT64, # Input, rho.
        utils._EngineIntArrayType.c_type, # Input/Output, state[lstate].
        utils._BYREF_INT, # Input, lstate.
        utils._EngineFloat64ArrayType.c_type, # Output, z[ns+1, s].
        utils._EngineFloat64ArrayType.c_type, # Output, xx[ns+1].
        utils._EngineErrbuf,
        utils._BYREF_INT, # Output, ifail.
        utils._EngineState._chrlen_ctype, # Input, hidden chrlen of errbuf.
    )

    _m_sorder = utils._get_mutual_sorder(
        {'fun_name': fun_name},
        ((z, 'z'),),
    )
    _en_data = utils._EngineData(py_c_print_rec=cb_py_c_print_rec, sorder=_m_sorder)
    _ns = utils._EngineIntScalarType(
        ns, {'fun_name': fun_name, 'entity': 'ns'},
    ).to_ctypes()
    _s = utils._EngineIntScalarType(
        s, {'fun_name': fun_name, 'entity': 's'},
    ).to_ctypes()
    _m = utils._EngineIntScalarType(
        m, {'fun_name': fun_name, 'entity': 'm'},
    ).to_ctypes()
    _xmax = utils._EngineFloat64ScalarType(
        xmax, {'fun_name': fun_name, 'entity': 'xmax'},
    ).to_ctypes()
    _h = utils._EngineFloat64ScalarType(
        h, {'fun_name': fun_name, 'entity': 'h'},
    ).to_ctypes()
    _m_py = utils._EngineIntScalarType.to_py(_m)
    _lam = utils._EngineFloat64ArrayType(
        lam, {'fun_name': fun_name, 'entity': 'lam'},
        exp_shape=(_m_py,),
    ).to_ctypes()
    _rho = utils._EngineFloat64ScalarType(
        rho, {'fun_name': fun_name, 'entity': 'rho'},
    ).to_ctypes()
    utils._check_comm(
        statecomm, {'fun_name': fun_name, 'entity': 'statecomm'},
        req_keys=('state',),
    )
    _lstate_py = utils._EngineIntArrayType.get_shape(statecomm['state'], {'fun_name': fun_name, 'entity': 'statecomm[\'state\']'}, exp_rank=1)[0]
    _lstate = utils._EngineIntScalarType(
        _lstate_py, {'fun_name': fun_name, 'entity': '_lstate'},
    ).to_ctypes()
    _state = statecomm['state']
    _ns_py = utils._EngineIntScalarType.to_py(_ns)
    _s_py = utils._EngineIntScalarType.to_py(_s)
    _z_cls = utils._EngineFloat64ArrayType(
        z, {'fun_name': fun_name, 'entity': 'z'},
        exp_shape=(_ns_py+1, _s_py),
        mutual_sorder=_m_sorder,
        check_contents=False,
    )
    _z = _z_cls.to_ctypes(can_reference_contents=False)
    _xx_cls = utils._EngineFloat64ArrayType(
        xx, {'fun_name': fun_name, 'entity': 'xx'},
        exp_shape=(_ns_py+1,),
        check_contents=False,
    )
    _xx = _xx_cls.to_ctypes(can_reference_contents=False)
    _errbuf = utils._EngineErrbuf()
    _ierr = utils._EngineIntScalarType.empty_ctypes()
    engine_ctypes(
        _en_data,
        _ns,
        _s,
        _m,
        _xmax,
        _h,
        _lam,
        _rho,
        _state,
        _lstate,
        _z,
        _xx,
        _errbuf,
        _ierr,
        _errbuf._length_,
    )
    utils._handle_cb_exception(excs)
    _ierr_py = utils._EngineIntScalarType.to_py(_ierr)
    if _ierr_py == utils._IERR_SUCCESS:
        _err_class = utils._EngineNoErrorExit()
    else:
        _ifmt, _eb_data = utils._EngineState._handle_global_errors(fun_name, _errbuf, _ierr_py)
        if _ierr_py == 1 and _ifmt == 1:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: ns >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 2 and _ifmt == 2:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, s = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: s >= 1.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 3 and _ifmt == 3:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, m = {:s}, and ns = {:s}.'.format(
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[1]),
                        utils._STR_FMT_INT_AS_STR.format(_eb_data[2]),
                    ),
                    'Constraint: m >= max(1, 2*(ns-1)).',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 4 and _ifmt == 4:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, xmax = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: xmax > 0.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 5 and _ifmt == 5:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, h = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < h < 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 6 and _ifmt == 6:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, at least one element of lam was negative.',
                    'Constraint: all elements of lam must be non-negative.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 7 and _ifmt == 7:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, rho = {:s}.'.format(
                        utils._STR_FMT_FLOAT_AS_STR.format(_eb_data[1]),
                    ),
                    'Constraint: 0.0 < rho <= 1.0.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        elif _ierr_py == 8 and _ifmt == 8:
            _err_class = utils._EngineErrorExit(
                fun_name=fun_name,
                msg_lines=[
                    'On entry, statecomm[\'state\'] vector has been corrupted or not initialized.',
                ],
                errcodes=(_ierr_py, _ierr_py, _ifmt),
            )
        else:
            _err_class = utils._EngineUnexpectedExit(
                fun_name=fun_name,
                errcodes=(utils._IERR_UNEXPECTED, _ierr_py, _ifmt),
            )
    if isinstance(_err_class, utils._EngineNoErrorExit):
        _z_cls.refresh_py(_z)
        _xx_cls.refresh_py(_xx)
        return_data = None
    _err_class.manage()
    return return_data
